#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Analyzer.pm 806 2011-07-14 14:33:28Z tm $
#---------------------------------------------------------------------
# Noark::Analyzer - Framework for Analyzing Noark deposits
#
# Author: Thomas Malt           <tm@linpro.no>
#
# Copyright (C) 2003-2011  Riksarkivaren.
# Copyright (C) 2003-2011  Thomas Malt
#
#---------------------------------------------------------------------

=head1 NAME

Noark::Analyzer - Framework for Analyzing Noark deposits

=head1 SYNOPSIS

my $an = Noark::Analyzer->new();

=head1 DESCRIPTION

This modules contains functionality for analyzing the content of a
Noark deposit. Different types of integrity checks can be implemented.

=cut

package Noark::Analyzer;

use strict;
use warnings;

use Carp qw(carp cluck confess croak);
use Carp::Assert;
use Data::Dumper;
use File::Basename;
use File::Find::Rule;
use File::Spec;
use File::Spec::Win32;	     # Not automatically loaded by File::Spec.
use List::MoreUtils qw(first_index pairwise);
use POSIX qw(ceil);
use Time::HiRes qw(time);
use XML::Checker::Parser;
use XML::Twig;

use Noark::Report;
use Noark::Config qw($dtds);
use Noark::Parser;
use Noark::Error;
use Noark::Error::BadOrderMsg;
use Noark::Error::XMLMissingMandatoryAttribute;

use base ('Noark::Parser');

my $ISOK = 1;

# All global data used by self should end up here:
my %of_self = (
    SELF              => undef, 
    CURR              => [],
    XMLTITABELL       => undef, # Contains the current TI.TABELL:
    CURRENT_FILE      => undef,
); 

# Handlers for the Checker::Parser... element start.
my %SAX_ELEM_HANDLERS = (
    'JOURNPOST.JO' => \&handle_journpost_jo,
    'NOARKSAK.JO'  => \&handle_noarksak_jo,
    'NOARKSAK.SD'  => \&handle_noarksak_sd,
    'JOURNPOST.SD' => \&handle_journpost_sd,
);

# Handlers for character data..
my %SAX_CHAR_HANDLERS = (
    'TI.TABELL'    => \&handle_sax_ti_tabell,
    'TI.ANTPOSTER' => \&handle_sax_ti_antposter,
    'JP.JAAR'      => \&handle_sax_jp_jaar,
    'SA.SAAR'      => \&handle_sax_sa_saar,
    'VE.FILREF'    => \&handle_sax_ve_filref, 
    'VE.DOKFORMAT' => \&handle_sax_ve_dokformat,
);

# Stack of elements open in the Checker::Parser;
my @SAXSTACK = ();

# Stack of CDATA used by the character data handlers.  At every start
# element an array reference is pushed onto the stack, and at every
# end element the topmost element gets popped.
my @SAX_CDATA = ();

use IO::Handle;
open my $alog, '>', '/tmp/analyzer.'.$$.'.log' or die "could not open logfile";
$alog->autoflush(1);

print $alog "-" x 74, "\n";
print $alog localtime().": Initializing analyzer log file\n";
print $alog "-" x 74, "\n";


=head1 FUNCTIONS

=over

=item new (%args)

Constructor. Returnes a blessed analyzer object.

Arguments:
  debug            => INT  - Debug level
  verbose          => INT  - Verbosity level.
  do_fkcheck       => BOOL - Foreign key check.
  validate_reports => BOOL - Should be run validation on reports.
  do_doc_ref_check => BOOL - Check that documents that are referred to exists.

=cut
sub new {
    my ($proto, %args) = @_;

    my $class = ref($proto) || $proto;
    my $self  = {};

    print "\n  Initializing Parser object ... " 
        if defined $args{verbose} && $args{verbose} > 2;

    # Call the construtior of Noark::Parser
    $self = $class->SUPER::new(%args);

    print "  Done.\n" 
        if defined $args{verbose} && $args{verbose} > 2;

    $self->{DTDDIR}            = dirname($INC{'Noark/Analyzer.pm'}).'/DTD';
    $self->{VAL_REPS}          = $args{validate_reports};
    $self->{VALIDATE_DOC_REFS} = $args{do_doc_ref_check};

    # Counters:
    $self->{'DTDMISSING'}     = 0;
    $self->{'FILECOUNTERROR'} = 0;
    $self->{'FATALS'}         = 0;
    $self->{'WARNINGS'}       = 0;

    # Containers:
    $self->{'XMLERR'}         = {};
    $self->{'DATAERR'}        = $self->get_data_errorlist();
    $self->{'EI'}             = undef; # Reference to parsed exportinfo HASH.
    $self->{'JOURNPOST'}      = [];
    $self->{'FILVER'}	      = [];
    # $self->{'ERRMSG'}         = \%errmsg;
    $self->{'EXTRAFILES'}     = [];
    $self->{'MISSINGFILES'}   = [];
    $self->{'NOT_DEPOSITED'}  = [];
    $self->{'METAINFO'}       = {}; # Storage for various statistical info.
    $self->{'TABLEMAP'}       = $self->map_tablenames();
    $self->{'DIR_MAP'}	      = {}; # Between dirnames in deposit and on disk.
    $of_self{SELF} = $self;

    return $self;
}



=item is_noarkih_valid

Checks noarkih for validity

=cut
sub is_noarkih_valid {
    my ($self) =  @_;
    return 1 if not $self->{VAL_REPS};

    my $filemap = $self->get_filemap();
    my $start   = time();
    my $progress = $self->get_progress();
    my $string  = "  Is NOARKIH.XML valid";

    $progress->setup_status(time => $start, string => $string);

    if (!$filemap->is_file_in_deposit('NOARKIH.XML')) {
	    return undef if ! $self->_handle_no_noarkih_error();
    }

    $of_self{CURRENT_FILE} = 'NOARKIH.XML';
    $progress->print_status() if $self->{VERB} > 1;

    my $valid = $self->is_file_valid(
        $filemap->get_file_by_name($of_self{CURRENT_FILE})
    );

    if (! $valid) {
	push(
            @{$of_self{CURR}}, 
            { 
                name => 'FAIL',
                text => ('NOARKIH.XML er ikke gyldig XML i henhold '.
                         'til Noark-4.1.')
            }
        );

	push(@{$self->{'XMLERR'}->{'NOARKIH'}}, @{$of_self{CURR}});

	$self->{'FATALS'} = 0;
	$self->{'WARNINGS'} += 1;
    }

    $self->empty_current_errorlist();

    $progress->print_status() if $self->{VERB} > 1;
    $progress->print_rapport_verb(
        is_valid   => $valid,
        start_time => $start
    ) if $self->{VERB};

    return $valid;
}

=item is_dokver_valid ()

Validates the DOKVER file(s).

=cut
sub is_dokver_valid {
    my ($self) = @_;
    return 1 if not $self->{VAL_REPS};

    my $start = time();
    my $progress = $self->get_progress();

    $progress->setup_status(time   => $start,
				            string => '  Is JOURNAL valid');

    # Det kan vre fler enn en journalfil.
    my $files = $self->get_files_by_doctype("DOKVER");
    $self->{'METAINFO'}->{'JOURNAL'} = {JOURNPOST => {COUNT => 0}};

    # Hvis vi ikke finner noen filer er det en feil.
    if (!scalar(@{$files})) {
	$self->_add_missing_journal_error();
	return undef;
    }

    $progress->print_status() if $self->{VERB} > 1;

    my $valid = 1; # Checker to see if  both files are valid.
    for my $file (@{$files}) {
        $of_self{CURRENT_FILE} = $file;
        my $filename = basename($file);
        if (! $self->is_file_valid($file)) {
            $self->add_xmlerror('JOURNAL', 0,
                                name => 'FAIL',
                                text => "'".$filename."' er ikke gyldig XML i "
                                      . "henhold til Noark-4.1."); 
            
            push(@{$self->{'XMLERR'}->{'JOURNAL'}}, @{$of_self{CURR}});
            $self->{'WARNINGS'} += 1;
            $valid = 0;
        }
    }

    $self->empty_current_errorlist(); 

    $progress->print_status() if $self->{VERB} > 1;
    $progress->print_rapport_verb(start_time => $start,
				                  is_valid   => $valid) if $self->{VERB};

    return $valid;
}

=item is_journal_valid ()

Validates the JOURNAL file(s)

=cut
sub is_journal_valid {
    my ($self) = @_;
    return 1 if not $self->{VAL_REPS};

    my $filemap  = $self->get_filemap();
    my $start    = time();
    my $progress = $self->get_progress();

    $progress->setup_status(
        time   => $start,
        string => '  Is JOURNAL valid'
    );
    
    # Det kan vre fler enn en journalfil.
    my $files = $filemap->get_files_by_doctype("JOURNAL");
    $self->{'METAINFO'}->{'JOURNAL'} = {JOURNPOST => {COUNT => 0}};

    # Hvis vi ikke finner noen filer er det en feil.
    if (!scalar(@{$files})) {
        $self->_add_missing_journal_error();
        $progress->print_rapport_verb(
            is_valid   => 0, 
            start_time => $start
        ) if $self->{VERB};

        return undef;
    }

    $progress->print_status() if $self->{VERB} > 1;

    my $valid = 1; # Checker to see if  both files are valid.
    for my $file (@{$files}) {
        $of_self{CURRENT_FILE} = $file;
        my $filename = basename($file);
        if (! $self->is_file_valid($file)) {
            $self->add_xmlerror(
                'JOURNAL', 
                0,
                name => 'FAIL',
                text => "'".$filename."' er ikke gyldig XML i "
                      . "henhold til Noark-4.1."
            ); 
            
            push(@{$self->{'XMLERR'}->{'JOURNAL'}}, @{$of_self{CURR}});
            $self->{'WARNINGS'} += 1;
            $valid = 0;
        }
    }

    $self->empty_current_errorlist(); 

    $progress->print_status() if $self->{VERB} > 1;
    $progress->print_rapport_verb(
        start_time => $start,
        is_valid   => $valid
    ) if $self->{VERB};

    return $valid;
}



=item is_sakdok_valid

Validates the SAKSDOKOVERSIKT files.

=cut
sub is_sakdok_valid {
    my ($self) = @_;

    return 1 if not $self->{VAL_REPS};

    my $filemap  = $self->get_filemap();
    my $start    = time();
    my $progress = $self->get_progress();
    
    $progress->setup_status(
        time   => $start, 
        string => '  Is SAKDOKOVERSIKT valid'
    );

    my $files = $filemap->get_files_by_doctype("SAKDOKOVERSIKT");

    # Initializing the counter
    $self->{'METAINFO'}->{'SAKDOKOVERSIKT'} = { NOARKSAK  => { COUNT => 0 },
						JOURNPOST => { COUNT => 0 },
						FILREF    => { COUNT => 0 },
					      };

    if (! scalar(@{$files})) {
        $self->_add_sakdok_files_warning();
        $progress->print_rapport_verb(
            is_valid        => 0, 
            start_time      => $start,
            noarksak_count  => $self->{'METAINFO'}->{'SAKDOKOVERIKT'}->{'NOARKSAK'}->{'COUNT'},
            journpost_count => $self->{'METAINFO'}->{'SAKDOKOVERIKT'}->{'JOURNPOST'}->{'COUNT'},
        ) if $self->{VERB};
        return;
    }

    my $valid = 1;
    for my $file (@{$files}) {
        $of_self{CURRENT_FILE} = $file;
        my $filename = basename($file);
        if (! $self->is_file_valid($file)) {
            $self->add_xmlerror(
                'SAKDOKOVERSIKT', 0,
                name => 'FAIL',
                text => ("'".$filename."' er ikke gyldig XML i"
                      . " henhold til Noark-4.1."),
                hint => ("Jamfr Noark-4.1 Kap. 15.3.4.")
            );
            push(@{$self->{'XMLERR'}->{'SAKDOKOVERSIKT'}}, @{$of_self{CURR}});
            $self->{'WARNINGS'} += 1;
            $valid = 0;
        }
    }

    $self->empty_current_errorlist(); 

    $progress->print_rapport_verb(
        is_valid   => $valid,
        start_time => $start,
        noarksak_count  => $self->{'METAINFO'}->{'SAKDOKOVERIKT'}->{'NOARKSAK'}->{'COUNT'},
        journpost_count => $self->{'METAINFO'}->{'SAKDOKOVERIKT'}->{'JOURNPOST'}->{'COUNT'},
	) if $self->{VERB};

    return $valid;
}


sub _add_missing_journal_error {
    my $self = shift;

    $self->{FATALS} += 1;
    $self->add_xmlerror(
        'JOURNAL', 1,
        name => 'FAIL',
        text => "Vi kunne ikke finne noen fil(er) med DOCTYPE JOURNAL\n"
             . "           i avleveringen. JOURNAL er en av de rapportene "
             . "som skal vre\n           med i en avlevering.\n",
	hint => 'Jamfr Noark-4.1 Kap. 15.3.1, K15.18'
    );

    return 1;
}


sub _handle_no_noarkih_error {
    my $self = shift;

    if ($self->{'VERB'} > 1) {
	print STDERR "Couldn't read file NOARKIH.XML: $!\n";
    }

    my $fc = $self->check_filename('NOARK.IH', 'NOARKIH.XML');

    $self->add_error_noarkihname('NOARKIH.XML', 0, $fc);
    push(@{$self->{XMLERR}->{NOARKIH}}, @{$of_self{CURR}});

    if (not $fc->{'EXISTS'}) {
	$self->empty_current_errorlist();
	$self->{'FATALS'} += 1;
	return undef;
    }

    $self->{'WARNINGS'} += 1;

    return 1;
}



sub _add_sakdok_files_warning {
    my $self = shift;

    $self->{'WARNINGS'} += 1;
    $self->add_xmlerror(
        'SAKDOKOVERSIKT', 1,
        name => 'FAIL',
        text => "Vi kunne ikke finne fil(er) med DOCTYPE SAKDOKOVERSIKT\n"
            . "           i avleveringen. SAKDOKOVERSIKT er en av rapportene som alltid skal\n"
			      . "           vre med i avleveringen.\n",
        hint => ('Jamfr Noark-4.1 Kap. 15.3.1, K15.18.')
       );

    return 1;
}

=item is_file_valid($file)

Runs a given file through the validator.

  param  $file   -- Complete path to file in deposit.
  return BOOLEAN -- TRUE or FALSE

=cut
sub is_file_valid {
    my ($self, $file) = @_;

    my $filemap  = $self->get_filemap();
    my $progress = $self->get_progress();

    assert(-r $file);

    $progress->set_start_time();
    $progress->set_current_size( (stat($file))[7] );

    my %handlers = (Start      => \&handle_element_start,
		    End        => \&handle_element_end,
		    Char       => \&handle_element_cdata,
		   );

    my $parser = XML::Checker::Parser->new(
        KeepCDATA      => 1,
        Handlers       => \%handlers,
        SkipInsignifWS => 1,
    );

    XML::Checker::Parser::set_sgml_search_path($self->{'DTDDIR'});

    $self->{SAX_PARSER} = $parser;

    $ISOK = 1; # Make sure isok = 1 before we start;
    my $MAX_DTD_RETRY = 2;
    for (my $ctr = 0; $ctr < $MAX_DTD_RETRY; $ctr++) {
	eval {
	    local $XML::Checker::FAIL = \&handle_xml_fail;
	    $parser->parsefile ($file);
	};

	return 1 if ($ISOK && !$@);
	return 0 if not ($@ =~ /Couldn.t find external DTD \[(.*)\]/);

        my $key = $filemap->get_key_by_filename($file);
        my $dtd = $self->get_xmlmeta_info($key);

	if (not defined $dtd) {
	    $self->add_error_dtd_missing($key, $1);
	    $self->{'FATALS'} += 1;
	    return 0;
	}

	$self->add_error_dtd_found($key, $dtd);
	$self->{'WARNINGS'} += 1;
	$file = $dtd->{'FILE'};
    }

    return 0;
}



=item get_eksportinfo

Returns an HASHREF with eksportinfo.

=cut
sub get_eksportinfo {
    my $self = shift;

    return $self->{'EI'} if defined($self->{'EI'});

    $self->{'EI'} = $self->SUPER::get_eksportinfo();
    return $self->{'EI'};
}



sub get_current_errorlist {
    my $self = shift;
    return $of_self{CURR};
}

sub empty_current_errorlist {
    my $self = shift;

    $of_self{CURR} = [];
}

=item validate_import

This function parsers through all <TABELLINFO> elements and calls all
subroutines to do validation and integrity_checks on them.

=cut
sub validate_import {
    my $self      = shift;
    my $start     = time();
    my $info_list = $self->get_tabellinfo_list();
    my $filemap   = $self->get_filemap();
    my $progress  = $self->get_progress();
    
    my %fi = ('NOARKIH.XML' => 1,);

    for (@{$filemap->get_filename_keys()}) {
        $fi{$_} = 1 if (/(JOURNAL|SAKDOK)\d*\.XML/);
    }

    $self->check_for_nonexisting_tables();

    my $length   = $info_list->get_tabellinfo_count;
    my $ctr      = 0;
    my $ctr_ok   = 0;

    # TABLE:
    for my $info (@{$info_list}) {
        my $inner = time();
        my $ok    = 1;
        my $name  = $info->get_table_name();

        $ctr += 1;
        print $self->_get_status_head($name, $ctr, $length) if ($self->{VERB});

        $ok = ($ok && $self->check_tabellinfo( $info )) ? 1 : 0;
        $ok = ($ok && $self->are_files_valid( $info ))  ? 1 : 0;

        my $res = undef;
        $res = $self->handle_deposited_table( $info );
        $ok  = ($ok && $res) ? 1 : 0; 

        for my $file (@{$info->get_files_ref}) {
            $fi{$file->{'TI.FILNAVN'}} = 1;
        }

        $self->{XMLERR}->{$name} = $of_self{CURR};
        $self->empty_current_errorlist();
        $ctr_ok += $ok;

        print $self->_get_verb_errors($name) if ($self->{VERB} > 2);
        print $self->_get_status_res($ok, time()-$inner) if ($self->{VERB});
    }

    $self->{EXTRAFILES} = [grep {!$fi{$_}} @{$filemap->get_filename_keys}];
    $self->check_duplicate_tables();

    if ($self->{DO_FKCHECK}) {
        $self->verify_or_repair_foreign_key_integrity(verify => 1);
        $self->verify_or_repair_foreign_key_integrity(repair => 1);
    }

    if ($self->{'VERB'}) {
        print "FAIL: ".($length-$ctr_ok).", OK: $ctr_ok, Total: $ctr -- ";
        print "Fatals: ",$self->{'FATALS'},", Warnings: ", $self->{'WARNINGS'};
        print " -- Time: ". $progress->get_timestr(time()-$start), "\n";
    } 
}


=item are_files_valid ($info)

Parses through the file or files given in info. Stores information
about errors in the class structure for later use.

Returns TRUE or FALSE.

=cut
sub are_files_valid {
    my ($self, $info) = @_;

    my $filemap = $self->get_filemap();
    my $progress = $self->get_progress();

    # Setting up metainfo:
    $self->{METAINFO}->{$info->{'TI.TABELL'}.'.TAB'} = {};

    for my $file_hash (@{$info->get_files_ref}) {
        my $filename = $file_hash->{'TI.FILNAVN'};

        if (! $file_hash->{EXISTS}) {
            $file_hash->{VALID} = 0;
            if ($self->{VERB} > 2) {
                printf "     File: '%s' does NOT exist.\n", $filename;
            }
            next; 
        }
        my $file = ($file_hash->{SAME}) ? uc($filename) 
                                        : uc($file_hash->{BASENAME});
        $of_self{CURRENT_FILE} = $file;
    
        $progress->setup_status(
            string => "     Validating '".$file."'"
        ) if ($self->{VERB} > 1);

        $file_hash->{VALID} = $self->is_file_valid(
            $filemap->get_file_by_name($file)
        );

        $progress->finish_and_print_status(ok => $file_hash->{VALID}) 
            if $self->{VERB}> 1;
        
        if (! $file_hash->{VALID}) {
            push(@{$of_self{CURR}}, {name => 'FAIL', 
                            text => "$file er ikke gyldig XML i henhold ".
                            "til Noark-4.1."});
            last;
        }
    }

    my $all_ok = 1;

    for my $fhash (@{$info->get_files_ref}) {
        $all_ok = 0 if (! $fhash->{VALID});
    }
    return $all_ok;
} 



=item map_tablenames

This functions makes a list of all the valid shortnames, so we can
check against it when validating shortnames in NOARKIH.

  return  HASHREF with the shortnames.

=cut
sub map_tablenames {
    my $self = shift;

    my %names = ();

    for (keys(%{$dtds})) {
	my ($name) = /(\w+)\.TAB/;
	next unless $name;
	$names{$name} = 1;
    }

    return \%names;
}





=item check_filename ($name)

Does various checks on XML filenames to see whether they exist or not.

  param  STRING  $tablename -- Tabellnavnet denne gjelder for.
  param  STRING  $filename  -- Key in FILEMAP to search for.
  param  HASHREF $seen   -- Reference to hash keeping seen  files.
  return HASHREF         -- Result of parsing.

=cut
sub check_filename {
    # TODO Overgang fra egen res-hash til bruk av eget Validator objekt.
    my ($self, $tablename, $filename, $seen) = @_;

    my $filemap = $self->get_filemap();

    $filename = "" if not defined ($filename); # $filename can be empty in
                                               # sertain circumstances.
    my %res = ();

    $res{UPCASE}  = ($filename eq uc($filename)) ? 1 : 0;
    $res{XMLNAME} = $filename;

    if ($filemap->is_file_in_deposit(uc($filename))) {
        $res{EXISTS}   = 1;
        $res{BASENAME} = basename($filemap->get_file_by_name(uc $filename));
    } 
    else {
        $res{EXISTS} = 0;
        for (@{$filemap->get_files_by_doctype($tablename.".TAB")}) {
            next if (defined $seen && $seen->{$_});
            $res{EXISTS} = 1;
            $res{BASENAME} = basename($_);
        }
    }

    $res{BASENAME} = '' if not $res{EXISTS};
    $res{SAME}     = ($res{BASENAME} eq $filename) ? 1 : 0;

    my ($pre,  $suf)  = split(/\./, $filename);
    my ($fpre, $fsuf) = split(/\./, $res{BASENAME});

    $pre  = "" if not defined($pre);
    $suf  = "" if not defined($suf);
    $fpre = "" if not defined($fpre);
    $fsuf = "" if not defined($fsuf);

    $res{'PRELEN'}  = length($pre)  || 0;
    $res{'SUFLEN'}  = length($suf)  || 0;
    $res{'FPRELEN'} = length($fpre) || 0;
    $res{'FSUFLEN'} = length($fsuf) || 0;

    if ($res{'UPCASE'}       && $res{'SAME'}         && $res{'EXISTS'} &&
	$res{'PRELEN'}  <= 8 && $res{'SUFLEN'}  == 3 && $res{'FPRELEN'} <= 8 &&
	$res{'FSUFLEN'} == 3) {
	$res{'OK'} = 1;
    } else {
	$res{'OK'} = 0;
    }
    
    $seen->{uc($filename)} = 1 if defined($seen);
    return \%res;
}


=item is_valid_table_name

Returns true if tabellname is a valid NOARK-4 tabellnavn.

=cut
sub is_valid_table_name {
    my ($self, $name) = @_;

    return 1 if $self->{TABLEMAP}->{$name};
    return 0;
}


=item check_tabellinfo ($ti)

Takes a TABELLINFO HASH as argument and parses it for needed
information.

Generates errormessages where appropriate:
 * If file case is wrong.
 * If the file name is invalid.
 * If the table name is invalid.

Tries to fix the filenames if they are invalid, and a file with the
correct doctype exists in the deposit.

Returns TRUE or FALSE, whether it is OK to continue or not.

=cut
sub check_tabellinfo {
    my ($self, $info) = @_;

    my $is_ok = 1;
    my $name  = $info->get_table_name();
    my $data_errors = $self->get_data_errorlist();

    if (not $self->is_valid_table_name(uc $name)) {
        $data_errors->add_error('NOARKIH', 'Noark::Error::TableName', {
            line  => $info->get_line_number,
            table => $info->get_table_name,
        });
        $self->{'FATALS'} += 1;

        $is_ok = 0; # Fatal error
    } 
    elsif ($name ne uc($name)) {
        $data_errors->add_error('NOARKIH', 'Noark::Error::TableName', {
            line  => $info->get_line_number,
            table => $info->get_table_name,
        });
        $self->{WARNINGS} += 1;
    } 

    # Converting table name to upper case no matter what:
    $name = uc($name);
    $info->set_table_name($name);
    
    my $ih_count = $info->get_ih_number_of_files();
    my $an_count = $info->get_file_count();
    my $line     = $info->get_line_number();
    if ( $ih_count != $an_count ) {        
        $data_errors->add_error('NOARKIH', 'Noark::Error::FileCount', {
            line          => $line,
            table         => $info->get_table_name,
            noarkih_count => $ih_count,
            deposit_count => $an_count,
        });

        $self->{WARNINGS}       += 1;
        $self->{FILECOUNTERROR} += 1;
    }

    # List of files.
    my %SEEN = (); # HASH to keep track of the files we've seen.
    for my $file (@{$info->get_files_ref()}) {
        next if $SEEN{uc($file->{'TI.FILNAVN'})};

	my $filecheck = $self->check_filename($name, $file->{'TI.FILNAVN'},
					      \%SEEN);

        for my $k (keys %{$filecheck}) {
            $file->{$k} = $filecheck->{$k};
        } 
        if (not $filecheck->{'OK'}) {            
            my $t = $self->add_error_filename($name,
                                              $info->get_line_number,
                                              $file);
            $is_ok = ($t && $is_ok) ? 1 : 0;
            unless ($is_ok) {
                push(@{$of_self{CURR}}, 
                     {name => "FAIL",
                      text => ("Kunne ikke finne '".$filecheck->{'XMLNAME'}.
                               "' i avleveringen.")});
            }  
        } 
    }

    return $is_ok; 
}


=item check_for_nonexisting_tables ()

Reads through all tables defined in NOARK.IH and verifies that no illegal,
nonexisting, tables are deposited

Adds an errormessage if an illegal table is found.

=cut
sub check_for_nonexisting_tables {
    my $self = shift;

    my $list           = $self->get_tabellinfo_list();
    my @unknown_tables = sort { 
        $a->get_line_number <=> $b->get_line_number 
    } values %{$list->get_unknown_tabellinfos()};
 
    for my $table (@unknown_tables) {
        my $data_errors = $self->get_data_errorlist();
        $data_errors->add_error('NOARKIH', 'Noark::Error::UnknownTable', {
            line  => $table->get_line_number(),
            table => $table->get_table_name(),
        });
    }

    return 1;
}


=item _add_registered_postnumbers ()

Checks out postnumbers already inserted when inserting postnumbers.

=cut
sub _add_registered_postnumbers {
    my ($self) = @_;

    my %pk_seen = ();
	my $db  = $self->get_db_handle();
	my $sth = $db->prepare_and_execute(qq{SELECT po_postnr FROM postnr});

	while ( my $row = $sth->fetchrow_hashref() ) {
	    $pk_seen{$row->{'po_postnr'}} += 1;
	}
	
	return \%pk_seen;
}

=item handle_deposited_table ($info)

Iterates through the deposit initiating all consistency checks outside
XML-validation.

  param   HASHREF $info  Hashref containing parsed information about files.
  return  UNDEF

=cut
sub handle_deposited_table {
    my ($self, $info) = @_;

    my $table_name       = $info->get_table_name();
    my $start            = time();
    my $is_ok            = 1;
    my $rows_inserted    = 0;
    my %extra_attributes = ();
    my %used_attributes  = ();
    my $file_exists      = 0;
    my $filemap          = $self->get_filemap();
    my $progress         = $self->get_progress();

    # Hashes used for detecting duplicate values in "unique" and
    # "primary key" fields.
    my %pk_seen     = ();	# Stores all primary keys found.
    my %unique_seen = ();	# Store all values that are unique.

    # If current table is POSTNR, make sure we compare foreign keys in
    # the deposit with foreign keys already registered (during
    # database initialization).
    if ($table_name eq 'POSTNR') {
	    %pk_seen = %{$self->_add_registered_postnumbers()};
    }

    # Run through all the files of table in case the table is split into
    # multiple files.
    for my $file (@{$info->{FILES}}) {
        next if ! $file->{EXISTS};
        
        my $filename = uc $file->{BASENAME};

        # Escape in a cloud of green smoke if current file is not found.
        if (! $filemap->is_file_in_deposit($filename)) {
            print "     NO such file: $filename.\n" if ($self->{'VERB'} > 1);
            $is_ok = 0;
            $self->{'FATALS'} += 1;
            next;
        }

        $file_exists = 1;
	    
        my $filepath = $filemap->get_file_by_name($filename);
        my @stat     = stat($filepath);
        
        # Setup progress bar.
        $progress->setup_status( 
            string => sprintf("     Reading '%s'", $filename),
            size   => $file->{'TI.ANTPOSTER'},
        ) if $self->{VERB} > 1;

        # Iterate over the file
        my $rows_count  = 0;
        my @lines_count = ();
        my $filesize    = $stat[7];
        my $ti_poster   = $file->{'TI.ANTPOSTER'} || 0;

        $progress->set_current_string(
            sprintf("     Importing '%s'", $filename)
        );        
        
        my $doc = XML::Twig->new(
            start_tag_handlers => {
                $table_name => sub {
                    my ($t, $elt) = @_;
                    $elt->set_att('#line' => $t->current_line);
                }, 
            },
            twig_handlers => {
                $table_name => sub {
                    my ($t, $elt) = @_;
                    my ($ins_c) = $self->get_data_and_insert($elt, {
                        info        => $info,
                        file        => $filename,
                        extra       => \%extra_attributes,
                        used        => \%used_attributes,
                        pk_seen     => \%pk_seen,
                        unique_seen => \%unique_seen,
                    }) || 0;
                    $is_ok = ($is_ok && $ins_c) ? 1 : 0;
                    $rows_inserted += $ins_c;
                    $rows_count    += 1;
                    $t->purge();
                },
            },
        );

        $doc->parsefile($filepath);
        $doc->purge();
        $doc = undef;        

        $file->{'XML.ANTPOSTER'} = $rows_count;

        # Check for dummy deposit files
        if ($ti_poster == 0 && $rows_count == 1) {
            $self->add_error_antposter_dummy($info, $file, $rows_count);
            return undef;
        }

        # Check for missing / extra deposit rows
        if ($ti_poster != $rows_count) {
            $self->add_error_antposter($info, $file, $rows_count);
            $is_ok = 0;
        }

        $progress->finish_and_print_status(ok => $is_ok) if $self->{VERB}> 1;
    }

    if (%extra_attributes) {
        # Check if attributes are used in FILE but not in NOARKIH.
        $is_ok = 0;
        $self->add_error_attr_extra($table_name, [keys %extra_attributes]);
    }

    if ($file_exists) {
        my @missing = grep { ! $used_attributes{$_} } @{$info->{ATTRIBUTTER}};
        if (@missing) {
            # Check if attributes are in NOARKIH, but not in FILE:
            $is_ok = 0;
            $self->add_error_attr_missing($table_name, \@missing);
	        
            # Check if missing attributes are /cursive/ attributes
            $self->check_cursive_attributes(
                attributes => \@missing,
                table      => $table_name,
                missing    => 1,
	    );
        }   
    }
    
    # Check if any used attributes are /cursive/ attributes.
    $self->check_cursive_attributes(
        attributes => [keys %used_attributes],
        table      => $table_name,
    );


    if ($self->{'VERB'} > 1) {
        $progress->print_traverse_deposit_res(
            reported  => $info->get_total_antposter(),
            inserted  => $rows_inserted,
            start     => $start,
            is_ok     => $is_ok,
        );
    }

    return $is_ok;
}


sub check_cursive_attributes {
    my ($self, %args) = @_; 
    
    my $attributes  = $args{attributes};
    my $table       = $args{table};
    my $missing     = $args{missing} || undef;
    my $schema      = $self->get_noark_schema();
    my $data_errors = $self->get_data_errorlist();
    
    if (! ref $schema eq 'Noark::Schema') {
        croak("Did not get Noark::Schema as expected") 
    }

    for my $att (@{$attributes}) {
        my $answer = $schema->is_attr_exportable($table, $att);

        if (!defined($answer)) {
            $data_errors->add_error($table, 'Noark::Error::UnknownAttribute', {
                attribute => $att,
            });
            $self->{FATALS} += 1;
            next;
        } 
        elsif ($answer) {
            next;
        } 
        else {
            # If we're here we got a cursive attribute.
            $data_errors->add_error($table, 'Noark::Error::CursiveAttribute', {
                attribute => $att,
            });
            $self->{'FATALS'} += 1;
        }
    }
    return 1;
}

=item check_duplicate_tables

=cut
sub check_duplicate_tables {
    my $self = shift;

    my $infolist = $self->get_tabellinfo_list();

    for my $table (@{$infolist}) {
        next if ($table->get_counter < 2);
        my $data_errors = $self->get_data_errorlist();
        $data_errors->add_error(
            'NOARKIH', 
            'Noark::Error::DuplicateTabellinfo', {
                table => $table->get_table_name,
                line  => $table->get_line_number,
            }
        );
        $self->{'WARNINGS'} += 1;

    }

    return 1;
}


=item add_xmlerror($doc, %error)

Adds an error for the document given as argument with the ERROR HASH
given as second argument.

  param STRING $doc
  param BOOL   $single -- Tell if the error should be added in an
			  array or not.
  param HASH   %error

=cut

sub add_xmlerror {
    my ($self, $doc, $single,  %error) = @_;

    if ($single) {
	push(@{$self->{'XMLERR'}->{$doc}},
	     \%error);
    } else {
	push(@{$self->{'XMLERR'}->{$doc}},
	    [\%error]);
    }
    return 1;
}


=item add_error_filename ($name, $line, $fs)

Adds an error to the ERRORS hash for the given file. This error is
added after all the filechecks has been performed, and the output
varies according to the results of those checks.

  param  STRING  $name
  param  INTEGER $line
  param  HASHREF $fs   -- Information from check_filename.

  Contents of the $fs hashref.
    'EXISTS'   => 1
    'BASENAME' => 'UTVBEH.XML'
    'FPRELEN'  => 6
    'FSUFLEN'  => 3
    'OK'       => 1
    'PRELEN'   => 6
    'SAME'     => 1
    'SUFLEN'   => 3
    'UPCASE'   => 1
    'XMLNAME'  => 'UTVBEHANDLING.XML'

  return UNDEF

=cut
sub add_error_filename {
    my ($self, $name, $line, $fs) = @_;

    my $is_ok       = 1;
    my $data_errors = $self->get_data_errorlist();

    if (not $fs->{'UPCASE'}) {
        $data_errors->add_error('NOARKIH', 'Noark::Error::FileCase', {
            file => $fs->{XMLNAME},
            line => $line,
        });

	$self->{'WARNINGS'} += 1;
    }

    if (not $fs->{'SAME'}) {
        if ($fs->{EXISTS}) {
            $data_errors->add_error(
                'NOARKIH', 
                'Noark::Error::MissingFileFound',
                {
                    noarkih_file => $fs->{'TI.FILNAVN'},
                    deposit_file => $fs->{'BASENAME'},
                    line         => $line,
                }
            );
            $self->{'WARNINGS'} += 1;
        } else {
            $data_errors->add_error(
                'NOARKIH',
                'Noark::Error::MissingFile',
                {
                    file => $fs->{'TI.FILNAVN'},
                    line => $line,
                }
            );
            $is_ok = 0;
            $self->{'FATALS'} += 1;                
        }
    } 

    if (not $fs->{'PRELEN'} <= 8 || not $fs->{'SUFLEN'} <= 3) {
        $data_errors->add_error(
            'NOARKIH',
            'Noark::Error::WrongFilename',
            {
                noarkih_file => $fs->{'TI.FILNAVN'},
                deposit_file => $fs->{'BASENAME'},
                line         => $line,
            }
        );
        $self->{'WARNINGS'} += 1;
    }

    return $is_ok;
}


=item add_error_antposter ($info, $fi, $count)

This error is added if the number of rows in NOARK.IH is diffrent from
the number of elements in the deposited files.

=cut
sub add_error_antposter {
    my ($self, $info, $fi, $count) = @_;

    my $data_errors = $self->get_data_errorlist();
    $data_errors->add_error('NOARKIH', 'Noark::Error::RowCount', {
        table => $info->get_table_name,
        line => $info->get_line_number,
        file => $fi->{'TI.FILNAVN'},
        noarkih_count => defined $fi->{'TI.ANTPOSTER'} 
                      ? $fi->{'TI.ANTPOSTER'} : '-',
        file_count => $count,
    });
    
    $self->{'FATALS'} += 1;

    return 1;
}



=item add_error_antposter_dummy ($info, $fi, $count)

This error is added if we suspect they have submittet a dummy XML-file
in the deposit. This is typical for deposits from DocuLive 6.*.

Adds an error to the DATAERR hash, increases WARNINGS by one.

=cut
sub add_error_antposter_dummy {
    my ($self, $info, $fi, $count) = @_;
    
    my $data_errors = $self->get_data_errorlist();
    $data_errors->add_error('NOARKIH', 'Noark::Error::RowCountDummy', {
        table => $info->get_table_name,
        line => $info->get_line_number,
        file => $fi->{'BASENAME'},
    });
    
    $self->{'WARNINGS'} += 1;

    return 1;
}


=item add_error_attr_missing ($table, $missing)

Adds an error to the error-hashes with information about missing
attributes. Kicks in when attributes are defined in NOARKIH, but are
missing in the deposit XML-file.

Denne funksjonen blir kalt en gang under parsingen av NOARKIH.

Arguments:
  $table   - tablename
  $missing - list of missing elements.
Returns 1 when done.

=cut
sub add_error_attr_missing {
    my ($self, $table, $missing) = @_;
    
    # Verify attributes first: If they are all invalid do not get the errors.
    # We pick up errors for unknown attributes elsewhere.
    my $schema = $self->get_noark_schema();
    my @input = grep { 
        $schema->is_attributename_valid($table, $_); 
    } @{$missing};
    
    return 0 if ! @input;
    
    my $data_errors = $self->get_data_errorlist();
    $data_errors->add_error($table, 'Noark::Error::MissingAttribute', {
        attributes => \@input,
        table      => $table,
    });
    $self->{'WARNINGS'} += 1;

    return 1;
}


=item add_error_attr_extra ($fi, $el, $attrlist)

Adds a fatal error when a file contains more attributes than specified in
NOARKIH.XML

Denne funksjonen kan bli kalt for hver post i tabell-fil.

=cut
sub add_error_attr_extra {
    my ($self, $table, $attrlist) = @_;
    
    # Verify attributes first: If they are all invalid do not get the errors.
    # We pick up errors for unknown attributes elsewhere.
    my $schema = $self->get_noark_schema();
    my @input = grep { 
        $schema->is_attributename_valid($table, $_); 
    } @{$attrlist};

    return 0 if ! @input;
    
    my $data_errors = $self->get_data_errorlist();
    $data_errors->add_error($table, 'Noark::Error::ExtraAttribute', {
        attributes => \@input,
        table      => $table,
    });
    $self->{'FATALS'} += 1;

    return 1;
}


=item add_error_xml_order_missing_mand

Adds an error when mandatory elements are missing

=cut
sub add_error_xml_order_missing_mand {
    my ($self, $file, $line, $bo) = @_;
    
    my $error = $self->_get_xml_order_missing_mand_error_object();
    
    my $attr = $bo->get_missing_mand_attr();
    for (keys(%{$attr})) {
        $error->add_attributes($_, $line);
    }
    
    $self->{'FATALS'} += 1;
    return 1;
}

=item _get_xml_order_missing_mand_error_object

=cut
sub _get_xml_order_missing_mand_error_object {
    my $self = shift;
    
    for my $value (@{$of_self{CURR}}) {
        return $value if (ref($value) eq 'Noark::Error::XMLMissingMandatoryAttribute');
    }
    
    my $error = Noark::Error::XMLMissingMandatoryAttribute->new();
    push @{$of_self{CURR}}, $error;
    return $error;
}

sub add_error_xml_order_missing {
    my ($self, $file, $line) = @_;

    my $name  = "Obligatoriske elementer mangler.";
    my $text  = ("Det er mulig elementene bare ".
		 "kommer i feil rekkeflge. ".
		 "Dette er en alvorlig feil som gjr at en eventuell import ".
		 "av systemet vil mislykkes.");

    push(@{$of_self{CURR}}, {name => $name, text => $text, hint => "",
		    FILE => $file, LINE => $line});

    $self->{'FATALS'} += 1;
}


sub add_error_xml_wrong_order {
    my ($self, $file, $line, $bo) = @_;

    my $name = "Elementer i feil rekkeflge";
    my $text = 'Elementer kommer i feil rekkeflge i henhold til Noark-4.1. '
	     . 'Dette gjr at valideringen feiler. Feilen er ikke kritisk, '
	     . 'men genererer en feilmelding. Rekkeflgen var:';

    my $err = Noark::Error->new( name => $name,
				 text => $text,
				 file => $file,
			       );
    $err->set_list_labels('I avlevering', ' ', 'Riktig rekkeflge');
    $err->set_list_label_lengths(24, 1, 24);

    my @ordered = @{$bo->get_ordered_attr()};
    my @deposit = @{$bo->get_found_attr()};

    my %in_dep = ();
    for (@deposit) {
	$in_dep{$_} = 1;
    }

    my %dep_seen = ();
    my @real_dep = ();
    my @real_ord = ();
    while (my $ord = shift @ordered) {
	next if (!$in_dep{$ord}); # Skipping attributes not in deposit
	my $dep = shift @deposit || '';

	if ($dep eq $ord) {
	    push @real_dep, $dep;
	    push @real_ord, $ord;
	    $dep_seen{$dep} = 1;
	    next;
	}

	if ($dep_seen{$ord}) {
	    push @real_dep, '';
	    push @real_ord, $ord;
	    unshift @deposit, $dep;
	}
	else {
	    push @real_dep, $dep;
	    push @real_ord, '';
	    unshift @ordered, $ord;
	    $dep_seen{$dep} = 1;
	}
    }

    for (0..$#real_ord) {
	my $dep = $real_dep[$_];
	my $ord = $real_ord[$_];

	my $yes = $dep eq $ord ? '' : '+';

	$err->add_listing($dep, $yes, $ord);
    }

    push @{$of_self{CURR}}, $err;

    $self->{'WARNINGS'} += 1;

    return 1;
}





=item add_error_noarkihname ($file, $line, $fc)

Adds an error to the ERRORS hash with information about the NOARKIH file.

  param STRING  $file -- filename given.
  param INTEGER $line -- Line number of error.
  param HASHREF $fc   -- HASHREF with data.

=cut
sub add_error_noarkihname {
    my ($self, $file, $line, $fc) = @_;

    my $name = "Feil med filnavnet til NOARKIH.";
    my $text = "";
    my $hint = "";

    if ($fc->{'XMLNAME'} ne $fc->{'BASENAME'}) {
	$text .= ("Vi kunne ikke finne en fil med navn 'NOARKIH.XML' ".
		  "i avleveringen. ");
	$hint .= ("Filnavnet skal skrives med STORE BOKSTAVER. ".
		  "Jamfr Noark-4.1. K15.8. ");
    }
    if ($fc->{'EXISTS'}) {
        $text .= ("Vi fant likevel en fil med navn '".$fc->{'BASENAME'}."' ".
                  "og importen kan fortsette.");        
    } 
    else {
	$text .= ("Vi kunne ikke finne 'NOARKIH.XML' i avleveringen, og ".
		  "importen vil derfor bli avbrutt.");
	$hint .= "NOARKIH.XML m ha riktig filnavn og flge med avleveringen.";
    }

    my %err = (name => $name, text => $text, hint => $hint, LINE => $line,
	       FILE => "NOARKIH.XML");

    push(@{$of_self{CURR}}, \%err);

    return 1 if ($fc->{'EXISTS'});
    return 0;
}


=item  get_frequency_analysis ()

Fetches the data neccessary for doing the frequency analysis.

  return LIST

=cut
sub get_frequency_analysis {
    my $self = shift;

    my $ei = $self->{'EI'};

    my $map     = Noark::Schema::Map->new();
    my $dbh     = $self->get_db_handle();
    my $sa_saar = $map->get_schema_field_name('SA.SAAR');
    my $jp_jaar = $map->get_schema_field_name('JP.JAAR');

    my @saks = $dbh->get_table_rows(
        table => $map->get_schema_table_name('NOARKSAK')
    );

    my @posts = $dbh->get_table_rows(
        table => $map->get_schema_table_name('JOURNPOST')
       );

    my %posts_by_said = ();
    for my $post (@posts) {
        $posts_by_said{$post->{jp_said}} += 1;
    }

    my %sak_aar = ();
    my @sak_jpost = ();
    my @sastatus_u = ();
    for my $sak (@saks) {
        next if $sak->{$sa_saar} == 9999; # skip dummy data.

        $sak_aar{$sak->{$sa_saar}} += 1;
        # my $count = $dbh->get_journpost_by_noarksak_count($sak->{sa_id});
        my $count = $posts_by_said{$sak->{sa_id}} || 0;
        $sak->{arkn4_count} = $count;

        if ($count < 1 || $sak->{sa_antjp} != $count) {
            push @sak_jpost, $sak;
        }

        if ($sak->{sa_status} eq 'U') {
            push @sastatus_u, $sak;
        }
    }

    my %jp = ();
    my @jposts = $dbh->get_table_rows(
        table => $map->get_schema_table_name('JOURNPOST')
    );

    my @jpost_u = grep { $_->{jp_status} eq 'U' } @jposts;
    for my $post (@jposts) {
        next if $post->{$jp_jaar} == 9999; #Skip dummy data
        $jp{$post->{$jp_jaar}} += 1;
    }

    my $jo  = $self->{METAINFO}->{JOURNAL}->{JOURNPOST}->{YEAR};
    my $sds = $self->{METAINFO}->{SAKDOKOVERSIKT}->{NOARKSAK}->{YEAR};
    my $sdj = $self->{METAINFO}->{SAKDOKOVERSIKT}->{JOURNPOST}->{YEAR};

    return {
        noarksak       => \%sak_aar,
        sakjpost       => \@sak_jpost,
        sastatus       => \@sastatus_u,
        jpstatus       => \@jpost_u,
        journpost      => \%jp,
        journal        => $jo,
        sakdok_poster  => $sdj,
        sakdok_saker   => $sds,
        fradato        => $self->__wash_ei_datestr($ei->{'EI.FRADATO'}),
        tildato        => $self->__wash_ei_datestr($ei->{'EI.TILDATO'}),
    };
}


sub get_mandatory_tables_test {
    my ($self) = @_;

    my $schema = $self->get_noark_schema();
    my $list   = $self->get_tabellinfo_list();
    my $seen   = $list->get_deposited_hash();
    my $tables = $schema->get_table_list();

    my @extra   = ();
    my @missing = ();

    for my $table (@$tables) {
        if ($table->is_mandatory()) {
            next if $seen->{$table->get_noark_name()};
            push @missing, $table;
        } 
        else {
            next if not $seen->{$table->get_noark_name()};
            push @extra, $table;
        }
    }

    return {
        extra   => \@extra,
        missing => \@missing,
    };
}

sub do_numbering_test {
    my $self    = shift;
    my $map     = Noark::Schema::Map->new();
    my $dbh     = $self->get_db_handle();
    my $sa_id   = $map->get_schema_field_name('SA.ID');
    my $sa_saar = $map->get_schema_field_name('SA.SAAR');
    my %years   = ();

    my @byid    = $dbh->get_table_rows(
        table => $map->get_schema_table_name('NOARKSAK'), 
        order_by => $sa_id
    );

    for (@byid) {
        next if $_->{$sa_saar} == 9999; 
        # Since we are using the db we are victim of dummy data. 
        push(@{$years{$_->{$sa_saar}}}, $_);
    }

    # Og til slutt en liste med oversikt over alt.
    my %res = (
        FIRSTID => $byid[0]->{$sa_id},
        LASTID  => $byid[$#byid]->{$sa_id},
    );

    my %YEARS = ();
    for (keys(%years)) {
        # Sorter saker per r etter seknr.
        @{$years{$_}} = sort { 
            $a->{sa_seknr} <=> $b->{sa_seknr} 
        } @{$years{$_}};
        
        my $count = scalar(@{$years{$_}});  # Antall saker

        my $firstid = $years{$_}->[0]->{sa_seknr};
        my $lastid  = $years{$_}->[$count-1]->{sa_seknr};

        # Intervall mellom frste og siste ID per r.
        my $idcnt = ($lastid - $firstid)+1 || 1;
        my $freq  = $idcnt/$count;

        my @missing = ();
        my $first = undef;
        my $curr  = undef;
        for (@{$years{$_}}) {
            $curr = $_->{'SA.ID'};
            if (defined $first && $curr > $first+$freq) {
                push(@missing, {FIRST => $first, LAST => $curr});
            }
            $first = $curr;
        }
        
        $YEARS{$_} = {
            MISSING => \@missing,
            COUNT   => $count,
            IDCOUNT => $idcnt,
            IDFREQ  => $freq,
            FIRSTID => $firstid,
            LASTID  => $lastid,
        };
    }

    $res{'YEARS'} = \%YEARS;
    return \%res;
}

sub __wash_ei_datestr {
    my ($self, $str) = @_;

    # Antar dato har formatet dd.mm.YYYY og fikser.
    if ($str =~ /(\d\d)\D(\d\d)\D(\d\d\d\d)/) { 
        return $3.$2.$1; 
    }

    return $str;
}

=item do_interval_test

Cheks for cases that is out of bounds.

=cut
sub do_interval_test {
    my $self = shift;

    my $ei      = $self->get_eksportinfo();
    my $map     = Noark::Schema::Map->new();
    my $dbh     = $self->get_db_handle();
    my $sa_dato = $map->get_schema_field_name('SA.DATO');
    my $sa_saar = $map->get_schema_field_name('SA.SAAR');
    my $fradato = $self->__wash_ei_datestr($ei->{'EI.FRADATO'});
    my $tildato = $self->__wash_ei_datestr($ei->{'EI.TILDATO'});

    # Filterere vekk rader med 9999 som sa_saar. Det er genererete rader
    # som er satt inn for  f avhengigheter til  g opp.
    my @sa  = grep { $_->{$sa_saar} != 9999 } $dbh->get_table_rows(
        table    => $map->get_schema_table_name('NOARKSAK'),
        order_by => $sa_dato,
    );    

    my $fra     = 1;
    my $til     = 1;
    my $fra_ctr = 0;
    my $til_ctr = 0;

    for (@sa) {
        $_->{$sa_dato} =~ s/-//g; # Strip "-" from iso dates from db. 
        if ($_->{$sa_dato} < $fradato) {
            $fra = 0;
            $fra_ctr++;
        }
        if ($_->{$sa_dato} > $tildato) {
            $til = 0;
            $til_ctr++;
        }
    }
        
    my %res = (
        FROM     => $fra, TO => $til,
        FROMCTR  => $fra_ctr, TOCTR => $til_ctr,
        FROMDATE => $sa[0]->{$sa_dato},
        TODATE   => $sa[$#sa]->{$sa_dato},
        EIFROM   => $fradato,
        EITO     => $tildato,
    );

    return %res;
}


=item do_noarksak_test

Compares the number of NOARKSAK entries counted in the deposit versus
the number reported in JOURNAL.XML.

Takes no arguments. Returns an HASHREF with the results.

=cut
sub do_noarksak_test {
    my ($self) = @_;
    my %res = %{$self->{'METAINFO'}};

    $res{'NOARKSAK'}->{'NOARKSAK'}   = scalar(@{$self->{'NOARKSAK'}});
    $res{'JOURNPOST'}->{'JOURNPOST'} = scalar(@{$self->{'JOURNPOST'}});



    return %res;
}


=item do_doccount_test

Counts the number of DOKVERSJON entries found in DOKVERSJON.XML.

Takes no arguments. Returns an HASHREF with the results.

=cut
sub do_doccount_test {
    my ($self) = @_;
    unless ($self->{'DOKVERSJON'}) {
	return (NUMDOCS => 0);
    }

    return (NUMDOCS => scalar(@{$self->{'DOKVERSJON'}}));
}



#=====================================================================
# HANDLERS:
#=====================================================================


=item handle_xml_fail ($code, $msg, %info)

Default handler for failures when validating noarkih.

=cut
sub handle_xml_fail {
    my ($code, $msg, %info) = @_;

    assert($of_self{CURRENT_FILE});

    my %handlers = (154 => \&handle_bad_order,
		    157 => \&handle_undefined_element,);

    if (defined $handlers{$code} && ref($handlers{$code}) =~ /CODE/) {
        $handlers{$code}($of_self{SELF}, $msg, $of_self{CURRENT_FILE}, %info);
    } 
    else {
        croak("Got unhandled XML error: code: '$code', mesg: '$msg'");
    }

    $ISOK = 0;
}


sub handle_element_start {
    my ($self, $element, %attr) = @_;

    push(@SAXSTACK, $element);

    # A new array reference that will hold chunks of CDATA occurring
    # inside the current element.
    push(@SAX_CDATA, []);

    if (defined($SAX_ELEM_HANDLERS{$element})) {
	$SAX_ELEM_HANDLERS{$element}($of_self{SELF}, $element, %attr);
    }
}


# Default element handler for end elements.
sub handle_element_end {
    my ($self, $element) = @_;

    if ($of_self{SELF}->{VERB} > 1) {
        if (scalar(@SAXSTACK) < 4) {
            my $progress = $of_self{SELF}->get_progress();
            $progress->print_status(byte => $self->current_byte());
        }
    }

    $of_self{XMLTITABELL} = undef;

    # Remove the topmost (youngest) element from the SAX CDATA stack.
    my $raw_cdata = pop(@SAX_CDATA);

    # Peek into the topmost element, but don't remove it.  SAX char handlers
    # might depend on the stack to be intact.
    # Note: Not sure why I did it this way instead of just using $element
    #       (tm: 2011-06-01)
    my $el = $SAXSTACK[-1];

    if (defined($SAX_CHAR_HANDLERS{$el})) {
	my $cdata = join("", @$raw_cdata); # Only join if necessary.
        $SAX_CHAR_HANDLERS{$el}($of_self{SELF}, $el, $cdata);
    }

    # Remove the topmost element from the SAX element stack.
    $el = pop(@SAXSTACK);

    if ($element ne $el) {
	    warn("Elements not matching: '$element::$el'");
	    return undef;
    }
    return 1;
}


=item handle_element_cdata

Default handler for CDATA.  Simply adds the $string to the current
CDATA buffer.  The buffer used is always the newest buffer array on
the stack (@SAX_CDATA).

The reason for this is that there is no guarantee that $string always
contains _all_ the CDATA of the current XML element.  Hence, this
handler only accumulates the data while the handler that uses the data
is called when the current XML element is closed (in
handle_element_end).

=cut
sub handle_element_cdata {
    my ($self, $string) = @_;

    # Add $string to the current CDATA buffer (on the top of the
    # stack).  Using the index -1 instead of $#SAX_CDATA gives a
    # speedup of roughly 15 % (for that operation).
    push(@{$SAX_CDATA[-1]}, $string);
}

sub handle_noarksak_jo {
    my ($self, $element, $string) = @_;
    $self->{'METAINFO'}->{'JOURNAL'}->{'NOARKSAK'}->{COUNT} += 1;
}

sub handle_journpost_jo {
    my ($self, $element, $string) = @_;
    $self->{'METAINFO'}->{'JOURNAL'}->{'JOURNPOST'}->{COUNT} += 1;
}


sub handle_noarksak_sd {
    my ($self, $el, $str) = @_;
    $self->{'METAINFO'}->{'SAKDOKOVERSIKT'}->{'NOARKSAK'}->{COUNT} += 1;
}

sub handle_journpost_sd {
    my ($self, $el, $str) = @_;

    my $stack = \@SAXSTACK;
    my $cdata = \@SAX_CDATA;

    my $meta = $self->{METAINFO}->{SAKDOKOVERSIKT};
    $meta->{JOURNPOST}->{COUNT} += 1;
    
}

=item handle_sax_sa_saar ($element, $string)

Fikser tellingen av saker per r i JOURNAL.XML, bare der.

=cut
sub handle_sax_jp_jaar {
    my ($self, $element, $string) = @_;

    my $parent = $SAXSTACK[-2];
    return if (! $parent eq 'JOURNPOST.JO' ||
	       ! $parent eq 'JOURNPOST.SD');

    $self->{METAINFO}->{JOURNAL}->{JOURNPOST}->{YEAR}->{$string}
	+= 1 if ($parent eq 'JOURNPOST.JO');
    $self->{METAINFO}->{SAKDOKOVERSIKT}->{JOURNPOST}->{YEAR}->{$string}
	+= 1 if ($parent eq 'JOURNPOST.SD');

}

sub handle_sax_sa_saar {
    my ($self, $element, $string) = @_;

    my $parent = $SAXSTACK[-2];
    return unless $parent eq 'NOARKSAK.SD';

    $self->{METAINFO}->{SAKDOKOVERSIKT}->{NOARKSAK}->{YEAR}->{$string} += 1;
}


=item handle_sax_ti_tabell ($element, $string)

Character data handler. This is called for cdata in TI.TABELL

=cut
sub handle_sax_ti_tabell {
    my ($self, $element, $string) = @_;

    $of_self{XMLTITABELL} = uc($string);
    return 1;
}


=item handle_sax_ti_antposter (element, $string)

Handler for <TI.ANTPOSTER> in NOARKIH.XML. If we are here we know we
are in NOARK.IH.

Does the following:
 * Counts the number of 'NOARKSAK' and 'JOURNPOST' and stores in
   $self->{'METAINFO'} for later comparison.

=cut
sub handle_sax_ti_antposter {
    my ($self, $element, $string) = @_;

    return undef if ! defined $of_self{XMLTITABELL};
    if (! defined $self->{METAINFO}->{NOARKIH}->{$of_self{XMLTITABELL}}) {
	return undef;
    }
    if (!($string =~ /^\d+$/)) {
	carp("Input from <TI.ANTPOSTER> is not numeric!");
	return undef;
    }

    $self->{'METAINFO'}->{'NOARKIH'}->{$of_self{XMLTITABELL}} += $string;
}



# Several helper functions for handle_sax_ve_filref:
sub _path_win32_to_unix {
    my ($self, $win32_path) = @_;
    
    return File::Spec->join( File::Spec::Win32->splitdir($win32_path) );
}


sub _get_current_filename_and_line {
    my ($self) = @_;

    my $parser = $self->{SAX_PARSER}->{Checker}->{Expat};

    return $parser->current_line, fileparse($parser->base());
}

=item handle_sax_ve_filref (element, $string)

Handler for <VE.FILREF> in SAKDOK.XML (and probably DOKVER.XML).

Does the following:

=cut
sub handle_sax_ve_filref {
    my $root_element = $SAXSTACK[0];
    my %do_for = (
        'SAKDOKOVERSIKT' => 1,
        'DOKVERSJON.TAB' => 1,
    );
    return undef if (not $do_for{$root_element});

    my ($self, $element, $string) = @_;

    my $docdir = $self->{DIR}.'/'.$self->{DOCDIR};
    my $path   = ($string =~ qr/\\/xms) ? $self->_path_win32_to_unix($string)
				      : $string;

    # Consider all paths to be relative (remove '/'-prefix if found).
    $path =~ s/\A\///xms;
    $path = $docdir.'/'.$path;

    my $doc_exists = (-e $path) ? 1 : 0;

    my $meta_root = ($root_element eq 'SAKDOKOVERSIKT')
                  ? $self->{METAINFO}->{'SAKDOKOVERSIKT'}->{FILREF}
                  : $self->{METAINFO}->{'DOKVERSJON.TAB'};

    $meta_root->{FOUND}->{ $doc_exists ? 'TRUE' : 'FALSE' } += 1;
    $meta_root->{COUNT}++;

    push @{$meta_root->{LIST}->{$doc_exists ? 'TRUE':'FALSE'}}, $path;

    if (! $doc_exists) {

        my ($current_line, $current_filename)
            = $self->_get_current_filename_and_line();

        my $table_name = ($root_element eq 'SAKDOKOVERSIKT') ? 'SAKDOKOVERSIKT'
                                                             : 'DOKVERSJON';

        my $data_errors = $self->get_data_errorlist();
        $data_errors->add_error(
            $table_name, 
            'Noark::Error::BrokenDocumentReference',
            {
                file => $current_filename,
                name => $string,
                line => $current_line,
            }
        );
        $self->{FATALS} += 1;
    }
}

sub handle_sax_ve_dokformat {
    my ($self, $element, $string) = @_;

    chomp $string; # Just in case.

    my $root_element = $SAXSTACK[0];

    # This has to be dokversjon.
    #   Maybe we should assert or croak.
    return undef if $root_element ne 'DOKVERSJON.TAB';

    my $meta_root = $self->{METAINFO}->{'DOKVERSJON.TAB'};

    $meta_root->{FILETYPES}->{$string} += 1;
    
    return 1;
}


sub add_error_dtd_found {
    my ($self, $file, $info) = @_;

    my $dtdfile = $self->{'DTDMAP'}->{$info->{'TABLE'}};

    my %err = (name => "'$file' mangler gyldig referanse til DTD.",
	       text => ("Referansen i fila var '".$info->{'DTDREF'}."', mens ".
			"riktig referanse skal vre '$dtdfile'. ".
			"Jamfr Noark-4.1 Kapittel 15.1. ".
			"Systemet prver likevel  validere fila med riktig ".
			"DTD."),
	       hint => ("Avleverte XML-filer skal bruke samme navn p ".
			"DTD'er som Riksarkivaren bruker."),
	       FILE => $file,
	       LINE => $info->{'DTDLINE'});

    push(@{$of_self{CURR}}, \%err);

}

sub add_error_dtd_missing {
    my ($self, $file, $dtd) = @_;

    my %err = (name => "'$file' mangler gyldig referanse til DTD.",
	       text => ("Vi kunne ikke finne noen gyldig referanse til en ".
			"Noark-4.1 DTD i fila (Referansen var: '".
			$dtd."'). Dette tyder p at fila ikke ".
			"er en gyldig Noark-4.1 fil. Dette genererer en ".
			"kritisk feil."),
	       hint => ("Avleverte XML-filer skal bruke samme navn p ".
			"DTD'er som Riksarkivaren bruker."),
	       FILE => $file,
	       LINE => "ikke avgitt");

    push(@{$of_self{CURR}}, \%err);
    $self->{'DTDMISSING'} += 1;
}


sub handle_undefined_element {
    my ($self, $msg, $file, %info) = @_;

    my ($elem) = $msg =~ /\[(.*)\]/;

    my ($name, $text, $hint) = "";
    my $line = $info{'line'};

    $name = "Fant ugyldig element";
    $text = ("Elementet '$elem' ble funnet p linje '$line'. Dette elementet ".
	     "er ikke en del av Noark-4.1.");
    $hint = ("Arkivsystemet m utbedres slik at ugyldige elementer ikke ".
	     "kommer med i avleveringen.");

    push(@{$of_self{CURR}}, {name => $name, text => $text, hint => $hint,
		    FILE => $file, LINE => $line});

    return 0;
}


=item handle_bad_order ($msg)

Bad Order is triggered whenever a mandatory element is missing, or an 
element occurs out of place in the xml. This handler figures out which 
is which and triggers the correct error message.

Tries to produce meaningful output.

=cut
sub handle_bad_order {
    my ($self, $msg, $file, %info) = @_;

    my $bo = Noark::Error::BadOrderMsg->new(msg => $msg) ||
      croak("Did not get a valid BadOrderMsg object: $!");

    if (not $bo->has_found_attr()) {
        $self->add_error_xml_order_missing($file, $info{'line'});
        return 0;
    }

    if ($bo->is_missing_mand_attr()) {
        $self->add_error_xml_order_missing_mand($file, $info{'line'}, $bo);
        return 0;
    }

    $self->add_error_xml_wrong_order($file, $info{'line'}, $bo);
    return 1;
}


=item get_xml_errors_by_name ($name)

Returns an ARRAYREF of all the errors recorded by the XML-Parser for
this table/file.

  param   STRING   $name  Name of table to get errors for.
  return  ARRAYREF        List of errors for table.

=cut
sub get_xml_errors_by_name {
    my ($self, $name) = @_;

    return $self->{'XMLERR'}->{$name} || [];
}

=item get_data_errors_by_name ($name)

Returns an ARRAYREF of all data errors for the given table

  param  STRING    $name Name of table with errors.
  return ARRAYREF  List of errors for table

=cut
sub get_data_errors_by_name {
    my ($self, $name) = @_;
    my $list  = $self->get_data_errorlist();
    my $errors = $list->get_errors_as_arrayref_by_key($name); 

    return $errors;
}


=item get_data_errors ()

Returns an ErrorList of all consistency errors reported;

=cut
sub get_data_errorlist {
    my $self = shift;

    if (ref $self->{DATAERR} eq 'Noark::Parser::ErrorList') {
        return $self->{DATAERR};
    }

    $self->{DATAERR} = Noark::Parser::ErrorList->new();
    return $self->{DATAERR};
}


=item get_extra_files ()

Returns an ARRAYREF of all extra files found in the deposit.

=cut
sub get_extra_files {
    my $self = shift;

    return $self->{'EXTRAFILES'} || [];
}

=item get_docuements_mentioned ()

Does a loopup in the METAINFO hash/object and returns the total count of all
documents referenced in the deposit.

=cut
sub get_documents_mentioned {
    my $self = shift;

    my $docs = $self->{METAINFO}->{'DOKVERSJON.TAB'};
    
    return defined $docs->{COUNT} ? $docs->{COUNT} : 0;
}

=item get_documents_seen ()

Does a lookup and returns the total of all documents found.

=cut
sub get_documents_seen {
    my $self = shift;

    return scalar $self->get_documents_seen_list();
}

=item get_documents_seen_list ()

Returns a list of all the documents that are seen in the deposit regardless 
if they are mentioned in the DATA files or not.

=cut
sub get_documents_seen_list {
    my $self = shift;

    my @list = grep { 
        !/\.svn/ 
    } File::Find::Rule->file()->name('*')
        ->in( $self->{DIR}.'/'.$self->{DOCDIR} );

    return @list;
}

sub get_documents_extra_list {
    my $self = shift;

    my @seen  = $self->get_documents_seen_list;
    my %found = map { uc $_ => 1 } $self->get_documents_found_list; 

    my @extra = ();
    for (@seen) {
        next if $found{uc $_};
        push @extra, $_;
    }

    return @extra;
}

=item get_document_filetypes

Returns the document filetype distribution of the files mentioned in
DOKVERSJON.  Uses the attribute <VE.DOKFORMAT> to sort the filetypes.

VE.DOKFORMAT is an O2 mandatory attribute and must be set. Currently we do not
checke if it contains valid values

Returns an HASHREF.

=cut 
sub get_document_filetypes { 
    my $self = shift; 
    return $self->{METAINFO}->{'DOKVERSJON.TAB'}->{FILETYPES};
}

sub get_documents_found {
    my $self = shift;

    return $self->{METAINFO}->{'DOKVERSJON.TAB'}->{FOUND}->{TRUE} || 0;
}

sub get_documents_missing {
    my $self = shift;

    return $self->{METAINFO}->{'DOKVERSJON.TAB'}->{FOUND}->{FALSE} || 0;
}

sub get_documents_found_list {
    my $self = shift;

    if (defined $self->{METAINFO}->{'DOKVERSJON.TAB'}->{LIST}->{TRUE}) {
        return  @{$self->{METAINFO}->{'DOKVERSJON.TAB'}->{LIST}->{TRUE}}; 
    }

    return ();
}

sub get_documents_missing_list {
    my $self = shift;

    if (defined $self->{METAINFO}->{'DOKVERSJON.TAB'}->{LIST}->{FALSE}) {
        return  @{$self->{METAINFO}->{'DOKVERSJON.TAB'}->{LIST}->{FALSE}}; 
    }

    return ();
}

my $OFFSET = 0;

sub _get_status_head {
    my ($self, $name, $ctr, $len) = @_;

    my $dlen = 24;
    my $slen = length($name);

    $ctr = sprintf("%02i", $ctr);

    my $s = "  [$ctr/$len] Analyzing: $name ";
    $s .= "\n" if ($self->{'VERB'} > 1);

    $OFFSET = length($s);
    return $s;
}

sub _get_status_res {
    my ($self, $ok, $time) = @_;
    my $progress = $self->get_progress();
    
    $OFFSET = 37 if ($self->{'VERB'} > 1);

    my $space = 39-$OFFSET;

    my $s = " " x $space;
    $s .= ($ok) ? "  [OK]" : "[FAIL]";
    $s .= " -- Time: ".$progress->get_timestr($time);
    $s .= " [F:".$self->{'FATALS'}."/W:".$self->{'WARNINGS'};
    $s .= "/S:".scalar(@{$self->{NOT_DEPOSITED}})."]\n";

    $OFFSET = 0;
    return $s;
}

=item _get_verb_errors ($name)

Helper method to assist pretty verbose output.
Takes name of table as argument (ie: "NOARKSAK")

Returns formatted STRING.

=cut
sub _get_verb_errors {
    my ($self, $name) = @_;

    my $mesg   = "";
    my @errors = grep {
	!($_->{'name'} =~ /(FAIL|OK)/)
    } @{$self->get_xml_errors_by_name($name)};

    push(@errors, @{$self->get_data_errors_by_name($name)});

    return $mesg unless (scalar(@errors));

    $mesg .= " " x 5;
    $mesg .= "Feilmelding:\n";

    for (@errors) {
	next if ($_->{'name'} =~ /(FAIL|OK)/);

	$mesg .= " " x 7;
	$mesg .= $_->{'name'}."\n";

	my @words = split(/ /, $_->{'text'});
	my $line = " " x 9;

	for my $word (@words) {
	    $line .= ' '.$word;
	    if (length($line) > 66) {
		$mesg .= $line."\n";
		$line = " " x 9;
	    }
	}

	$mesg .= $line."\n";
    }

    return $mesg;
}

1;

__END__

=back

=head1 SEE ALSO

=over

=item Noark::Analyzer

=item Noark::Db

=item Noark::Report

=item Noark::Parser

=back

=head1 AUTHOR

 Thomas Malt <tm@linpro.no>
 Kjell-Magne ierud <kjellm@linpro.no>

=head1 COPYRIGHT

=over

=item Copyright (c) 2003-2011  Riksarkivaren

=back

=head1 LICENSE

Whatever the licence Riksarkivaren choose to put this under.

=cut
