#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Config.pm 587 2011-06-01 13:01:33Z tm $
#---------------------------------------------------------------------
# Noark::Config - Maintainer of configuration files.
#
# Author: Thomas Malt <thomas@malt.no>
#
# Copyright (C) 2003-2011  Riksarkivaren
# Copyright (C) 2003-2011  Thomas Malt <thomas@malt.no>
#
#---------------------------------------------------------------------

package Noark::Config;

=head1 NAME

Noark::Config - Reads and give access to the Noark configuration files.

=head1 SYNOPSIS

To use both $config and $tables:

  use Noark::Config qw($config $tables);

To only use $tables:

  use Noark::Config qw($tables);

=head1 DESCRIPTION

This module gives you access to the objects 
B<$config> and 
B<$tables>.
$config contains all the configuration variables available, and $table
is a map between Noark-4.1 XML tablenames and the tablenames in your
spesific implementation of the Noark 4 database.

Noark::Config first tries to look in the standing directory, from the
perspective of the user, looking for the files B<importer.conf> and
B<tablemap.conf>. If it can't find valid configuration files in that
directory it tries to look in B</etc/noark/>.

The configuration files themselves should be fairly self explanatory.

=head1 CONFIGURATION OPTIONS

=head2 Database connection

Variables for setting up the database connection correctly.

=over

=item B<DB_HOST> - Hostname of database server

=item B<DB_NAME> - Name of database on server

=item B<DB_USER> - Name of user on database

=item B<DB_PASS> - Password of user on database

=back

=head2 Log directories

Tell where importer should shave its log reports.

=over

=item B<LOG_DIR>    - Directory to save log files.

=item B<IMPORT_LOG> - Filename for import log.

=item B<ERROR_LOG>  - Filename for error log.

=item B<DEBUG_LOG>  - Filename for debug log.

=item B<ANALOG>     - :) Filename for analyze log.

=back

=head2 Use syslog

Tell if we should report state to syslog or not. Nice if importer is
run automatically from cron or a web-interface.

=over

=item B<USE_SYSLOG> - 1 or 0, TRUE or FALSE, yes or no.

=back

=head2 Integrity routines

It might be a good idea to be able to turn some integrity routines
on and off. Put a comma separated list of routines to run here.

=over

=item B<INTEGRITY_CHECKS> - list of checks to run.

=back

=head2 Fallback location of data

This is a fallback location for NOARK-data. (XML-avleveringer).  If
the user does not tell where files are on the command line This is
used.

=over

=item B<IMPORTDIR> - Complete PATH to directory containing files.

=back

=cut


use strict;
use warnings;

use Exporter;
use File::Basename;
use Carp;

my $GLOBALDIR = '/etc/arkn4';

my $pwd = qx|pwd|;
chomp($pwd);

our @LOOKDIRS = ($pwd, $ENV{HOME}, '/etc/arkn4', __get_config_dir());


our $config;
our $tables;
our $dtds;

our @ISA;
our @EXPORT_OK;
our ($VERSION) = '$Revision: 587 $' =~ /Revision: (\S+)/;

@ISA       = qw(Exporter); 
@EXPORT_OK = qw($config $tables $dtds $VERSION parse_config);


#---------------------------------------------------------------------
# Preparing config files:

$config = parse_config(get_file_str('arkn4.conf')); 
$tables = parse_config(get_file_str('tablemap.conf'));
$dtds   = parse_config(get_file_str('dtdmap.conf'));


#---------------------------------------------------------------------
# SUBROUTINES
#---------------------------------------------------------------------

=head1 SUBROUTINES

=over

=item parse_config ($file)

Parses any configuration file and returns it as an HASHREF

  return  HASHREF
  param   STRING $file

=cut
sub parse_config {
    my ($file) = @_;
    
    my %res = ();
    open(my $F, $file) or die("Couldn't open file: $file: $!");
    for (<$F>) {
        next if (/^(#|\s*$)/);

        chomp;
        my ($key, $val) = split(/\s*=\s*/);

        # Strip leading space from key:
        $key =~ s/^\s+//;
        $res{$key} = $val;         
    }
    close($F) or croak("Couldn't close file: $file: $!");
    
    # Store config filename in hash. good for debug.
    # $res{__CONFIGFILE} = $file;
    return  \%res;
}


sub get_file_str {
    my $file = shift;

    my $msg = "";
    for (@LOOKDIRS) {
        my $ifi = $file;
        return "$_/$ifi" if -r "$_/$ifi";
        $msg .= "$_/$ifi, ";
        $ifi = ".".$ifi if ($_ eq $ENV{HOME});
        return "$_/$ifi" if -r "$_/$ifi";
        $msg .= "$_/$ifi, ";
    }
    $msg =~ s/,\s*$//;

    croak("Could not find configuration file.\n".
          "Tried looking for '$msg'\n".
          "Make sure you have copied a configuration file to ".
          "the correct location");
}

sub __get_config_dir {
    return dirname($INC{'Noark/Config.pm'})."/Config";
}

=back

=head1 SEE ALSO

=over

=item Noark::Analyzer

=item Noark::Db

=item Noark::Report

=item Noark::Parser

=back

=head1 AUTHOR

Thomas Malt <thomas@malt.no>

=head1 COPYRIGHT

=over

=item Copyright (C) 2003  Riksarkivaren

=back

=head1 LICENSE

Whatever the licence Riksarkivaren choose to put this under.

=cut


1;

__END__

#---------------------------------------------------------------------
# $Log$
# Revision 1.8  2005/11/03 12:56:36  tm
# Siste endringer
#
# Revision 1.7  2005/10/06 13:50:41  tm
# Fikset bugger .Siste oppdateringer.
#
# Revision 1.6  2005/10/04 14:01:02  tm
# Cleanup and fixes
#
# Revision 1.5  2005/10/04 13:42:34  tm
# fixes
#
# Revision 1.4  2005/10/03 14:05:05  tm
# changes
#
# Revision 1.3  2005/10/03 13:14:12  tm
# Better error handling.
#
# Revision 1.2  2005/09/29 14:21:40  tm
# Updated arkn4.conf
#
# Revision 1.1  2005/09/29 13:55:14  tm
# Added all files to repository
#
# Revision 1.8  2005/03/17 11:56:28  tm
# Added import_order file.
#
# Revision 1.7  2005/03/14 10:54:45  tm
# Oppdateringer.
#
# Revision 1.6  2005/03/14 09:53:12  tm
# Fine endringer.
#
# Revision 1.5  2005/02/07 09:31:08  elark
# Endringer gjort hjemme.
#
# Revision 1.4  2004/12/17 11:54:43  elark
# Siste endringer.
#
# Revision 1.3  2004/12/14 16:00:04  elark
# Dagens arbeid.
#
# Revision 1.2  2004/08/20 15:05:50  elark
# Siste versjoner av kode.
#
# Revision 1.1  2004/08/12 14:45:27  elark
# Fixing where its broke.
#
# Revision 1.5  2003/10/15 12:14:48  tm
# Ny versjon av importer og moduler. Frste versjon jeg regner som ferdig.
#
# Revision 1.4  2003/10/13 11:06:13  tm
# Added support for configuration of map between NOARK.XML and database
# table names.
#
# Revision 1.3  2003/10/10 16:33:09  tm
# Various changes hardening the code. More logging is added. Next step is
# real database interaction.
#
# Revision 1.2  2003/10/07 09:34:33  tm
# Added Noark::Parser object. Noark::Parser is a wrapper around the
# XML::DOM::Parser object.
#
# Revision 1.1  2003/10/06 12:51:01  tm
# First framework of importer.
#
#
