#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Db.pm 781 2011-06-23 14:45:21Z tm $
#---------------------------------------------------------------------
# Noark::Db - A module for communicating with a Mysql NOARK Database.
#
# Authors: Thomas Malt <thomas@malt.no>
#	       Martin Thorsen Ranang <mtr@linpro.no>
#
# Copyright (C) 2003-2009  Riksarkivaren.
#
#---------------------------------------------------------------------


=head1 NAME

Noark::Db - A module for communicating with a Mysql based Noark
database.

=head1 SYNOPSIS

my $db = Noark::Parser->new();

my $sth = $db->prepare_and_execute($sql, @arg);

=head1 DESCRIPTION

This module implements a framework for communicating with a Noark
Database. The module is designed for being easy to expand, and can be
easliy made to handle different database backends.

It should be straight forward to add functionality for handling
connections to most exsisting Relational Database Systems.

=cut


package Noark::Db;

use strict;
use warnings;

use Carp qw(cluck carp confess croak);
use Data::Dumper;
use DBI;
use English;
use Env::Path;
use FindBin;
use IO::Prompt;
use POSIX qw( WIFEXITED );	# ("if-exited".)
use Readonly;

use Noark::Config qw($config); #);
use Noark::Report;
use Noark::Helper;

our ($VERSION) = '$Revision: 781 $' =~ /: (\S+)/;

# Information used to check whether the database is pristine or not.
my $pristine_db_tables = 95;
my $pristine_db_table_len_exceptions = {postnr => 4566,};

Readonly my $LEGAL_DB_NAME => qr/[A-Za-z][A-Za-z0-9_]*/xms;

# Used to implement the Db as a singleton class, so that only one
# database connection is made per process using this module.  Hence,
# the single database connection will be shared between all instances
# of the Db class.  If threading is ever considered, this variable
# must be shared and protected, through locking, against malignant
# use.
my $is_connected = {};

# Returns a connection id, used to implemen singleton per connection
# behavior.
sub _get_connection_id {
    my ($self, $args) = @_;

    my $connection_id = join(':',
			     $args->{host}     || $self->{DB_HOST},
			     $args->{database} || $self->{DB_NAME},
			     $args->{user}     || $self->{DB_USER},);
    return $connection_id;
}

=head1 FUNCTIONS

=over

=item new ({[host         => <hostname>], 
            [database     => <db name>],
            [user         => <username>], 
            [password     => <password>],
            [dont_connect => <boolean>],})

Constructor. Tries to connect to the database set in the config file,
or given on the commandline. Returns a blessed reference to itself, or
undef.

Can also take arguments for connection to the database on construction. If
arguments are given to the subroutine, they will override any defaults.

=cut
sub new {
    my ($proto, $args) = @_;

    # Let's us create an new object both from a call to the Object
    # itself and from a reference to an object.
    my $class = ref($proto) || $proto;
    my $self = {};

    $self->{VERSION} = $VERSION;

    $self->{DB_HOST} = $args->{host}     || $config->{DB_HOST};
    $self->{DB_NAME} = $args->{database} || $config->{DB_NAME};
    $self->{DB_USER} = $args->{user}     || $config->{DB_USER};
    $self->{DB_PASS} = $args->{password} || $config->{DB_PASS};

    bless($self, $class);

    if (! ($args->{dont_connect} || $self->{dbh})) {
	    if (! $self->db_connect($self->{DB_HOST}, $self->{DB_NAME}, $self->{DB_USER}, $self->{DB_PASS})) {
	        confess($OS_ERROR);
	    }
    }

    return $self;
}

=item DESTROY ()

Destructor.  Will close the current connection.

=cut

sub DESTROY {
    my ($self) = @_;

    $self->db_disconnect();
}

=item db_connect ($host, $db, $user, $pass)

=over

=item $host - hostname to connect to

=item $db   - name of database

=item $user - username to connect as

=item $pass - password of user

=back

Connects to the specified database. If no connection can be made
returns undef. If connection is successful returns a B<Noark::Db>
object containing a DBI connection.

=cut
sub db_connect {
    my ($self, $host, $db, $user, $pass) = @_;

    # Create an id uniquely identifying the _requested_ connection.
    # The user's password does not seem relevant for identifying a
    # database connection.
    my $connection_id = $self->_get_connection_id({
						   host => $host,
						   database => $db,
						   user => $user,
						  });

    # If the requested connection already has been made, return it.
    if (! exists $is_connected->{$connection_id}) {
	# Store a reference to the newly created connection.  (Used
	# for the singleton implementation.

	# mysqladmin variables | grep sock
	# | socket                      | /var/lib/mysql/mysql.sock          |
	# Workaround for finding which socket the database uses.
	my $sockstr = '';
	my $sock    = $self->find_socket_for_db($host, $user, $pass) 
	    if $host eq 'localhost';

	$sockstr = ":mysql_socket=".$sock if (defined $sock);

	my $dbh = undef;
	eval {
	    $dbh = DBI->connect("dbi:mysql".$sockstr.":database=$db;host=$host", 
				$user, $pass,
				{ PrintError => 0,
				  RaiseError => 1 });
	};

	# Define a helpful error message.
	my $cannot_connect_message = <<"END_CANNOT_CONNECT";


Unable to connect to the database. The application will exit.
The connection variables used was:

  Username: $user
  Password: $pass
  Database: $db
  Hostname: $host
  Socket..: $sock

Please check to make sure this is correct.

The error message from mysql was:
END_CANNOT_CONNECT

	if ($@) {
	    croak($cannot_connect_message . $DBI::errstr);
	}
	unless ($dbh) {
	    croak($cannot_connect_message . $DBI::errstr);
	}
	
	$is_connected->{$connection_id} = {
            dbh => $dbh,
            refereneces => 0,
        };
    }

    $self->{dbh} = $is_connected->{$connection_id}->{dbh};

    # Increase the reference count for this connection.
    $is_connected->{$connection_id}->{references} += 1;

    return $self->{dbh};
}

=item db_disconnect ()

Decreases the reference count by one and disconnects the "physical"
connection if the reference count equals zero.

=cut
sub db_disconnect {
    my ($self) = @_;

    # Retrieve our connection id.
    my $connection_id = $self->_get_connection_id();

    if ($is_connected->{$connection_id} && $self->{dbh}) {
	# Decrease the reference count for this connection.
	$is_connected->{$connection_id}->{references} -= 1;

	# When the reference count for the connection that THIS object
	# represents reaches 0, the connection will truly be disconnected.
	if ($is_connected->{$connection_id}->{references} == 0) {
	    my $rc  = $self->{dbh}->disconnect;
	    delete $self->{dbh};

	    # Remove the item from the IS_CONNECTED hash.
	    delete $is_connected->{$connection_id}
	}
    }
}

=item find_socket_for_db (<hostname>)

Tries to discover which socket mysql uses when running on same server as
script. DBI sometimes tries the wrong socket

=cut
sub find_socket_for_db {
    my ($self, $host, $user, $password) = @_;

    open(my $F, "/usr/bin/mysqladmin --host=$host "
	 . "--user=$user --password=$password variables |") or
      croak("Got fatal error: $OS_ERROR");
    my $sock = undef;
    while (<$F>) {
        chomp;
        next unless /\|\s+socket\s+\|\s+(\S+)\s+/;
        $sock = $1;
        last;
    };
    close ($F);
    return $sock;
}

=item get_db_handle ()

Get a reference to the db handle in this module

Returns: a reference to the new database handle (DBI)

=cut

sub get_db_handle {
    my ($self) = shift;
    return $self->{dbh};

}


=item get_journpost_by_noarksak_count()

Returns the number of journalposts for a give noarksak

=cut
sub get_journpost_by_noarksak_count {
     my ($self, $sa_id) = @_;
 
     my $sql = qq|SELECT jp_id FROM journpost WHERE jp_said=?|;
     my $sth = $self->prepare_and_execute($sql, $sa_id);
     return $sth->rows;
}


=item get_table_rows ($table)

Returns the content of a table as an array. 

=cut
sub get_table_rows {
    my ($self, %args) = @_;
    
    my @params = ();
    push @params, $args{order_by} if defined $args{order_by};

    my $sql = qq|SELECT * FROM $args{table}|;
    $sql .= qq| ORDER BY ?| if defined $args{order_by};
    
    my $sth = $self->prepare_and_execute($sql, @params);
    
    my @rows = ();
    while (my $row = $sth->fetchrow_hashref()) {
        push @rows, $row;
    }
    return @rows;
}


=item is_table_in_db ($table)

Takes table name as argument. Returns true or false.

=cut
sub is_table_in_db {
    my ($self, $table) = @_;

    my $dbh = $self->get_db_handle();
    my $sql = qq{SELECT * FROM $table};

    eval {
        my $rows = $dbh->do($sql);
    };

    return 0 if $@;
    return 1;
}


sub is_attribute_in_table {
    my ($self, $args) = @_;

    my $attribute = $args->{attribute};
    my $table     = $args->{table};
    my $value     = $args->{value};

    my $sql = sprintf (
        "SELECT $attribute FROM $table WHERE $attribute = %s",
	$self->quote($value)
    );
    my $sth = $self->prepare_and_execute($sql);

    my $rows = $sth->rows();

    return $rows;
    
}

=item prepare_and_execute ($sql, @args) 

Prepare, execute (and log) a sql statement

=cut
sub prepare_and_execute {
    my $self = shift;
    my $sql  = shift;
    my @args = @_;

    my $sth = $self->{dbh}->prepare($sql) or croak( "Prepare failed: $OS_ERROR");

    my $num = 1;
    for my $arg (@args) {
         $sth->bind_param($num, $arg);
         $num += 1;
     }

    # Wrapped this in an eval so the app don't die on an error without
    # us checking it first.
    eval {
        $sth->execute;
    };
    
    return $sth if ! $@;
    
    if ($@ =~ /Duplicate.*entry.*key/) {
	    print "\n";
	    print "$@";
	    print "-" x 78, "\n";
	    print " " x 8, "DB Error: Skipping insertion into database.\n";
	    print " " x 10, "Data with identical primary key already exists.\n";
	    print " " x 10, "Please tell the developer(s) about this.\n";
	    print $sql, "\n";
	    print Dumper(\@args);
	    print "-" x 78, "\n";
            cluck($@);
    }
    elsif ($@ =~ /a foreign key constraint fails/) {
        croak("Execute of SQL failed because you tried to insert data into ".
              "a database with foreign keys inserted. You have to initialize ".
              "the database without foreign keys, and insert them later.");
    } elsif ($@) {
        my $sepp = "-" x 78;
        confess($sepp."\n".
                "ERROR: Prepare and execute failed.\n".
                $sepp."\n".
                "This represents a new and undiscovered error in ".
                "the system.\n".
                "SQL: '$sql'.\n".
                "ARGS: ".Dumper(\@args)."\n".
                "MESSAGE: |$@|\n".
                $sepp."\n");
    }

    return undef;
}

=item quote ($string)

=over

=item $string - The string to quote.

=back

External wrapper for the database-object's quote method.

=cut
sub quote {
    my ($self, $string) = @_;
    return $self->{dbh}->quote($string) if $self->{dbh};
}

=item prepare ($sql)

=over

=item $sql - The sql sentence to prepare.

=back

External wrapper for the database-object's prepare method.

=cut
sub prepare {
    my ($self, $sql) = @_;
    return $self->{dbh}->prepare($sql) if $self->{dbh};
}

#
# What follows are a lot of internal subroutines used by check_db_context:
#
sub _db_exists {
    my ($self, $db_name) = @_;

    my $sth = $self->prepare_and_execute(
        "SELECT schema_name FROM information_schema.schemata " .
        "WHERE schema_name = '$db_name'"
    );
    return $sth->rows;		# != 0 if the database exists.
}

# 'pristine'.  Etymology: Latin pristinus; akin to Latin prior 1:
# belonging to the earliest period or state: ORIGINAL <the
# hypothetical pristine lunar atmosphere>.  2 a: not spoiled,
# corrupted, or polluted (as by civilization): PURE <a pristine
# forest>.  b: fresh and clean as or as if new <pristine hard-backs in
# uniform editions to fill our built-in bookcases -- Michiko Kakutani>
#
sub _db_is_pristine {
    my ($self, $db_name) = @_;
    my $exceptions = $pristine_db_table_len_exceptions;

    my $sth = $self->prepare_and_execute("SELECT table_name "
					 . "FROM information_schema.tables "
					 . "WHERE table_schema = '$db_name'");
    # Checking for the right number of tables in the database.
    if ($sth->rows != $pristine_db_tables) {
	return 0;
    }

    # Check every table's length to see whether the database is in an
    # unused condition.
    #
    # Note: We cannot use the information from
    # information_schema.tables.table_rows because (at least) for
    # InnoDB-driven MySQL databases, that field is only an
    # approximation.
 TABLE:
    while (my $row = $sth->fetchrow_hashref()) {
	my $table_name = $row->{table_name};
	my $full_table_name = "$db_name.$table_name";

	# Get the length of the table.
	my $table_sth = $self->prepare_and_execute("SELECT count(*) "
						   . "FROM $full_table_name");
	if (!defined $table_sth) {
	    # If not even this step succeeds, we do definitively not
	    # have a pristine database.
	    return 0;
	}

	my $table_length = int($table_sth->fetchrow_array());

	# For all but a few exceptions, the table length should be 0.
	next TABLE if ((exists $exceptions->{$table_name})
			   && ($table_length == $exceptions->{$table_name}));

	# The default table length is 0.
	next TABLE if ($table_length == 0);

	# If we get this far, this table has been touched, and hence
	# the database is NOT pristine.
	return 0;
    }

    # Conclude that the database is (probably) pristine.
    return 1;
}

sub _find_createdb {
    my $current_path = $FindBin::Bin; # The (dir)path of this script.
    my $path = Env::Path->PATH;	      # The environment variable PATH.

    # Prefer to use CURRENT_PATH, iff the createdb script can be found
    # there.
    $path->Prepend($current_path);

    if (! exists $config->{CREATEDB_NAME}) {
	croak "Can't find the CREATEDB_NAME item in the current "
	    . "configuration.";
    }
    my @createdb = $path->Whence($config->{CREATEDB_NAME});
    return $createdb[0];
}

sub _recreate_db {
    my ($self, $create_only) = @_;

    my $name = $self->{DB_NAME};
    my $user = $self->{DB_USER};
    my $pass = $self->{DB_PASS};

    # Unless the user wants verbose output, ignore the output from the
    # child processes.
    my $redirection = $config->{VERB} ? ''
	                              : '> /dev/null 2>&1';

    # The switch which tells if we should add post-numbers or not.
    my $postnr = $config->{ADD_POSTNR} ? '--add-postnr'
                                       : '';

    # Find the most probable instance of the arkn4_createdb executable.
    my $createdb = _find_createdb();

    my $library_path = join(':', @INC);

    # Exporting the library path.
    $ENV{PERLLIB} = $library_path;

    if ($create_only == 0) {
	print "Dropping old database '$name'... " if $config->{VERB};

	my $command = "$createdb --user=$user --name=$name --password=$pass --drop-db $redirection";
	WIFEXITED(system $command)
	    or croak "Couldn't run '$command' ($OS_ERROR).";

	print "done\n" if $config->{VERB};
    }

    print "Creating new database '$name'... " if $config->{VERB};

    my $command = "$createdb --user=$user --name=$name --password=$pass $postnr $redirection";
    WIFEXITED(system $command)
	or croak "Couldn't run '$command' ($OS_ERROR).";

    print "done\n" if $config->{VERB};
}

sub _abort_by_user {
    my ($self, $answer) = @_;

    print "Your answer was: '$answer'.\n"
	. "The program cannot continue under the current conditions.\n"
	. "Aborting.\n";

    exit 0;
}

sub _interactive_new_db_name {
    my ($self) = @_;

    my $requirement_prompt = <<'END_REQUIREMENT';
Please enter a legal name (a letter followed by any combination of
letters, digits, and underscores):
END_REQUIREMENT

    my $new_name = prompt("Please enter a new database name: ",
			  -require => {$requirement_prompt => $LEGAL_DB_NAME});

    $config->{DB_NAME} = $new_name;
    print "Will now use database '$config->{DB_NAME}'.\n";

    $self->{DB_NAME} = $config->{DB_NAME};

    print "Checking whether the database is ready ...\n" if ($config->{VERB});
    $self->check_db_context();
}


sub _interactive_recreate_db {
    my ($self, $args) = @_;

    my $reason = $args->{reason} || "is not ready";

    my $create_choice = <<'END_CREATE_CHOICE';
  Press 'c' to automatically create the database.
END_CREATE_CHOICE

    my $recreate_choice = <<'END_RECREATE_CHOICE';
  Press 'r' to automatically recreate the database.  This is done by
  dropping the old database and creating a new one with the same name.
END_RECREATE_CHOICE

    my $context_dependent_choice = $args->{create_only} ? $create_choice
	                                                : $recreate_choice;


    my $prompt = <<"END_NOTICE";

You specified the database '$self->{DB_NAME}' for this operation, but
$reason.  This program can try to handle the situation for you.

$context_dependent_choice
  Press 's' to supply a new database name.

  Press 'a' to abort this operation and handle the problem manually.

The default choice, chosen by pressing <enter>, is:

END_NOTICE
    my $answer;

    if ($args->{create_only}) {
	if ($config->{FORCE_DB_NAME}) {
	    $answer = 'c';	# Create database.
	} else {
	    $answer = prompt($prompt, -onechar,
			     -default=>'c', -echo => q{},
			     -require => {'Must be a, c, or s: '
					  => qr/[acs]/xms});
	}
    } else {
	if ($config->{FORCE_DB_NAME}) {
	    $answer = 'r';	# Recreate database.
	} else {
	    $answer = prompt($prompt, -onechar,
			     -default=>'s', -echo => q{},
			     -require => {'Must be a, r, or s: '
					  => qr/[ars]/xms});
	}
    }
    # User chose to abort the program.
    $self->_abort_by_user($answer)            if ($answer eq q{a});

    # If we haven't exited above, we call _recreate_db(), letting the
    # $args->{create_only} value determine what will be done.
    $self->_recreate_db($args->{create_only}) if ($answer =~ qr/[cr]/xms);

    # Ask the user for a new database name and, if necessary, create it.
    $self->_interactive_new_db_name()         if ($answer eq q{s});
}


=item check_db_context ()

=cut

sub check_db_context {
    my ($self, $args) = @_;

    # Disconnect, just to be sure.
    $self->db_disconnect();

    # Connect to the database that's used for book-keeping.
    $self->db_connect($self->{DB_HOST}, 'mysql',
		      $self->{DB_USER}, $self->{DB_PASS});

    print " +-- Does the database '$self->{DB_NAME}' "
	. "exist? ... " if $config->{VERB};

    if ($self->_db_exists($self->{DB_NAME})) {
	print "yes\n" if $config->{VERB};

	print "   +-- Is the database empty and OK?   ... " if $config->{VERB};
	if ($self->_db_is_pristine($self->{DB_NAME})) {
	    print "yes\n" if $config->{VERB};
	} else {
	    print "no\n" if $config->{VERB};
	    # Suggest to the user that we drop the old database and
	    # create a new one.
	    $self->_interactive_recreate_db({reason => "it is not empty",
					     create_only => 0,});
	}
    } else {
	print "no\n" if $config->{VERB};
	# Suggest to the user that we create a new database.
	$self->_interactive_recreate_db({reason => "it does not exist",
					 create_only => 1,});
    }

    # print "OK, ready to run!\n";
}

=back

=head1 SEE ALSO

=over

=item Noark::Analyzer

=item Noark::Config

=item Noark::Parser

=item Noark::Report

=back

=head1 AUTHORS

Thomas Malt <tm@linpro.no>,
Martin Thorsen Ranang <mtr@linpro.no>

=head1 COPYRIGHT

=over

=item Copyright (C) 2003-2007  Riksarkivaren

=back

=head1 LICENSE

Whatever the licence Riksarkivaren choose to put this under.

=cut

1;
__END__
# 

