#!/usr/bin/perl
#-----------------------------------------------------------------------------
# $Id: Error.pm 749 2011-06-22 08:35:23Z tm $
#-----------------------------------------------------------------------------
# Noark::Error - Standard error class.
# 
# Author: Thomas Malt <tm@linpro.no>
#
# TODO All errors should derive from this object.
# TODO Move existing objects to this syntax
# TODO Derive classes for each error from this class.
#
#-----------------------------------------------------------------------------
package Noark::Error;

=head1 NAME

Noark::Error - Standard error class for a whole bunch of errors. 

=head1 SYNOPSIS

  my $err = Noark::Error->new(name => "foobar error", 
                              text => "something is foobar");

=head1 DESCRIPTION

Standard error message. Contains simple variable substitution logic.

Text can contain variables with the syntax '##<variable>##'. If text
contains ##VAR##, then this will be substituted with the provided variable:

  $err->add_variable(VAR => 'Text to be inserted');

=cut

use strict;
use warnings;

use Carp;
use Carp::Assert;
use Data::Dumper;

sub new {
    my ($proto, %args) = @_;
    
    my $class = ref($proto) || $proto;
    my $self  = {};
    
    $self->{name}        = $args{name}        || "";
    $self->{text}        = $args{text}        || "";
    $self->{keys}        = $args{keys}        || [];
    $self->{file}        = $args{file}        || "";
    $self->{__vars}        = $args{variables}   || {};
    $self->{__list}        = $args{listing}     || [];
    $self->{__count} = 0;
    $self->{__keys}  = {};
    $self->{__key_values} = [];
    bless($self, $class);

    $self->set_list_labels($args{__list_labels} || []);


    return $self;
}


sub add_line {
    my ($self, $args) = @_;

    my $j = $self->{__keys};
    for my $k (@{$self->{__key_values}}) {
        if (not defined $args->{$k}) { 
            croak("You must supply a value for '$k' in arguments");
        }

        $j->{$args->{$k}} = {};
        $j = $j->{$args->{$k}};
    }

    return 1;
}

=item add_variable (key => value)

Adds one or more variables to the variables storage. Used to generate error
message.

=cut
sub add_variable {
    my ($self, %vars) = @_;
    
    for (keys %vars) {
        $self->{__vars}->{$_} = $vars{$_};
    }

    return 1;
}

sub add_listing {
    my $self = shift;
    my @list = @_;

    # Verify that @list has content
    croak("You must supply a list of strings to add_listing") if(!scalar @list);

    my $lab_cnt = scalar @{$self->{__list_labels}};
    my $lst_cnt = scalar @list;
    
    push @{$self->{__list}}, \@list;

    return 1 if (!$lab_cnt);


    if ($lst_cnt != $lab_cnt) {
        pop @{$self->{__list}}; # clean up in case we catch this in an eval
        croak( "Number of fields in listing ($lst_cnt) must be of same "
             . "length as number of labels ($lab_cnt).");
    }
    
    return 1;
}

sub get_listing {
    my $self = shift;

    return $self->{__list};
}

=item get_text

Returns the formatted text from the error.

=cut
sub get_text {
    my ($self) = @_;

    my $text = $self->{text};
    chomp $text;

    for my $key (keys %{$self->{__vars}}) {
        my $value = $self->{__vars}->{$key};
        $text =~ s/##$key##/$value/g;
    } 

    my $keys = "";
    for (@{$self->{keys}}) {
        $keys .= $_.", ";
    } 
    $keys =~ s/,\s*$/./;
    $text .= $keys;

    return $text; 
} 



=item get_name

Returns name of error

=cut
sub get_name {
    my ($self) = @_;
    
    return $self->{name};
}


=item set_file

Set the filename an error occurs in.

=cut
sub set_file {
    my ($self, $file) = @_;

    assert($file);

    $self->{file} = $file;

    return 1;
}


=item get_file

Returns filename error occurs in.

=cut
sub get_file {
    my ($self) = @_;

    return $self->{file};
}


=item increment_counter

Adds one to the counter for this type of error.

=cut
sub increment_counter {
    my $self = shift;
    
    $self->{__count} += 1;

    return $self->{__count};
}


sub get_error_count {
    my $self = shift;

    return $self->{__count};
}

=item get_list_labels

Returns a list of labels for displaying the error list. as an ARRAYREF.

=cut
sub get_list_labels {
    my $self = shift;

    return $self->{__list_labels};
}


=item set_list_labels(label, label, label)

Sets a list of strings to be the labels for error listing.

=cut
sub set_list_labels {
    my $self = shift;

    if (ref $_[0] eq 'ARRAY') {
        $self->{__list_labels} = shift;
    } 
    else {
        $self->{__list_labels} = \@_;
    }

    return 1 if (!scalar @{$self->{__list}});

    my $lst_cnt = scalar @{$self->{__list}->[0]};
    my $lab_cnt = scalar @{$self->{__list_labels}};
    if ($lst_cnt != $lab_cnt) {
        croak("Number of labels must be equal to number of fields in list");
    }

    return 1;
}


=item set_list_label_lengths

Fields in error listing might be of different lengths for test output.
This gives us a chance of setting them.

=cut
sub set_list_label_lengths {
    my $self = shift;

    if (scalar @{$self->{__list_labels}} != scalar @_) {
        croak("Length list must be equal to number of labels");
    }

    $self->{__list_labels_length} = \@_;

    return 1;
}


=item get_list_label_lengths

Fields in an error listing might be of different lengths for text output. 
This gives us a chance of getting them.

Returns ARRAYREF.

=cut
sub get_list_label_lengths {
    my $self = shift;

    if (!defined $self->{__list_labels_length} ||
        !@{$self->{__list_labels_length}}) {
        print Dumper $self;
        confess("List label lenghts must be set before they can be fetched.");
    }

    return $self->{__list_labels_length};
}


# Replacement for FORMATS. Simple formatting of text strings.
# Does not return a newline at the end.
sub __format_text {
    my ($self, $text, $indent) = @_;

    $indent ||= 0;
    my $pre  = " " x $indent;

    my @words = split / /, $text;
    my $done  = "";
    my $line  = "";

    while (my $word = shift @words) {
        my $return = undef;
        $line .= $pre if length($line) == 0;
        if ($word =~ /(\S*)(\n+)(\S*)/) {
            # word with newline in it
            $word = $1 || "";
            $return = $2;
            unshift(@words, $3) if defined $3;
            $word =~ s/\s*//g;
        }
        if (length($line.' '.$word) > 78) {
            $done .= $line."\n";
            $line = '';
            redo;
        }

        $line .= $word." ";
        if (defined $return) {
            $done .= $line.$return;
            $line = '';
        }
    }

    $done .= $line;

    return $done;
}


1;

__END__
