#!/usr/bin/perl
#-----------------------------------------------------------------------
# $Id: Log.pm 493 2007-03-01 10:19:44Z mtr $
#-----------------------------------------------------------------------
# Noark::Log - logging module for Noark
#
# Author: Knut Haugen <knuthaug@linpro.no>
#         Thomas Malt <tm@linpro.no>
#         Peder Stray <peder@linpro.no>
#
# Copyright (C) 2002-2003  Linpro AS
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
# USA.
#-----------------------------------------------------------------------

=head1 NAME

Noark::Log - logging module for Noark

=head1 SYNOPSIS

use Noark::Log qw(logname);

where <logname> is the name of the script to log for, or any other
name you wish to have in your logfile for this script.

=cut

=head1 DESCRIPTION

Module for initiating log in a uniform way for all scripts. Will open
a syslog connection.

=cut

package Noark::Log;


use Unix::Syslog qw(:subs :macros);
use Exporter;
use CGI qw(:standard);
use vars qw(@ISA @EXPORT $VERSION %levels %facilities);

use strict;

@ISA       = qw(Exporter);
@EXPORT    = qw(logg close_log);

($VERSION) = '$Revision: 493 $' =~ /: (\S+)/;

%levels = (
           0   => 'EMERG', 
           1   => 'ALERT', 
           2   => 'CRIT',
           3   => 'ERROR', 
           4   => 'WARNING', 
           5   => 'NOTICE', 
           6   => 'INFO', 
           7   => 'DEBUG'
           );

my $opened = 0;

my %conf = (
            pid => 1, 
            name => '', 
            facility => LOG_LOCAL0,
            debug => 1,
            );


=pod 

=head1 Functions

=over 4

=item import ($class, $name)

Local override of import () from module Exporter. This exports the
Unix::Syslog macros up to the main script and also the methods we
need. Called automatically. This method opens up a syslog connection.

=cut
sub import {

    my($class,$name) = @_;

    $class->export_to_level (1,undef);
    Unix::Syslog->export_to_level(1, undef, ':macros');
    
    $name = $conf{name} unless $name;
    
    if ($conf{pid} && !$opened) {
        openlog ($name, LOG_PID, $conf{facility});
        $opened = 1;
    }elsif (!$opened) {
        openlog ($name, 0, $conf{facility});
        $opened = 1;
    }
  
}

=pod

=item logg ($pri, $msg, @args)

The method to use for all logging. Can handle sprintf strings and
arguments (in @args)

Arguments:

=over 8

=item  - $pri - the syslog priority to log for ( see man syslog)

=item  - $msg - the message to print. This may be a sprintf format string

=item  - @args (optional - the arguments to print if $msg is sprintf format.

=back

Returns: undef

=cut

sub logg {
    my($pri,$msg,@args) = @_;
    my ($package, $filename, $line) = caller;
    my ($subroutine) = (caller (1))[3];

    if ($conf{debug}) {
        $subroutine ||= "main";
        $subroutine =~ s/.*:://;
        $msg = "${package}::${subroutine}($line): $msg";    
    }else {
        $msg = "$package: $msg"; 
    }

    #if (-t STDOUT) {
    #    warn($levels{$pri} . " [" .remote_host .  "] " . sprintf($msg,@args)."\n");
    #}

    syslog ($pri, $levels{$pri} . " [" . remote_host . "] " . $msg, @args);
    
}

=pod

=item close_log ()

Close the connection to the syslog daemon.
No params, returns the return from the closelog call, which is undef.

=cut

sub close_log {
    return Unix::Syslog::closelog();
}


=item DESTROY ()

Cleanup function for closing the connection to the syslog daemon. No
params, returns the return from the closelog call, which is undef.

=cut

sub DESTROY {
    return close_log();
}


=back 

=head1 SEE ALSO

Unix::Syslog

=head1 AUTHORS

  Thomas Malt <tm@linpro.no>
  Knut Haugen <knuthaug@linpro.no>
  Peder Stray <peder@linpro.no>

=head1 COPYRIGHT

Copyright (C) 2002-2003 Linpro AS.

=head1 LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

=cut

1;

__END__

#-----------------------------------------------------------------------
# $Log$
# Revision 1.1  2005/09/29 13:55:14  tm
# Added all files to repository
#
# Revision 1.1  2004/08/12 14:45:27  elark
# Fixing where its broke.
#
# Revision 1.2  2003/10/15 16:41:17  tm
# Cleaned up stuff and made a default analyzer handerl.
#
# Revision 1.1  2003/10/10 11:12:45  tm
# Added Helper module for various misc functions. Various updates WIP.
#
# Revision 1.3  2002/12/06 10:26:50  knuthaug
# doc fixes
#
# Revision 1.2  2002/08/20 13:46:58  knuthaug
# cleanups on import
#
# Revision 1.1  2002/08/20 12:52:35  knuthaug
# import of logging module from limacute/VG projects
#
