#!/usr/bin/perl
#-----------------------------------------------------------------------------
# $Id: Parser.pm 803 2011-07-14 08:45:21Z tm $
#-----------------------------------------------------------------------------
# Noark::Parser - Main parser object for parsing xml data.
#
# Author: Thomas Malt           <tm@linpro.no>
#         Kjell Magne ierud    <kjellm@linpro.no>
#
# Copyright (c) 2003-2011  Riksarkivaren
#
#-----------------------------------------------------------------------------

=head1 NAME

Noark::Parser - Parser for deposited XML data.

=head1 SYNOPSIS

my $p = Noark::Parser->new(dir => "path/to/deposit")

=head1 DESCRIPTION

This parser uses the XML::DOM module for reading and parsing the
deposited Noark data. Currently it uses the NOARKIH.XML file to get
information about tables and fields included in the deposit.

NOARKIH.XML serves as a starting point for a framework for integrity
checking and analysis of the deposit. This feature can be used as an
error tracking system, and as a feedback mechanism to the various
depositors.

I had some doubt whether to inherit from XML::DOM::Parser or wrap
it. In the end I decided on wrapping it since the addidional
functionality we need is so minimal.

=head1 DEFINITIONS

foo

=cut

package Noark::Parser;

use strict;
use warnings;

use lib qw(..);

use Carp qw(carp cluck confess croak);
use Carp::Assert;
use Data::Dumper;

use File::Basename;
use File::Find::Rule;
use IO::File;
use List::Util qw( first );
use Time::HiRes qw( time );

use XML::Twig;

use Noark::Config qw( $dtds );
use Noark::Db;
use Noark::Schema;
use Noark::Schema::Map;
use Noark::Parser::ErrorList;
use Noark::Parser::Tabellinfo;
use Noark::Parser::TabellinfoList;
use Noark::Parser::Filemap;
use Noark::Progress;

our ($VERSION) = '$Revision: 803 $' =~ /: (\S+)/;

# Different XML-Elements needs to be handled differently, and this
# HASHREF, defines handlers for handling the different ellements.
my %handlers = ('TI.TABELL'         => \&handle_ti_tabell,
                'ATTRIBUTTER'       => \&handle_attributter,
                'TI.ANTFILER'       => \&handle_ti_fildeler,
                'FIL'               => \&handle_fil
               );


=head1 CONSTRUCTOR

=over

=item new (%args)

Constructor. Creates a new Parser object.

Arguments:
  dir     => PATH - Where to find the deposit.
  usedb   => BOOL - Whether to use a Database or not. Default '0'.
  verbose => INT - set verbosity level
  debug   => INT - whether to print meaningless output to STDOUT
  tmpdir  => STRING - Which temp. dir to use for various temporary needs.
The argument is mandatory.

=back

=cut
sub new {
    my ($proto, %args) = @_;
    
    my $class  = ref($proto) || $proto;
    my $self   = {};
    my $tmpdir = '/tmp/arkn4.'.$ENV{USER}.'.'.$$;

    # Variables
    $self->{VERB}       = $args{verbose}         || 0;
    $self->{DEBUG}      = $args{debug}           || 0;
    $self->{DIR}        = $args{dir};
    $self->{XMLDIR}     = $args{xmldir};
    $self->{DOCDIR}     = $args{docdir};
    $self->{REPORTDIR}  = $args{reportdir};    
    $self->{DO_FKCHECK} = $args{'do_fkcheck'};   
    $self->{DEL_FK_ERR} = $args{del_fk_err}      || 0;
    $self->{TMPDIR}     = $args{tmpdir}          || $tmpdir; 

    $self->{SQLDIR}     = dirname($INC{'Noark/Parser.pm'}).'/SQL';

    if (!-d $self->{DIR}) {
        croak("You need to supply a valid directory for the deposit in 'dir'");
    }

    if (!-d $self->{SQLDIR}) {
        croak("Unable to find a valid SQL-directory in installation:\n    ".
              $self->{SQLDIR});
    } 

    # Objects
    $self->{XP}      = undef; # Reference to the parser object.
    $self->{IH}      = undef; # Reference to the XML::Document.
    $self->{JO}      = undef;
    $self->{DB}      = undef;

    $self->{VERSION} = $VERSION;

    bless($self, $class);

    # Containers;
    $self->{__dberrors}        = $self->get_db_errorlist();
    $self->{__pk_normalized} = {}; # Used by verify_primary_key_integrity

    print "\n    Parsing tabellinfo list ... " if $self->{VERB} > 2;
    $self->{INFO}           = $self->get_tabellinfo_list();
    print "Done.\n" if $self->{VERB} > 2;

    print "    Parsing the filemap ... " if $self->{VERB} > 2;
    $self->{FILEMAP}        = $self->get_filemap();
    print "Done.\n" if $self->{VERB} > 2;

    print "    Parsing the schema ... " if $self->{VERB} > 2;
    $self->{SCHEMA}         = $self->get_noark_schema();
    print "Done.\n" if $self->{VERB} > 2;

    $self->{DTDMAP}         = $dtds;
    $self->{DBDONE}         = {}; # Check if we have inserted into db.
    $self->{REAL_TABLES}    = {};

    if (defined($args{noarkih}) && -r $args{noarkih}) {
        $self->{FILEMAP}->{FILEMAP}->{'NOARKIH.XML'} = $args{noarkih};
    }

    return $self;
}


=head1 SUBROUTINES

=over

=cut

sub _handle_duplicate_pk {
    my ($self, $table, $pk, $file, $line) = @_;

    if ($self->{VERB} > 2) {
        print " " x 8;
        print "Found duplicate value ($pk) for primary key\n";
    }

    my $db_errors = $self->get_db_errorlist();
    $db_errors->add_error($table, 'Noark::Error::DuplicateKey', {
        table => $table,
        pk    => $pk,
        file  => $file,
        line  => $line,
    });

    return 1;
}


sub _handle_potential_duplicate_pk {
    my ($self, $table, $pk, $norm, $file, $line) = @_;
    
    my $db_errors = $self->get_db_errorlist();
    $db_errors->add_error($table, 'Noark::Error::PotentialDuplicateKey', {
        pk => $pk,
        norm => $norm,
        file => $file,
        line => $line,
    });

    return 1;
}

#-----------------------------------------------------------------------------
# Feilmelding som skrives ut nr vi ikke finner en gyldig primrnkkel.
#
sub _handle_not_valid_primary_key {
    my ($self, $table, $what, $file_name, $line_number) = @_;

    if ($self->{VERB} > 2) {
        print " " x 8, "Found invalid key ($what) in table.";
        print "Line: $line_number\n";
    }
 
    my $db_errors = $self->get_db_errorlist();
    $db_errors->add_error($table, 'Noark::Error::PrimaryKey', {
        file => $file_name,
        line => $line_number,
        pk   => $what,
    });

    return 1;
}

sub _handle_null_error {
    my ($self, $table, $field, $file_name, $line_number) = @_;

    if ($self->{VERB} > 2) {
        print " " x 8;
        print "Mangler verdi i et felt ($field) som er NOT NULL.\n";
    } 

    my $db_errors = $self->get_db_errorlist();
    $db_errors->add_error($table, 'Noark::Error::Null', {
        file => $file_name,
        line => $line_number,
        field => $field,
    });
    
    return 1;
}

sub _handle_data_valid_error {
    my ($self, $table, $field, $row, $data_type, $file, $line) = @_;

    if ($self->{VERB} > 2) {
         print " " x 8, qq{'$field' var };
         print qq{$row nr datatypen var '$data_type'.\n};
    } 

    my $db_errors = $self->get_db_errorlist();
    $db_errors->add_error($table, 'Noark::Error::ValidData', {
        field    => $field,
        value    => $row,
        datatype => $data_type,
        file     => $file,
        line     => $line,        
    });

    return 1;
}

sub _handle_duplicate_unique {
    my ( $self, $table, $field, $value, $type, $file, $line_2, $line_1 ) = @_;

    if ($self->{VERB} > 2) {
        print ' ' x 8, qq{'$field' er UNIQUE, men verdien '$value' forekom };
        print qq{flere ganger.\n};
    }
    
    my $db_errors = $self->get_db_errorlist();
    $db_errors->add_error($table, 'Noark::Error::DuplicateUnique', {
        field    => $field,
        value    => $value,
        datatype => $type,
        file     => $file,
        lines    => [$line_1, $line_2],
    });

    return 1;
}

=item get_progress

Returns a Noark::Progress object for printing nice status messages.

=cut
sub get_progress{
    my $self = shift;

    if (defined $self->{PROGRESS} && 
        ref $self->{PROGRESS} eq 'Noark::Progress') {
        return $self->{PROGRESS};
    }
    
    $self->{PROGRESS} = Noark::Progress->new(verbosity => $self->{VERB});
    
    return $self->{PROGRESS};
}

sub get_db_errorlist {
    my $self = shift;

    if (ref $self->{__dberrors} eq 'Noark::Parser::ErrorList') {
        return $self->{__dberrors};
    }

    $self->{__dberrors} = Noark::Parser::ErrorList->new();
    return $self->{__dberrors};
}

sub get_filemap {
    my $self = shift;
    
    if (ref($self->{FILEMAP}) eq 'Noark::Parser::Filemap') {
        return $self->{FILEMAP};
    }

    $self->{FILEMAP} = Noark::Parser::Filemap->new({
        dir       => $self->{DIR},
        xmldir    => $self->{XMLDIR},
        docdir    => $self->{DOCDIR},
        reportdir => $self->{REPORTDIR},
        list      => $self->get_tabellinfo_list() 
    });

    return $self->{FILEMAP};
}


=item get_noark_schema ()

This subroutine is self documenting.

Creates a Noark::Schema object and returns a reference to it.

=cut
sub get_noark_schema {
    my $self = shift;

    return $self->{SCHEMA} if ref $self->{SCHEMA} eq 'Noark::Schema';

	my $eksportinfo = $self->get_eksportinfo();
	$self->{SCHEMA} = Noark::Schema->new(
	    debug       => $self->{DEBUG},
            verbosity   => $self->{VERB},
            eksportinfo => $eksportinfo
	);
	
	return $self->{SCHEMA};
}

sub _find_noarkih {
    my ($self) = @_;
    my $target_filename = 'NOARKIH.XML';
    my $dir = $self->{DIR};

    # Find the first occurence of $target_filename in $dir or any of
    # its sub-directories...
    my $path = first { 1 } File::Find::Rule->file
	                                   ->name(qr/$target_filename/i)
		                           ->in( ($dir) );

    return undef if !defined $path;

    # Return the found path, but add './' if it is not an absolute
    # filename...
    return ($path =~ qr/\A\//xms) ? $path
	                          : './'.$path;

}

=item set_noarkih_file ($file)

 $file - PATH to NOARKIH.XML

If NOARKIH.XML was not set on construction you can set it with this
function.

=cut
sub set_noarkih_file {
    my ($self, $file) = @_;

    if (-r $file) {
        $self->{FILES}->{'NOARKIH.XML'} = $file;
    } else {
        cluck("No such file: $file");
        return 0;
    }

    return 1;
}


=item get_db_handle ()

Returns a db handle for the object.

=cut
sub get_db_handle {
    my ($self) = @_;

    return $self->{DB} if (defined $self->{DB});

    $self->{DB} = Noark::Db->new();
    return $self->{DB};
}

=item get_eksportinfo ()

Extracts EKSPORTINFO from NOARKIH.XML and returns an HASHREF
containing the data for easy use other places.

  return HASHREF

=cut
sub get_eksportinfo {
    my ($self) = @_;

    # my $ih = $self->get_noarkih_dom();
    my $file = $self->_find_noarkih();
    return undef if not -r $file;

    my @exportinfo = undef;
    my $ih = XML::Twig->new(
        twig_handlers => { 
            EKSPORTINFO => sub { @exportinfo = $_->children; },
        }
    );
    $ih->parsefile($file);
    
    my %EI = map { $_->name => $_->text } @exportinfo;
    $ih->purge();
    return \%EI;
}


=item get_tabellinfo_list

Return a Noark::Parser::TabellinfoList object

=cut
sub get_tabellinfo_list {
    my $self = shift;

    if (ref($self->{INFO}) eq 'Noark::Parser::TabellinfoList') {
        return $self->{INFO};
    }

    $self->{INFO} = Noark::Parser::TabellinfoList->new(
        file => $self->_find_noarkih()
    );

    return $self->{INFO};
}


=item get_xmlmeta_info ($file)

This functions snarfs various information from the XML file without
going through an XML parser. It is meant to be a last resort solution
to find out what the people in the other end of the archive export has
been thinking.

  param  STRING  $file -- The key in the FILEMAP.
  return HASHREF $info -- Information about what we found.

=cut
sub get_xmlmeta_info {
    my ($self, $file) = @_;

    my $filemap = $self->get_filemap();

    mkdir $self->{TMPDIR} if ! -d $self->{TMPDIR};

    my $F = IO::File->new($filemap->get_file_by_name($file), 'r')
        || confess("Could not open file: $!");

    # TODO Lag en bedre fiks for hndtering av temp-filer.
    $file = $file.'.'.$$.'.DTDFIX.XML';

    my $OUT = IO::File->new($self->{TMPDIR}.'/'.$file, 'w') 
        || confess("Couldn't open file: $!");

    my $line    = 0;
    my $dtd     = undef;
    my $table   = undef;
    my $dtdline = 0;
    for (<$F>) {
        next if /^\s*$/;
        $line += 1;
        if (/^\<\!DOCTYPE/) {
            $dtdline = $line;
            chomp;

            my ($doc, $sys, $ref) = undef;
            ($doc, $table, $sys, $ref) = split(/\s/);
            ($dtd) = $ref =~ /[\"\']?([\w\.]*)[\"\']?\>/;

            my $real  = $self->{DTDMAP}->{$table};
            if (not defined $real) {
                $F->close;
                $OUT->close;
                return undef;
            }

            print $OUT "$doc $table SYSTEM \"$real\">\n";
        } else {
            print $OUT $_;
        }
    }
    $F->close;
    $OUT->close;

    return undef unless defined $table;

    my %info = (TABLE   => $table,
                DTDREF  => $dtd,
                LINE    => $line,
                DTDLINE => $dtdline,
                FILE    => $self->{TMPDIR}.'/'.$file);

    return \%info;
}


=item get_data_and_insert ($name)

Extracts the data given in $root and inserts it into the database.
Returns undef if something goes wrong. Returns TRUE if things go
right.

Arguments:
  info - HASHREF, Tableinfo gotten from noarkih.xml.
  element - XML::Twig::Elt
  file - Name of file we are currently parsing.

Generates error messages about the data and inserts them in the
DBERROR hash for later use.

TODO: Implement a more robust way of doing error-handling.

=cut
sub get_data_and_insert {
    my ($self, $element, $arg) = @_;

    my $info              = $arg->{info};
    my $table_name        = $info->get_table_name();
    my $file              = $arg->{'file'};  
    my $line              = $element->att('#line');
    my $extra_attributes  = $arg->{extra};
    my $used_attributes   = $arg->{used};
    my $noark_attributes  = $info->get_attributes;
    my $pk_seen           = $arg->{pk_seen};
    my $unique_seen       = $arg->{unique_seen};
    my %noark_attr        = map { $_ => 1 } @{$info->get_attributes};
    my %row               = map { $_->name => $_->text } $element->children;
    my $progress          = $self->get_progress();
    my $schema            = $self->get_noark_schema();
    my $table             = $schema->get_table($table_name);

    $progress->increment_done_by_amount();
    $progress->print_status() if $self->{VERB} > 1;
    
    # check attributes not in deposit for primary keys that can be null
    # Check primary keys that can be null (composite keys) if they are 
    # null and insert dummy value if so
    for my $att (@{$schema->get_pk_names($table_name)}) {
        my $a = $table->get_attribute($att);
        if (ref($a) ne 'Noark::Schema::Attribute') {
            croak "Did not get a Schema::Attribute";
        }
        
        if ($a->can_be_null && (not defined $row{$att})) {
            $row{$att} = $a->get_default_value();
        }
    }

    for my $att (keys %row) {
        $used_attributes->{$att}   = 1;
        $extra_attributes->{$att} += 1 if ! $noark_attr{$att};        
    }

    my $washed = $self->wash_attribute_values(\%row);

    my $pk_res = $self->verify_primary_key_integrity(
        table   => $table_name,
        row     => $washed,
        file    => $file,
        line    => $line,
        pk_seen => $pk_seen,
    );
    return undef if ! defined $pk_res;

    my $verified = $self->verify_attributes(
        row    => $washed,
        noark  => $noark_attributes,
        table  => $table_name,
        info   => $info,
        file   => $file,
        line   => $line,
        unique => $unique_seen,
    );
    return 0 if not defined $verified;
    my $sql = $self->build_sql($table_name, $noark_attributes);
    my $db  = $self->get_db_handle();
    my $sth = $db->prepare_and_execute($sql, @{$verified});

    return 0 if ! defined $sth;
    return 1;
}

sub wash_attribute_values {
    my ($self, $row) = @_;

    if (ref $row ne 'HASH') {
        croak("HASHREF needs to be given as argument to wash_attribute_values");
    }

    for my $k (keys %$row) {
        # Some keys can be undef (null in the db) and if they are part of a
        # key they are inserted with their default value in the row. This can
        # be undef, so to avoid problems we skip here if that is the case.
        next if ! defined $row->{$k}; 

        # Removing whitespace from beginning and end of values.
        $row->{$k} =~ s/^\s+//;
        $row->{$k} =~ s/\s+$//;
        
        # Cludge to accept strings with multiple spaces. Multiple spaces may
        # have been inserted on purpose, and might be integrals to maintaining
        # database consistensy.
        # 
        # Our workaround is to regard returns and carriage returns as added to
        # the xml for formatting and not part of the deposit, and ordinary 
        # whitespace as part of the deposit and should be untouched.
        # Beware that this might not be fool proof and might introduce errors
        # if some deposits actually intended to use returns and carriage returns
        # as values for keys.
        $row->{$k} =~ s/\r//g;
        $row->{$k} =~ s/\n\s+/ /g;
        $row->{$k} =~ s/\n//g;
    }

    return $row;
}

sub verify_primary_key_integrity {
    my ($self, %args) = @_;

    my $table   = $args{table};
    my $row     = $args{row};
    my $file    = $args{file};
    my $line    = $args{line};
    my $pk_seen = $args{pk_seen};

    # Handle primary keys (this includes handling of composite
    # primary keys as well).
    my ($pk, $what) = $self->make_primary_key_string($table, $row);

    if (! defined $pk) {
        $self->_handle_not_valid_primary_key($table, $what, $file, $line);
        return undef;
    }

    # NOTE: Primary key can be zero (0).
    # Empty pk ('') _should_ mean the key is auto-increment.
    if ((defined $pk) && ($pk ne '')) {
        $pk_seen->{$pk} += 1;
        if ($pk_seen->{$pk} > 1) {
            $self->_handle_duplicate_pk(
                $table, $pk, $file, $line
               );
            return undef;
        }

        my $norm = $pk;
        $norm =~ s/\s+//g; # Remove all whitespace from pk regardless.
        $self->{__pk_normalized}->{$table}->{$norm} += 1;
        if ($self->{__pk_normalized}->{$table}->{$norm} > 1) {
            $self->_handle_potential_duplicate_pk(
                $table, $pk, $norm, $file, $line
            );
            # Not critical so not returning.
        }
    }

    return 1;
}


=item make_primary_key_string ($table, $attrs)

Helper used by get_data_and_insert.  Concatenates all primary key
fields (normally one, but sometimes two or three) in a table into a
string for later use.

It also checks the primary key fields for nullness and returns with
any errors.

Returns a list with (undef, <attribute>) if the key is not valid.
Otherwise returns the string.

=cut
sub make_primary_key_string {
    my ($self, $table, $attrs) = @_;

    my $noark_schema = $self->get_noark_schema();
    my $primary_keys = $noark_schema->get_pk_names($table);

    my $primary_key = q||;
    for (@{$primary_keys}) {
        if ($noark_schema->is_attr_not_null($table, $_)         &&
            ! $noark_schema->is_attr_auto_increment($table, $_) &&
            ! defined($attrs->{$_})) {
            return (undef, $_);
        }
        $primary_key .= (defined($attrs->{$_})) ? $attrs->{$_} : "";
    }

    return $primary_key;
}

=item verify_or repair_foreign_key_integrity ()

Validates that all neccessary keys are present in the import.
Will delete rows of data that don't meet the requirements.

Author: Kjell Magne ierud, kjellm@linpro.no
        Thomas Malt, tm@linpro.no

=cut
sub verify_or_repair_foreign_key_integrity {
    my ($self, %arg) = @_;
    
    my $verify  = $arg{verify} || 0;
    my $repair  = $arg{repair} || 0;
    my $typestr = $verify ? 'Verifying' : 'Repairing';
 
    printf "%s foreign key constraints...\n", $typestr if ($self->{VERB});

    my $schema      = $self->get_noark_schema();
    my $infolist    = $self->get_tabellinfo_list();
    my $table_count = 0;
    my $table_total = $infolist->get_tabellinfo_count();
    my $inserted    = 0;

    for my $table (@{$schema->get_table_list()}) {
        my $name  = $table->get_noark_name();
        my $count = 0;
        
        next if not $infolist->get_tabellinfo_by_name($name);
        
        $table_count += 1;
        
        printf('  [%s/%s] %s %- 20s: ', sprintf("%02d", $table_count), 
               $table_total, $typestr, $name) if ($self->{VERB});

        for my $fk (@{$table->get_foreign_keys()}) {
            my $rows = $self->_get_rows_with_missing_reference($table, $fk);
            $count += scalar(@{$rows}) if $verify;

            for my $row (@{$rows}) {
                if ($verify) {
                    $self->_handle_missing_reference_error({
                        table        => $table, 
                        row          => $row, 
                        foreign_keys => $fk
                    });
                    next;
                }
                if ($repair) {
                    $count += $self->_fix_foreign_key_reference({
                        table           => $table,
                        attribute       => $fk->[0], 
                        ref_table       => $fk->[1],
                        attribute_value => $row->{$fk->[0]->get_name},
                        ref_attribute   => $fk->[2], 
                    });
                    next;
                }
            }
        }

        if ($self->{VERB}) {
            printf("Missing: % 6d keys\n", $count)  if $verify; 
            printf("Inserted: % 6d rows\n", $count) if $repair;
        } 
    }

    printf("%s foreign key constraints: [DONE]\n", $typestr) if $self->{VERB};
    return 1;
}

sub _handle_missing_reference_error {
    my ($self, $args) = @_;
    
    my $table          = $args->{table}->get_noark_name();
    my $row            = $args->{row};
    my $attr           = $args->{foreign_keys}->[0];
    my $ref_table      = $args->{foreign_keys}->[1];
    my $ref_attr       = $args->{foreign_keys}->[2];
    my $ref_attr_name  = $ref_attr->get_noark_name();
    my $ref_table_name = $ref_table->get_noark_name();
    my $attr_name      = $attr->get_noark_name();

    # By royal decree we skip all postnr-errors.
    return 0 if $ref_table_name eq 'POSTNR';

    my $db_errors = $self->get_db_errorlist();
    $db_errors->add_error($table, 'Noark::Error::ForeignKey', {
        field       => $attr_name,
        value       => $row->{$attr->get_name},
        requirement => 
            $attr->get_requirement." => ".$ref_attr->get_requirement, 
        ref_field   => $ref_attr_name,
        table       => $table,
        ref_table   => $ref_table_name,
    });

    return 1;
}



=item _fix_foreign_key_reference()

Repairs the missing foreign key reference by inserting a dummy row in the 
table missing a key.

Missing mandatory tables and attributes are already caught, and error
messages for that is created.
 
Calls itself recursively if dummy row also has foreign key dependencies.
Returns the number of rows inserted to make database valid.

=cut
sub _fix_foreign_key_reference {
    my ($self, $args) = @_;

    my $table         = $args->{table};
    my $attribute     = $args->{attribute};
    my $attr_value    = $args->{attribute_value}; 
    my $ref_table     = $args->{ref_table};
    my $ref_attr      = $args->{ref_attribute};
    my $ref_attr_name = $ref_attr->get_name();
    my $dbh           = $self->get_db_handle();

    if (! $attribute->is_exportable()) {
        croak(
            "Got called with an attribute which is not exportable. This " .
            "should never happen."
        );
    }
    
    # Escape if $attr_value is undef and can be undef
    if (!defined($attr_value)) {
        if ($attribute->can_be_null()) {
            return 0;
        }
        else {
            croak(
                "Got an undefined attribute value for attribute: " .
                $attribute->get_name()
            );
        }
    }

    # Don't insert if foreign key already exist.
    return 0 if $dbh->is_attribute_in_table({
        attribute => $ref_attr_name,
        table     => $ref_table->get_name(),
        value     => $attr_value,
    });
    my $inserted = 0;
    
    # Adding expection for postnummer, which should never be exported but is
    # heavily used.
    my $ref_attr_list = $ref_table->get_name eq 'postnr' 
                      ? $ref_table->get_attribute_list() 
                      : $ref_table->get_exportable_attribute_list();

    # Insert the referenced attribute into the list if it is not there.
    # This can happen if the attribute is not exportable by default, but
    # to fix the data integrity issue it is inserted. 
    #
    # Another way to fix this would be to change the database definition and
    # not include a foreign key reference for the attribute, but this would
    # make our database defintion less usable as a general purpose NORK4
    # implementation.   
    if (!grep { $_->get_name eq $ref_attr_name } @$ref_attr_list) { 
        push @$ref_attr_list, $ref_attr; 
    }

    my %attr_value = ();
    for (@$ref_attr_list) {
        if ($_->get_name() eq $ref_attr_name) {
            $attr_value{$_->get_noark_name()} = $attr_value;
            next;
        }
        next if ($_->can_be_null() && !defined $_->get_fk_reference());
        
        # If we got here this is not the key with meaning in the row and we
        # try to get an acceptable dummy value:
        my $value = $self->__get_dummy_value($_->get_data_type());

        while (( $_->is_primary_key() || $_->is_unique() ) &&
                   $dbh->is_attribute_in_table({
                       attribute => $_->get_name(),
                       table     => $ref_table->get_name(),
                       value     => $value
                   })
        ) {
            $value = $self->__get_dummy_value($_->get_data_type(), $value);
        }

        $attr_value{$_->get_noark_name()} = $value;
    } 

    my @keys   = keys %attr_value;
    my @values = @attr_value{@keys};
    
    if (scalar(@keys) != scalar(@values)) {
        # If these are not the same something is wrong
        croak(
            "Got different key and value counts (key/value mismatch) " .
            "before repairing foreign key reference."
        );
    }                              
                                    
    my $sql = $self->build_sql($ref_table->get_noark_name(), \@keys, 1);
    $dbh->prepare_and_execute($sql, @values);
    $inserted += 1;

    # If this reference also has references we must insert dummy rows for this
    # value also:
    for my $fkey (@{$ref_table->get_foreign_keys()}) {
        my ($f_attr, $f_ref_table, $f_ref_attr) = @{$fkey};
        next if ! $f_attr->is_exportable();
        my $f_value  = $attr_value{$f_attr->get_noark_name()};

        my %ref_args = ( table           => $ref_table,
                         attribute       => $f_attr,     
                         ref_table       => $f_ref_table,
                         attribute_value => $f_value,
                         ref_attribute   => $f_ref_attr,
                       );
        $inserted += $self->_fix_foreign_key_reference(\%ref_args);
    }

    return $inserted;    
}


sub __split_datatype_and_size {
    my ($self, $type) = @_;
    
    my $size = undef;
    if ($type =~ m/(\w+)\((\d+)\)/) {
        $type = $1;
        $size = $2;
    }

    return ($type, $size);
    
}

sub __get_dummy_value {
    my ($self, $datatype, $value) = @_;

    my ($type, $size) = $self->__split_datatype_and_size( $datatype );

    my %handler =  ( VARCHAR => \&__get_dummy_varchar,
                     BOOL    => sub { return 0; },
                     DATE    => \&__get_dummy_date,
                     INT     => \&__get_dummy_int,
                     TEXT    => sub { return 1; },
                     BLOB    => sub { return 1; },
                     TIME    => \&__get_dummy_time,
                   );

    return $handler{$type}($self, $size, $value);
}

sub __get_dummy_varchar {
    my ($self, $size, $value) = @_;
    croak("Size must be defined")   if !defined($size);
    croak("Size must be a number.") if (!($size =~ m/^\d+$/));

    return 'X' x $size if !defined($value); 

    # Ok: we got old value now deal with it:
    my $min  = ord('A');
    my $max  = ord('X');
    my @list = split('', $value);

    my @items = ();
  ITEM:
    while (my $item = pop @list) {
        if (ord($item) <= $min) {
            unshift @items, chr($max);
            next ITEM;
        }
        $item = chr(ord($item)-1);
        unshift @items, $item;
        last ITEM;
    }
    push @list, @items;
    
    return join('', @list);
}

sub __get_dummy_date {
    my $self = shift;
    my @ltime = localtime();

    my $year  = $ltime[5]+1900;
    my $month = sprintf("%02d", $ltime[4]+1);
    my $day   = sprintf("%02d", $ltime[3]+1);

    return $year.$month.$day;
}

sub __get_dummy_int {
    my ($self, $size, $value) = @_;
    croak("Size must be a number.") if (!($size =~ m/^\d+$/));
    return $value-1 if (defined $value);

    $size = 9 if $size >= 10;

    return '9' x $size;
}

sub __get_dummy_time {
    my $self = shift;

    my @ltime = localtlime();
    my $time = join('', sprintf("%02d", $ltime[2]), 
                    sprintf("%02d", $ltime[1]));

    return $time;
}

#-----------------------------------------------------------------------------
# Name:   _get_rows_with_missing_reference
# Author: Thomas Malt <tm@linpro.no>
#----------------------------------------------------------------------------- 
#
# Part of 'verify_foreign_key_references'. Queries the current table for any
# rows where the foreign keys are missing.
#
# Query structure: 
#         pk       t                    rt         a      ra          ra 
# SELECT $pk FROM $t l LEFT OUTER JOIN $rt r ON l.$a = r.$ra WHERE r.$ra
#
#-----------------------------------------------------------------------------
sub _get_rows_with_missing_reference {
    my ($self, $table, $foreign_keys_list) = @_;
    
    # Getting the Schema objects off the foreign key list.
    my $attr      = $foreign_keys_list->[0];
    my $ref_table = $foreign_keys_list->[1];
    my $ref_attr  = $foreign_keys_list->[2];
    my $dbh       = $self->get_db_handle();

    # Getting primary key list set up. l is the left table in join:
    my $pk_name   = join(', ', map { 'l.'.$_ } @{$table->get_pk_names()});

    my $query = qq{SELECT $pk_name, l.%s FROM %s l LEFT OUTER JOIN }
              . qq{%s r ON l.%s = r.%s WHERE r.%s IS NULL};
    
    # Stitching the query together
    my @params = ($attr->get_name(),
                  $table->get_name(),
                  $ref_table->get_name(),
                  $attr->get_name(),
                  $ref_attr->get_name(),
                  $ref_attr->get_name(),
                 );

    my $sql = sprintf($query, @params);
    if ($attr->can_be_null()) {
        $sql .= sprintf(' AND l.%s IS NOT NULL', $attr->get_name());
    }

    # Executing the query and accumulating list:
    #     my $sth  = $self->_execute_query($sql);
    my $sth = $dbh->prepare_and_execute($sql);
    my @rows = () ;

    while (my $row = $sth->fetchrow_hashref()) {
        push @rows, $row;
    }

    # Returning a reference to the list.
    return \@rows;
}

=item build_sql ($name, $attr)

Builds an SQL sentence for inserting data into the right tables in the
database.

Used by get_data_and_insert() and _fix_foreign_key_reference

  param   STRING   $name -- Name of table to insert into.
  param   ARRAYREF $attr -- ARRAYREF of attributes to insert
                            into table.
  param   BOOL     $greedy -- Include all attributes, even though they really 
                              shouldn't be exported
  return  STRING         -- sql statement

=cut
sub build_sql {
    my ($self, $name, $attributes, $greedy) = @_;

    my $dbmap  = Noark::Schema::Map->new();
    my $schema = $self->get_noark_schema();
    my $table  = $dbmap->get_schema_table_name($name);
    my $sql    = "INSERT INTO $table (";
    my $qmk    = q||;                    # Question marks for arguments.

    $greedy = 0 if ! defined $greedy;  # The default greedy setting.
    $greedy = 1 if $table eq 'postnr'; # be leanient towards postnr.

    my $strict = $greedy ? 0 : 1; # The opposite of greedy.
    
    for my $attr (@{$attributes}) {
        my $field = undef;
        eval {
            $field = $dbmap->get_schema_field_name($attr);
        };
        if ($@) {
            print STDERR "Found an attribute in NOARKIH that is not part "
                . "of Noark-4.1:\n";
            print $@;
            next;
        }

        # If strict skip attributes that shouldn't be exported.
        if ($strict) {
            next if ! $schema->is_attr_exportable($table, $attr);
        }
        
        $sql .= "$field, "; # Insert the parameters into SQL
        $qmk .= "?, ";      # Adding a questionmark for each parameter.
    }

    $sql =~ s/,\s*$//;             # Removes the last comma from string.
    $qmk =~ s/,\s*$//;             # Same here.

    $sql .= ") VALUES (".$qmk.")"; # Finish our SQL.

    return $sql;
}

=item verify_attributes (row => $row, noark => $def)

Takes attributes found in NOARK.IH (noark) and compares them to
attributes found in the current element in the deposit (row), building
a finished list of attributes to be inserted into the Db.

Two potential errors can occur: An attribute from noark is not
included in the row, that might be ok. An attribute in the row is not
included in noark. The last is serious and should be caught elsewhere
stopping the import.

  param  row      -- HASHREF
  param  noark    -- ARRAYREF
  param  table    -- STRING  Name of table, used for integrity checks.
  param  info     -- HASHREF
  param  file     -- Filename the program is currently parsing

  return ARRAYREF -- List of parsed attrbutes.

Currently this subroutine reports two errors and adds them to the
DBERROR hash: if a NOT NULL attribute is NULL, and if an attribute has
an illegal value according to the Db Schema.

=cut
sub verify_attributes {
    my ($self, %args) = @_;

    my @list        = ();
    my $schema      = $self->get_noark_schema();
    my $name        = $args{info}->{'TI.TABELL'};
    my $file        = $args{file};
    my $line        = $args{line};
    my $unique_seen = $args{unique};
    my $table_name  = $args{table};
    my $table       = $schema->get_table($table_name);
    my $attributes  = $args{noark};

 FIELD:
    for my $field (@$attributes) {
        my $attr  = $table->get_attribute($field);
        next FIELD if ref $attr ne 'Noark::Schema::Attribute';

        my $value = $args{row}->{$field};
        my $type  = $attr->get_data_type();

        # Hm.. where to check for attributes not supposted to be deposited.
        # Check 1: is the attribute supposed to be deposited
        # POSTNR is imported no matter what.
        if ($table_name ne 'POSTNR' && 
            !$schema->is_attr_exportable($table_name, $field) ) {
            next FIELD;
        }

        #---------------------------------------------------------------------
        # Handling null error.
        #---------------------------------------------------------------------
        if ((! defined $value) || ($value eq '')) {
            if ($schema->is_attr_not_null($table_name, $field)) {
                # Handling null error.
                $self->_handle_null_error($name, $field, $file, $line);
                # Replace missing data with dummy data so not to invalidate
                # the whole row.
                $value = $self->__get_dummy_value(
                    $schema->get_attr_data_type($table_name, $field)
                );
            } else {
                $args{row}->{$field} = undef if (! exists $args{row}->{$field});
                if (defined $value) {
                    $self->get_db_errorlist()->add_error(
                        $table_name, 
                        'Noark::Error::EmptyAttribute', 
                        {
                            field => $field,
                            file  => $file,
                            line  => $line,                            
                        }
                    );
                    $value = undef;
                }
            }
        }

        #---------------------------------------------------------------------
        # Handling data not valid error.
        #---------------------------------------------------------------------
        if (! $schema->is_attr_data_valid($table_name, $field, $value)) {
            # Handling data valid error.
            $self->_handle_data_valid_error($name, $field, $value,
                                            $type, $file, $line);

            # If the invalid attribute is a primary key we chose to let this
            # invalidate the whole row.
            my $attr = $schema->get_attribute($table_name, $field);
            return undef if $attr->is_primary_key();

            # Replace missing data with dummy data so not to invalidate
            # the whole row.
            $value = $self->__get_dummy_value(
                $schema->get_attr_data_type($table_name, $field)
            );
        }

        #---------------------------------------------------------------------
        # Handling uniqueness error.
        #---------------------------------------------------------------------
        if ((defined $value) && $schema->is_attr_unique($table_name, $field)) {
	    # Handle uniqueness-requirement.
            if ($unique_seen->{$field}->{$value}) {
		# Handle duplicate occurrence.
                my $first_seen_line = $unique_seen->{$field}->{$value};
                $self->_handle_duplicate_unique(
                    $name, 
                    $field, 
                    $value, 
                    $type,						
                    $file, 
                    $line, 
                    $first_seen_line
                );

                # This invalidates the whole row so we return undef to 
                # skip insertion of row.
                return undef;
            }
            else {
                $unique_seen->{$field}->{$value} = $line;
            }
        }
        push @list, $value;
    }

    return \@list;
}

 
#-----------------------------------------------------------------------------
# Destructor
#-----------------------------------------------------------------------------
# Rydder (fjerner) tempfiler fr avslutning.
#
sub DESTROY {
    my $self = shift;
    system('rm', '-rf', $self->{TMPDIR});
}

1;

__END__

=pod 

=back

=head1 SEE ALSO

=over

=item Noark::Analyzer

=item Noark::Config

=item Noark::Db

=item Noark::Report

=back

=head1 AUTHOR

 Thomas Malt <thomas@malt.no>
 Kjell-Magne ierud <kjellm@linpro.no>

=head1 COPYRIGHT

=over

=item Copyright (C) 2003-2006  Riksarkivaren

=back

=head1 LICENSE

Whatever the licence Riksarkivaren choose to put this under.

=cut
