#!/usr/bin/perl
#-----------------------------------------------------------------------------
# $Id: Filemap.pm 607 2011-06-01 17:14:05Z tm $
#-----------------------------------------------------------------------------
# Noark::Parser::Filemap - object representing a map between the files found 
#                          in NOARKIH and the filesystem.
#

package Noark::Parser::Filemap;

use Carp;
use IO::File;


=head1 CONSTRUCTOR

=over

=item new ($args)

Creates a new filmap object

=back

=cut
sub new {
    my ($class, $arg) = @_;

    if (! defined $arg->{list}) {
        confess("You must supply a tabellinfo list to constructor");
    }

    confess("You must supply xmldir as part of constructor")
        if not defined $arg->{xmldir};

    confess("'docdir' is a mandatory argument to constructor")
        if not defined $arg->{docdir};

    confess("'reportdir' is mandatory argument to constructor")
        if not defined $arg->{docdir};

    my $self = bless ({}, $class);

    $self->{ROOTDIR}   = $arg->{dir} || '';
    $self->{INFO}      = $arg->{list};
    $self->{XMLDIR}    = $arg->{xmldir};
    $self->{DOCDIR}    = $arg->{docdir};
    $self->{REPORTDIR} = $arg->{reportdir};
    $self->{ROOTDIR}   =~ s/\/$//; # Strip of trailing / if it exists.

    $self->_map_files($self->{ROOTDIR});

    return $self;
}


sub is_file_in_deposit {
    my ($self, $filename) = @_;

    return defined $self->{FILEMAP}->{$filename} ? 1 : 0;
}

sub get_file_by_name {
    my ($self, $name) = @_;

    return $self->{FILEMAP}->{$name} || undef;
}

sub get_key_by_filename {
    my ($self, $filename) = @_;

    for my $key (keys %{$self->{FILEMAP}}) {
        return $key if ($self->{FILEMAP}->{$key} eq $filename);
    }

    return undef;
}

=item get_filename_keys

Returns the keys in the filemap as an arrayref

=cut
sub get_filename_keys {
    my $self = shift;

    return [keys %{$self->{FILEMAP}}];
}


=item get_files_by_doctype

Returns the files in deposit with the given doctype.

=cut
sub get_files_by_doctype {
    my ($self, $doctype) = @_;
    
    return $self->{DOCTYPES}->{$doctype} || [];
}

# Fetches all the XML files into a filmap for later use
sub _map_files {
    my ($self, $dir) = @_;

    my @dirs = ($dir, $dir.'/'.$self->{XMLDIR}, $dir.'/'.$self->{REPORTDIR});

    for my $dd (@dirs) {
        opendir my $d, $dd or confess("Could not open dir: $!");

        for my $item (readdir $d) {
            next if $item =~ /^\./;
            next unless $item =~ /xml$/i;

            $doctype = $self->_get_doctype_of_file($dd.'/'.$item);
            if (defined $doctype) {
                push @{$self->{DOCTYPES}->{$doctype}}, $dd.'/'.$item;
            }
            $self->{FILEMAP}->{uc $item} = $dd.'/'.$item;
        }

        closedir($d);
    }

    return 1;
}


sub _get_doctype_of_file {
    my ($self, $file) = @_;

    my $fh = IO::File->new($file, 'r') or die("could not open '$file': $!");
    
    while (my $line = <$fh>) {
        chomp $line;

        next if !($line =~ /<\!DOCTYPE (\S+) /);
        my $doctype = $1;
        $fh->close();

        return $doctype;
    }

    $fh->close();

    return undef;
}


1;
