#!/usr/bin/perl
#-----------------------------------------------------------------------------
# $Id: Tabellinfo.pm 798 2011-06-28 13:41:47Z tm $
#-----------------------------------------------------------------------------
# Object for handling parsed tabellinfo objects
#

=head1 NAME

Noark::Parser::Tabellinfo - Holder of tabellinfos.

=head1 SYNOPSIS

my $info = Noark::Parser::Tabellinfo->singleton()

=head1 DESCRIPTION

This is a singleton. Only one object should exist within a process.

=cut

package Noark::Parser::Tabellinfo;

use strict;
use warnings;

use Carp;

sub new {
    my ($class, %args) = @_;

    my $self = bless({}, $class);

    confess("You must supply an element") if ! defined $args{element};
    
    $self->_parse_tabellinfo_element($args{element});
    $self->_setup_files();

    $self->{'TI.COUNT'}      = 0;    
    $self->{FILE_COUNT}      = @{$self->{FILES}};
    $self->{FILDEL_OK}       = $self->_calc_fildel_ok();

    return $self;
};


sub _setup_files {
    my $self = shift;

    for my $file (@{$self->{FILES}}) {
        $file->{'TI.FILNAVN'} ||= '';
    }
}

=head1 SUBROUTINES

=over

=item get_table_name

Returns the table name as a string.

=cut
sub get_table_name {
    my $self = shift;

    return $self->{'TI.TABELL'};
}


=item set_table_name ($name)

Sets the table name 

=cut
sub set_table_name {
    my ($self, $name) = @_;

    $self->{'TI.TABELL'} = $name;
}


=item is_fildel_ok

return whether all files has a TI.FILDEL

=cut
sub is_fildel_ok {
    my $self = shift;

    return $self->{FILDEL_OK};
}

=item get_ih_number_of_files

A Noark-4.1 table can be deposited on many files. This returns the number of
files in deposit.

=cut
sub get_ih_number_of_files {
    my $self = shift;

    return $self->{'TI.ANTFILER'};
}

sub set_file_count {
    my ($self, $count) = @_;
    $self->{FILE_COUNT} = $count;

    return 1;
}

sub get_file_count {
    my $self = shift;
    
    return $self->{FILE_COUNT};
}


=item get_files_ref

Returns an array ref with information about the files.

=cut
sub get_files_ref {
    my $self = shift;

    return $self->get_sorted_files_ref();
}


sub get_sorted_files_ref {
    my $self = shift;
    
    return $self->{SORTED_FILES} if defined $self->{SORTED_FILES};

    my @sorted = sort { 
        $a->{'TI.FILDEL'} <=> $b->{'TI.FILDEL'} 
    } @{$self->{FILES}};

    $self->{SORTED_FILES} = \@sorted;
    return $self->{SORTED_FILES};
}

=item get_line_number

Returns the line number of the tabellinfo in NOARKIH.XML

=cut
sub get_line_number {
    my $self = shift;

    return $self->{LINE_NUMBER};
}

=item increment_counter

The object has an internal counter keeping track of how many times the
tableinfo occurs.

=cut
sub increment_counter {
    my $self = shift;

    $self->{'TI.COUNT'} += 1;
    return $self->{'TI.COUNT'};
}


=item get_counter

Returns the internal counter.

=cut
sub get_counter {
    my $self = shift;

    return $self->{'TI.COUNT'};
}



=item get_attributes 

Returns an ARRAYREF of the attributes in the table.

=cut
sub get_attributes {
    my $self = shift;
    return $self->{ATTRIBUTTER};
}



=item get_total_antposter

Returns the sum of all TI.ANTPOSTER in NOARK.IH.

=cut
sub get_total_antposter {
    my $self = shift;
    
    return $self->{TOTAL_ANTPOSTER} if defined $self->{TOTAL_ANTPOSTER};

    $self->{TOTAL_ANTPOSTER} = 0;
    for my $file (@{$self->{FILES}}) {
        $self->{TOTAL_ANTPOSTER} += $file->{'TI.ANTPOSTER'} || 0;
    }
    
    return $self->{TOTAL_ANTPOSTER};
}

sub get_total_xmlposter {
    my $self = shift;

    return $self->{__total_xmlposter} if defined $self->{__total_xmlposter};
    $self->{__total_xmlposter} = 0;
    for my $file (@{$self->{FILES}}) {
        $self->{__total_xmlposter} += $file->{'XML.ANTPOSTER'} || 0;
    }

    return $self->{__total_xmlposter};
}

sub is_count_consistent {
    my $self = shift;

    return ($self->get_total_antposter == $self->get_total_xmlposter);
}

=pod

=back

=cut


sub _parse_tabellinfo_element {
    my ($self, $element) = @_;
    $self->{LINE_NUMBER} = $element->att('#line');
    
    for my $node ($element->children()) {
        my $name = $node->name;
        
        if ($name eq 'FIL') {
            my %fil = map {$_->name => $_->text} $node->children; 
            push @{$self->{FILES}}, \%fil;
        }
        elsif ($name eq 'ATTRIBUTTER') {
            my @attributes = map { $_->text } $node->children;
            $self->{$name} = \@attributes; 
        } 
        else {
            $self->{$name} = $node->text;
        }
    }    
}


sub _calc_fildel_ok {
    my $self = shift;

    return 1 if ($self->{'TI.ANTFILER'} == 1);
    
    for my $file (@{$self->{FILES}}) {
        next if defined $file->{'TI.FILDEL'};
        return 0;
    }
    
    return 1;
}

1;
__END__
