#!/usr/bin/perl
#-----------------------------------------------------------------------------
# $Id: TabellinfoList.pm 580 2011-06-01 11:33:53Z tm $
#-----------------------------------------------------------------------------
# Object for handling a list of tabellinfo objects.
#


=head1 NAME

Noark::Parser::TabellinfoList - List of tabellinfo objects

=head1 SYNOPSIS

  my $list = Noark::Parser::TabellinfoList->new({ih_dom => $noarkih_dom_ref});

=head1 DESCRIPTION

TabellinfoList is a singleton object containing a list of all
tabellinfo-elements in noarkih.

=cut

package Noark::Parser::TabellinfoList;

use strict; 
use warnings;

use Carp;
use IO::File;
use File::Basename;
use Noark::Parser::Tabellinfo;

use XML::Twig;

my %root = ();

=head1 SUBROUTINES

=over

=item new ({ih_dom => $dom_ref});

Constructor takes one mandatory argument, a reference to the
XML::LibXML::Document object of the NOARKIH.XML file.

=cut
sub new {
    my ($class, %arg) = @_;

    return $root{singleton} if ref($root{singleton}) eq $class;

    my $self = bless([], $class); # We want to use object as list
    if(!defined($arg{file})) {
        croak ("\nCould not find a NOARKIH file in deposit. A NOARKIH.XML " .
               "file must be present.\n" .
               "The filename must be uppercase. \n" .
               "Sometimes files are deposited with lowercase filenames, " .
               "but this is in conflict with the NOARK-4.1 standard."
              );        
    } 
    elsif(! -r $arg{file}) {
        croak "Could not open NOARKIH file (".$arg{file}.")";
    }
    else {
        $root{FILE}          = $arg{file};
    }
    $root{IMPORT_ORDER}  = $self->get_import_order();
    $root{DEPOSITED}     = {};
    $root{LIST}          = $self->_parse_tabellinfo_dom();

    for my $table (@{$root{IMPORT_ORDER}}) {
        $root{DEPOSITED}->{$table} = $root{LIST}->{$table} ? 1 : 0;
        next if ! $root{LIST}->{$table};

        push @{$self}, $root{LIST}->{$table};
    }

    # Put entries that are not know to arkn4 in an hash for storage
    $root{UNKNOWN} = {};
    for (keys %{$root{LIST}}) {
        next if exists $root{DEPOSITED}->{$_};
        $root{UNKNOWN}->{$_} = $root{LIST}->{$_};
    }

    $root{singleton}    = $self;

    return $self;
}

=item get_table_names_by_name

Returns an ARRAYREF with the table names ordered by name.

=cut
sub get_table_names_by_name {
    return [sort keys %{$root{LIST}}];
}

=item get_tabellinfo_by_name ($name)

Returns the tabellinfo object reffered to by name

=cut
sub get_tabellinfo_by_name {
    my ($self, $name) = @_;

    return $root{LIST}->{$name} || undef;
}

sub get_tabellinfo_count {
    my $self = shift;
    
    return scalar @{$self};
}

sub get_deposited_hash {
    return $root{DEPOSITED};
}

sub _parse_tabellinfo_dom {
    my $self = shift;
    my %list = ();
    my $doc  = XML::Twig->new( 
        start_tag_handlers => {
          TABELLINFO => sub {
            my ($t, $elt) = @_;
            # Add the line number of the start tag
            $elt->set_att('#line' => $t->current_line);
          },
        }, 
        twig_handlers => {
          TABELLINFO => sub {
            my ($t, $elt) = @_;
            my $tabellinfo = Noark::Parser::Tabellinfo->new(element => $elt);
            my $name       = $tabellinfo->get_table_name();
            $list{$name}   = $tabellinfo if ! $list{$name};
            $list{$name}->increment_counter();
            $t->purge();
          },
        }, 
    );
    
    eval {
        $doc->parsefile($root{FILE});
    };
       
    return \%list;
}

sub get_unknown_tabellinfos {
    my $self = shift;

    return $root{UNKNOWN};
}

sub get_import_order {
    my $self = shift;
    
    return $root{IMPORT_ORDER} if ( defined $root{IMPORT_ORDER} &&
                                    scalar @{$root{IMPORT_ORDER}} );

    my @list = ();

    my $def  = dirname($INC{'Noark/Parser.pm'}).'/Config/import_order.conf';
    my $file = "./import_order.conf";

    $file = "/etc/arkn4/import_order.conf" if not -r $file;
    $file = $def if not -r $file;
    
    unless (-r $file) { 
        croak("ERROR: Could not read '$file: $!'\nPWD: ".qx|pwd|."\n".$0);
    }
    
    my $fh = IO::File->new($file, 'r') or die("Oops: $!");
    for (<$fh>) { 
        chomp;
        next if /^\s*(#|\s*$)/;
        push (@list, $_);
    } 
    $fh->close;

    return \@list; 
}


return 1;

__END__

=pod

=back

=cut

