#-----------------------------------------------------------------------
# $Id$
#-----------------------------------------------------------------------
# Noark::Progress
#
# Module for printing progress reports when running arkn4.
#
# Author: Thomas Malt <thomas@malt.no>
#
# Copyright (c) 2009-2011 Riksarkivaren
#

package Noark::Progress;

use Carp::Assert;

use strict;
use warnings;

sub new {
    my ($proto, %args) = @_;
    my $class = ref $proto || $proto;
    
    my $self = {
        __verbosity             => $args{verbosity} || 0,
        __current_string        => $args{string}    || "",
        __start_time            => $args{time}      || time(),
        __current_size          => 0,
        __current_done          => 0,
        __last_output_time      => 0,
        __current_string_length => 0,
    };
    
    bless($self, $class);

    return $self;
}

sub setup_status {
    my ($self, %args) = @_;

    # print "Got size: ".$arg->{size}."\n" if defined $arg->{size};

    $self->{__start_time}            = $args{time} || time();
    $self->{__current_size}          = $args{size} || 0;
    $self->{__current_string}        = $args{string} || "";
    $self->{__current_done}          = 0;
    $self->{__last_output_time}      = 0;
    $self->{__current_string_length} = 0;

    return 1;
}

sub set_current_string {
    my ($self, $string) = @_;
    confess("String must be set!") if ! defined $string;

    $self->{__current_string} = $string;

    return 1;
}

sub set_start_time {
    my ($self, $time) = @_;
    
    $time ||= time();
    
    $self->{__start_time} = $time;
    return $time;
}

sub set_current_size {
    my ($self, $size) = @_;
    
    assert(defined $size);
    
    $self->{__current_size} = $size;
}

sub increment_done_by_amount {
    my ($self, $amount) = @_;
    
    # If amount is not given we assume by one.
    $amount ||= 1;
    
    $self->{__current_done} += $amount;
    
    return $self->{__current_done};
}

sub get_current_done {
    my $self = shift;
    return $self->{__current_done};
}

sub finish_and_print_status {
    my ($self, %args) = @_;

    my $result = $args{ok} ? '[OK]' : '[FAIL]';
    my $string = $self->{__current_string}.':';
    my $time   = $self->get_timestr( time() - $self->{__start_time} );

    print "\b" x $self->{__current_string_length};
    printf "%-55s % 6s -- Time: %7s\n", $string, $result, $time;

    return 1;
}

sub print_status {
    my ($self, %args) = @_;

    $self->{__current_done} = $args{byte} if $args{byte};

    return undef if ((time() - 1.5) < $self->{__last_output_time});
    
    $self->{__last_output_time} = $args{now} || time();
    
    my $time      = $self->{__last_output_time};
    my $start     = $self->{__start_time};
    my $size      = $self->{__current_size} || 0;
    my $size_len  = length($size);
    my $byte      = $self->{__current_done} || 0;
    my $prefix    = $self->{__current_string}.":";
    my $percent   = $size ? ($byte/$size) : '0';
    my $st_len    = 55 - length($prefix) - 4;
    my $time_done = $time-$start;
    my $time_left = $percent ? ($time_done/$percent)-$time_done : 0.00;
    my $time_tot  = $percent ? ($time_done/$percent) : 0.00;
    my $done_len  = $size ? sprintf("%.0f", ($st_len*($byte/$size))) : "0";
    my $done_str  = "=" x $done_len;
    my $done_pct  = $size ? sprintf("%.2f", (100*$percent)) : "0.00";
    my $done_time = $percent ? $self->get_timestr($time_left, 0) : 'xx:xx';
    my $done_tot  = $percent ? $self->get_timestr($time_tot, 0)  : 'xx:xx';

    my $string    = sprintf ("%s [%-".$st_len."s][%-4s] %5s%%|ETA %7s", 
                             $prefix, 
                             $done_str, 
                             $self->_human_size($size), 
                             $done_pct, 
                             $done_time);
    
    print "\b" x $self->{__current_string_length};
    print $string;
    
    $self->{__current_string_length} = length($string);
    return 1;
}

sub _human_size {
    my ($self, $size) = @_;
    
    my $hsize = "";
    if ($size < (1000**2)) {
        $hsize = int(($size/1000))."K";
        return $hsize;
    }
    
    if ($size >= (1000**2) && $size < (1000**3)) {
        $hsize = int(($size/(1000**2)))."M";
        return $hsize;
    }

    $hsize = int($size/(1000**3))."G";
    
    return $hsize;
}

sub print_rapport_verb {
    my ($self, %args) = @_;
    
    my $valid      = $args{is_valid};
    my $start      = $args{start_time};
    my $time       = $self->get_timestr( time() - $start );
    my $string     = $self->{__current_string}.':';

    print "\b" x $self->{__current_string_length};

    $self->{__current_string_length} = 0;
    $self->{__current_string}        = '';
    $self->{__start_time}            = undef;
    $self->{__current_size}          = 0;

    my $answer = $valid ? "[YES]" : "[NO]";

    printf "%-56s %5s -- Time: %7s\n", $string, $answer, $time;

    return 1 if ($self->{__verbosity} < 2);

    print "      Saker:  ", $args{noarksak_count},  "\n" if defined $args{noarksak_count};
    print "      Poster: ", $args{journpost_count}, "\n" if defined $args{journpost_count};

    return 1;
}

sub print_traverse_deposit_res {
    my ($self, %args) = @_;

    my $rows_reported  = $args{reported};
    my $rows_inserted  = $args{inserted};
    my $start          = $args{start};
    my $is_ok          = $args{is_ok};

    my $string = sprintf("     Consistency: Rows: [%7s|%7s]",
                         $rows_reported, $rows_inserted);

    printf "%-55s %6s -- Time: %7s\n", $string, $is_ok ? '[OK]' : '[FAIL]',
	$self->get_timestr( time() - $start);
}

sub get_timestr {
    # Utility function which returns a "pretty" formatted string.
    my ($self, $time, $decimal) = @_;
    assert(defined $time);

    $time = 0 if $time < 0;

    $decimal ||= 0;

    my $min = int($time/60);
    my $sec = $time-($min*60);

    $min = sprintf("%02i", $min);
    $sec = sprintf("%07s", sprintf("%.".$decimal."f", $sec)) if $decimal;
    $sec = sprintf("%02s", sprintf("%.".$decimal."f", $sec)) if ! $decimal;
    
    return "$min:$sec";
}

1;
__END__

=head1 NAME

Noark::Progress - Printing progress reports to STDOUT when running arkn4

=head1 SYNOPSIS

TBA

=head1 DESCRIPTION

TBA

=cut
