#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Report.pm 811 2011-07-15 09:38:58Z tm $
#---------------------------------------------------------------------
# Noark::Report - Framework for generating reports.
#
# Author: Thomas Malt <tm@linpro.no>
#
# Copyright (C) 2003-2006  Riksarkivaren.
# Copyright (C) 2003-2006  Thomas Malt <tm@linpro.no>
#
#---------------------------------------------------------------------

=head1 NAME

Noark::Report - A Noark module for generating reports.

=head1 SYNOPSIS

my $foo = Noark::Report->new();

$foo->write_log(LOGNAME, $message);

=head1 DESCRIPTION

This modules contains functionality for the various reports we define
the Importer should use. The names of the reports are gotten from the
configuration file, and this object keeps the filehandles fo the
various log files.

Noark::Report can easily be extended to handle reports for various
other Noark applications, like the Viewer or others.

=cut

package Noark::Report;

use strict;

use Exporter;
use IO::File;
use Data::Dumper;

use Noark::Helper;
use Noark::Schema::Map;
use Noark::Report::Text;
use Noark::Report::HTML;

our ($VERSION) = '$Revision: 811 $' =~ /: (\S+)/;


=head1 FUNCTIONS

=over 4

=item new ()

Constructor. Creates and keeps filehandles of all logfiles defined
in the configuration file.

=cut
sub new {
    my $class = shift;
    my %args  = @_;

    croak("Cannot clone copy of self.") if (ref($class) eq __PACKAGE__);

    my %formats = (
        txt  => 'Noark::Report::Text',
        html => 'Noark::Report::HTML',
        pdf  => 'Noark::Report::PDF'
    );

    $args{'dir'}      = "."              if not defined $args{'dir'};
    $args{'name'}     = "default_report" if not defined $args{'name'};
    $args{'format'}   = "txt"            if not defined $args{'format'};

    if (not $formats{$args{'format'}}) {
        croak("You must specify which format to generate report in.") 
    }

    my $self = $formats{$args{'format'}}->new(%args);

    $self->{'VERSION'}    = $VERSION;
    $self->{'FORMAT'}     = $args{'format'};
    $self->{'NAME'}       = $args{'name'};
    $self->{'DIR'}        = $args{'dir'};
    $self->{'DEBUG'}      = $args{'debug'}      || 0;
    $self->{'VERB'}       = $args{'verbose'}    || 0;
    $self->{'MAX_ERRORS'} = $args{'max_errors'} || 10;

    # Paragraph sections
    $self->{'SEC'}    = [0,0,0,0];
    $self->{'REPORT'} = $self->get_report_handle();

    return undef unless $self->{'REPORT'};
    return $self;
}



=item print_report ($an)

Main subroutine for printing out the report. The analyzer-object is
given as argument and the whole thing is parsed through.

  param Noark::Analyzer $an -- The analyzer object.

=cut
sub print_report {
    my ($self, $an) = @_;
    
    my $infolist    = $an->get_tabellinfo_list();
    my $map         = Noark::Schema::Map->new();
    my $table_names = $infolist->get_table_names_by_name();
    my $db          = $an->get_db_errorlist();

    
    print "  Writing XML Valid errors ..." if ($self->{VERB} > 1);
    $self->write_validity_head();
    #-------------------------------------------------------------------------
    # XML-gyldighetssjekker
    #-------------------------------------------------------------------------
    my $xml_count = 0;
    my $xml_error = 0;

    for (qw|NOARKIH JOURNAL SAKDOKOVERSIKT|, @{$table_names} ) {
        unless ($_ =~ /(NOARKIH|JOURNAL|SAKDOKOVERSIKT)/) {
            next if ! $map->is_noark_table_name($_);
        }
        my $__errors = $an->get_xml_errors_by_name($_);
        $xml_error += scalar @$__errors;
        $xml_count += 1;
        if (@$__errors) {
            $self->write_table(
                $__errors,
                {NAME => $_ },
            );
        }
    }
    $self->write_validity_tail($xml_count, $xml_error);
    print " Done\n" if ($self->{VERB} > 1);


    print "  Writing Consistency errors ..." if ($self->{VERB} > 1);
    $self->write_consistency_head();
    #-------------------------------------------------------------------------
    # Konsistenssjekker
    #-------------------------------------------------------------------------
    my $data_error = 0;
    my $__ih_errors = $an->get_data_errors_by_name('NOARKIH');
    my $data_errors = $an->get_data_errorlist();
    $data_error += scalar @$__ih_errors;

    if (@$__ih_errors) {
        $self->write_table(
            $__ih_errors,
            {NAME => 'NOARKIH'}
        );
    }
    print " Done\n" if ($self->{VERB} > 1);


    print "  Writing extra files test ..." if ($self->{VERB} > 1);
    $self->write_extrafiles_test($an->get_extra_files());
    for (@{$table_names}) {
        next if not $map->is_noark_table_name($_);

        my $__errors = $an->get_data_errors_by_name($_);
        my $__info   = $infolist->get_tabellinfo_by_name($_);
        
        if (@$__errors || ! $__info->is_count_consistent()) {
            $data_error += scalar @$__errors;
            $self->write_table($__errors, $__info);
        }
        
        if (! $__info->is_count_consistent()) {
            $data_error += 1;
            $self->write_count(
                $__info
            );
        }
    }
    if ($data_error == 0) {
        $self->write_consistency_tail();        
    }
    print " Done\n" if ($self->{VERB} > 1);

    print "  Writing mandatory tables test ..." if ($self->{VERB} > 1);
    $self->write_mandatory_tables_test($an->get_mandatory_tables_test());
    print " Done\n" if ($self->{VERB} > 1);

    print "  Writing Frequency Analysis ..." if ($self->{VERB} > 1);
    $self->write_freq_analysis($an->get_frequency_analysis());
    print " Done\n" if ($self->{VERB} > 1);

    print "  Writing numbering test ..." if ($self->{VERB} > 1);
    $self->write_numbering_test($an->do_numbering_test());
    print " Done\n" if ($self->{VERB} > 1);

    print "  Writing interval test ..." if ($self->{VERB} > 1);
    $self->write_interval_test($an->do_interval_test());
    print " Done\n" if ($self->{VERB} > 1);

    #-------------------------------------------------------------------------
    # Database import rapport
    #-------------------------------------------------------------------------
    print "  Writing Database import errors ..." if ($self->{VERB} > 1);
    $self->write_dbimport_head();
    for my $t (sort $db->get_keys()) {
        if (! $map->is_noark_table_name($t)) {
            croak("Invalid table name ($t) got inserted in db error list");
        }

        $self->write_table(
            $db->get_errors_as_arrayref_by_key($t),
            $infolist->get_tabellinfo_by_name($t)
        );
    }
    print " Done\n" if ($self->{VERB} > 1);

    #-------------------------------------------------------------------------
    # Document reference report
    #-------------------------------------------------------------------------
    print "  Writing document checks ..." if ($self->{VERB} > 1);
    $self->write_documents_head();
    $self->write_documents_check($an);
    print " Done\n" if ($self->{VERB} > 1);
    
    print "  Writing end summary ..." if ($self->{VERB} > 1);
    $self->write_summary($an->{'WARNINGS'}, $an->{'FATALS'});
    print " Done\n" if ($self->{VERB} > 1);
}



=item get_section ()

Return the current section number for the report 
ie. 1.

=cut
sub get_section {
    my $self = shift;
    return $self->{'SEC'}->[0];
}

=item get_subsection ()

Return the current subsection of the report.
ie. 1.1

=cut
sub get_subsection {
    my $self = shift;
    return $self->{'SEC'}->[1];
}

=item get_subsubsection ()

Return the current subsubsection of the report
ie. 1.1.1.

=cut
sub get_subsubsection {
    my $self = shift;
    return $self->{'SEC'}->[2];
}


=item get_max_errors ()

Returns the maximum number of errors to output.

=cut
sub get_max_errors {
    my $self = shift;

    return $self->{MAX_ERRORS};
}

=item get_section_string ($level)

Returns a string represening the current section.

Takes an integer signifying the level off section to return ie:
"<1>.<2>.<3>."

=cut
sub get_section_string {
    my ($self, $level) = @_;

    return undef unless defined $level;

    my $str = "";    
    my $i = 0;
    while ($i < $level) {
        $str .= $self->{'SEC'}->[$i].".";
        $i += 1;
    }
    return $str;
}

=item increment_section

=cut
sub increment_section {
    my $self = shift;
    $self->{'SEC'}->[0] += 1;
    $self->{'SEC'}->[1] = 0;
    $self->{'SEC'}->[2] = 0;
}

=item increment_subsection
=cut
sub increment_subsection {
    my $self = shift;
    $self->{'SEC'}->[1] += 1;
    $self->{'SEC'}->[2] = 0;
}

=item increment_subsubsection
=cut
sub increment_subsubsection {
    my $self = shift;
    $self->{'SEC'}->[2] += 1;
}


sub DESTROY {
    my ($self) = @_;

    undef $self->{'IMPORT'};
    undef $self->{'ERROR'};
    undef $self->{'DEBUG'};
    undef $self->{'ANALYZE'};
}


1;

__END__

=back

=head1 SEE ALSO

=over

=item Noark::Analyzer

=item Noark::Db

=item Noark::Report

=item Noark::Parser

=back

=head1 AUTHOR

Thomas Malt <tm@linpro.no>

=head1 COPYRIGHT

=over

=item Copyright (C) 2003  Riksarkivaren

=back

=head1 LICENSE

Whatever the licence Riksarkivaren choose to put this under.

=cut
