#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Text.pm 814 2011-07-15 12:05:44Z tm $
#---------------------------------------------------------------------
# Noark::Report::Text - Framework for generating reports.
#
# Author: Thomas Malt <thomas@malt.no>
#
# Copyright (C) 2003-2011  Riksarkivaren.
#
#---------------------------------------------------------------------
package Noark::Report::Text;

use strict;

use Data::Dumper;
use Carp;
use Carp::Assert;

use base ('Noark::Report');

use constant { SECSEP => "\n\n\n" };

sub new {
    my $proto = shift;
    my %args  = @_;

    my $class = ref($proto) || $proto;
    my $self  = {};

    $self->{'FORMAT'} = $args{'format'};
    $self->{'NAME'}   = $args{'name'};
    $self->{'DIR'}    = $args{'dir'};
    $self->{'DEBUG'}  = $args{'debug'};
    $self->{'VERB'}   = $args{'verbose'};

    $self->{'REPORT'} = undef;
    
    bless($self, $class);
    return $self;
}


#-----------------------------------------------------------------------------
# Name: get_report_handle
#
# Returns the filehandle of where to write the report.
# Why is this in Noark::Report::Text and not in Noark::Report.
#
# Hm.. Maybe because we want different filehandles for different output 
# formats
sub get_report_handle {
    my ($self) = shift;

    return $self->{'REPORT'} if ref($self->{'REPORT'}) =~ /IO::File/;

    $self->{REPORT} = IO::File->new();
    if ($self->{NAME} eq '-') {
        use I18N::Langinfo qw(langinfo CODESET);
        
        $self->{REPORT}->fdopen(fileno(STDOUT), 'w') or die("Oops: $!");
        $self->{REPORT}->binmode( ':encoding('.langinfo(CODESET).')' );
    } 
    else {
        $self->{'REPORT'}->open($self->{'DIR'}."/".$self->{'NAME'}.".txt",'w');
        $self->{'REPORT'}->binmode(':encoding(ISO-8859-1)');
    }

    $self->{'REPORT'}->autoflush(1);

    return $self->{'REPORT'}; 
}


sub write_noarkih_error {
    my ($self, $msg) = @_;

    my $fh = $self->get_report_handle();
    
    $self->increment_subsection();
    my $sec = $self->get_section_string(2);
    
    print $fh <<EOF;

$sec  FEIL MED NOARKIH

  En feil oppstod under parsingen av NOARKIH.XML. Dette er feilmeldingen
  fra NOARK-parseren: 

$msg

EOF

    return 1;
}



=item write_eksportinfo ($ei)

Prints eksportinfo information in the report.

Added extra space for date in cases where the deposit has wierd date
date.

=cut
sub write_eksportinfo {
    my ($self, $ei) = @_;

    my $fh = $self->get_report_handle();

    my @free = ('EI.ORGNR', 'EI.KOMMUNE', 'EI.BASEID', 'EI.MERKNAD');

    $self->increment_subsection();
    my $sec = $self->get_section_string(2);

    for (@free) {
        $ei->{$_} = "<ikke oppgitt>" if not defined $ei->{$_};
    }

    format EI =
@<<<< EKSPORTINFO
$sec

  Arkivskaper.......: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                      $ei->{'EI.ARKSKAPER'}
  Systemnavn........: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                      $ei->{'EI.SYSTEMNAVN'}
  Periode...........: @<<<<<<<<< - @<<<<<<<<<
                      $ei->{'EI.FRADATO'}, $ei->{'EI.TILDATO'}
  Produksjonsdato...: @<<<<<<<<<<<<<<<<<<<<<<
                      $ei->{'EI.PRODDATO'}
  Kommunenummer.....: @<<<<<<<<<<<<<<<<<<<<<<
                      $ei->{'EI.KOMMUNE'}
  Org.nummer........: @<<<<<<<<<<<<<<<<<<<<<<
                      $ei->{'EI.ORGNR'}
  ID for arkivbase..: @<<<<<<<<<<<<<<<<<<<<<<
                      $ei->{'EI.BASEID'}
  Merknad...........: ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                      $ei->{'EI.MERKNAD'}
                      ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<~~
                      $ei->{'EI.MERKNAD'}

.

    $fh->format_name('EI');
    $fh->format_write();

    return 1;
} 


sub write_documents_head {
    my $self = shift;

    $self->_write_head("KONTROLL AV AVLEVERTE DOKUMENTER");
}

sub write_dbimport_head {
    my $self = shift;
    $self->_write_head("IMPORT AV DATA TIL NOARK-DATABASE");
}


sub write_validity_head {
    my $self = shift;
    $self->_write_head("GYLDIGHETSSJEKK AV XML-FILENE");
}


sub write_consistency_head {
    my $self = shift;
    $self->_write_head("KONTROLL AV INFORMASJON I NOARKIH");
}

sub write_validity_tail {
    my ($self, $total, $errors) = @_;

    my $fh  = $self->get_report_handle();
    my $sep = ""; 

    if ($errors > 0) {
        $self->increment_subsection();
        my $sec = $self->get_section_string(2);
        $sep = "  ";
        print $fh "  $sec Oppsummering\n\n";
    }

    printf $fh "$sep  %-36s: %3s\n", 
        "Antall XML-filer nevnt i NOARKIH", $total;
    printf $fh "$sep  %-36s: %3s\n", 
        "Antall XML-filer som ikke validerte", $errors; 

    if ($errors == 0) {
        print $fh "\n";
        print $fh "$sep  Alle XML-filer var gyldige.\n";
    }

}

sub write_consistency_tail {
    my $self = shift;
    my $fh = $self->get_report_handle();

    print $fh "  All informasjon i NOARKIH var gyldig.\n";

    return 1;
}

sub _write_head {
    my ($self, $head) = @_;
    my $fh = $self->get_report_handle();
    $self->increment_section();
    my $sec = $self->get_section_string(2);

    print $fh SECSEP;
    print $fh "$sec $head";
    print $fh "\n\n";

}

sub write_summary {
    my ($self, $warnings, $fatals) = @_;

    my $fh = $self->get_report_handle();

    $self->increment_section();
    my $sec = $self->get_section_string(2);

    my $sum = ($warnings + $fatals);
    my $sep = "    -------------------------------";
    print $fh SECSEP, "$sec OPPSUMMERING\n\n";

    format OP = 
    Ikke kritiske feil...: @>>>>>>
                           $warnings
    Kritiske feil........: @>>>>>>
                           $fatals
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$sep
    Totalt...............: @>>>>>>
                           $sum 
    ===============================

.
    
    $fh->format_name('OP');
    $fh->format_write();
    
    return 1;
}


=item print_histogram ($data, $fh)

Helper used by write_freq_analysis. Takes two arguments: Datastructure
to write histogram of and type of data to write.

TODO This is not a general function

Returns TRUE on success.

=cut
sub print_histogram {
    my ($self, $data, $name) = @_;

    $name = sprintf("% 6s", $name);
    my $fh     = $self->get_report_handle();
    my $width  = 54;
    my $margin = 19;
    my $max    = $self->get_max($data) || 1;
    my $fac    = ($width/$max);
    my $sum    = 0;
    print $fh " " x $margin, "0", " " x ($width+1-(length($max)/2)),"$max\n";
    print $fh " " x  6, "r    ", $name, " |", "-" x ($width), "|\n";
    for my $year (sort { $a <=> $b } keys(%{$data})) {
        $sum += $data->{$year};
        
        my $cnt = sprintf("% 6d", $data->{$year});
        if (length($cnt) > 6 && $cnt =~ /^\s/) {
            ($cnt) = $cnt =~ /^\s(.*)/;
        }
        my $hsh = "#" x int($data->{$year}*$fac);

        print $fh "      $year  $cnt :$hsh\n";

    }
    print $fh " " x 6, '-' x 13, '|', '-' x 54, "|\n";
    print $fh " " x 6, "Sum: ", sprintf("% 7d", $sum), "\n";
    print $fh " " x 6, '=' x 69, "\n";
    print $fh "\n";

    return 1;
}

sub print_journalpost_count {
    my ($self, $list) = @_;

    my $fh = $self->get_report_handle();

    $self->increment_subsubsection();
    my $sec = $self->get_section_string(3);
    print $fh "    $sec Saker med null eller feil antall journalposter \n\n";

    if (not scalar @$list) {
        print $fh " " x 6;
        print $fh "Alle saker hadde riktig antall journalposter.\n\n";
        return 1;
    }

    my $plural = (scalar @$list > 1) ? "saker" : "sak";
    print $fh " " x 6;
    print $fh (scalar @$list)." $plural hadde 0 eller feil antall journalposter:\n\n";
    printf $fh "%s  %10s  %8s  %8s  %9s  %-24s\n", 
        " " x 6, "Sak ID", "SA.ANTJP", "Funnet", "SA.STATUS", "SA.TITTEL";
    printf $fh "%s  %10s  %8s  %8s  %9s  %24s\n", 
        " " x 6, "-" x 10, "-" x 8, "-" x 8, "-" x 9, "-" x 24;

    for my $item (@$list) {
        printf $fh 
            "%s  %10s  %8s  %8s  %9s  %-24.24s\n",
            " " x 6, 
            $item->{sa_id},
            $item->{sa_antjp},
            $item->{arkn4_count},
            $item->{sa_status},
            $item->{sa_tittel};
           
    }

    print $fh "\n";

}

sub print_sastatus_u {
    my ($self, $list) = @_;
    my $fh = $self->get_report_handle(); 

    $self->increment_subsubsection();
    my $sec = $self->get_section_string(3);
    print $fh "    $sec Saker med saksstatus 'U' \n\n";
    

    if (! @$list) {
        print $fh " " x 6;
        print $fh "Ingen saker hadde SA.STATUS = 'U'.\n\n";
        return 1;
    }
     
    my $plr = (scalar @$list > 1) ? "saker" : "sak";
    print $fh " " x 6;
    print $fh (scalar @$list)." $plr hadde SA.STATUS = 'U':\n\n";
    printf $fh "%s  %12s  %8s  %8s  %9s\n", 
        " " x 6, "Sak ID", "SA.ANTJP", "Funnet", "SA.STATUS";
    printf $fh "%s  %12s  %8s  %8s  %9s\n", 
        " " x 6, "-" x 12, "-" x 8, "-" x 8, "-" x 9;

    for my $item (@$list) {
        printf $fh 
            "%s  %12s  %8s  %8s  %9s\n",
            " " x 6, 
            $item->{sa_id},
            $item->{sa_antjp},
            $item->{arkn4_count},
                $item->{sa_status}; 
    }

    print $fh "\n\n";
    
    return 1;
}


sub print_jpstatus_u {
    my ($self, $list) = @_;

    my $fh = $self->get_report_handle();
    
    $self->increment_subsubsection();
    my $sec = $self->get_section_string(3);
    print $fh "    $sec Journalposter med journalstatus 'U'\n\n";
    if (! @$list) {
        print $fh " " x 6;
        print $fh "Ingen journalposter hadde JP.STATUS = 'U'.\n\n";
        return 1;            
    }

    my $plr = (scalar @$list > 1) ? "er" : "";
    print $fh " " x 6;
    print $fh (scalar @$list)." Journalpost$plr hadde JP.STATUS = 'U':\n\n";
    printf $fh "%s  %12s  %8s\n", 
        " " x 6, "JP.ID", "SA.STATUS";
    printf $fh "%s  %12s  %8s\n", 
        " " x 6, "-" x 12, "-" x 8;

    for my $item (@$list) {
        printf $fh 
            "%s  %12s  %8s\n",
            " " x 6, 
            $item->{jp_id},
            $item->{jp_status}; 
    }

    print $fh "\n\n";
    
}

=item write_freq_analysis ($sa, $jp, $fradato, $tildato)

Generates a frequency analyses of number of noarksak and journpost
entries.

Arguments:

  $sa      - HASHREF with NOARKSAK entries.
  $jp      - HASHREF with JOURNPOST entries.
  $fradato - from date.
  $tildato - to date.

=cut
sub write_freq_analysis {
    my ($self, $args) = @_;

    my $fh = $self->get_report_handle();

    $self->increment_section();
    my $sec = $self->get_section_string(2);
    print $fh SECSEP, "$sec Antall saker, journalposter og dokumenter\n\n";

    $self->increment_subsection();
    my $ssec1 = $self->get_section_string(2);
    print $fh "  $ssec1 NOARKSAK\n\n";
    $self->increment_subsubsection();
    $sec = $self->get_section_string(3);
    print $fh "    $sec Antall saker fordelt p r\n";
    $self->print_histogram($args->{noarksak}, "Saker");
    $self->print_journalpost_count($args->{sakjpost});
    $self->print_sastatus_u($args->{sastatus});

    $self->increment_subsection();
    my $ssec2 = $self->get_section_string(2);
    print $fh "  $ssec2 JOURNPOST\n\n";
    $self->increment_subsubsection();
    $sec = $self->get_section_string(3);

    print $fh "    $sec Antall journalposter fordelt p r\n\n";
    $self->print_histogram($args->{journpost}, "Poster");
    $self->print_jpstatus_u($args->{jpstatus});

    $self->increment_subsection();
    my $ssec3 = $self->get_section_string(2);
    print $fh "  $ssec3 JOURNAL: Antall journalposter fordelt p r:\n\n";
    $self->print_histogram($args->{journal}, "Poster");

    $self->increment_subsection();
    my $ssec4 = $self->get_section_string(2);
    print $fh "  $ssec4 SAKDOKOVERSIKT: Antall saker og journalposter fordelt p r:\n\n";
    $self->print_histogram($args->{sakdok_saker}, "Saker");
    $self->print_histogram($args->{sakdok_poster}, "Poster");
    
    return 1;
}

sub write_mandatory_tables_test {
    my ($self, $args) = @_;

    my $fh = $self->get_report_handle();
    
    $self->increment_section();
    my $sec = $self->get_section_string(2);
    print $fh SECSEP, "$sec Kontroll av obligatoriske tabeller iht. O-krav\n\n";

    $self->increment_subsection();
    $sec = $self->get_section_string(2);
    
    print $fh "    $sec Obligatoriske tabeller som mangler\n\n";

    if (scalar @{$args->{missing}} < 1) {
        print $fh " " x 6;
        print $fh "Alle obligatoriske tabeller var med i avleveringen.\n\n";
    } 
    else {
        print $fh " " x 6;
        print $fh "Flgende obligatoriske tabeller var ikke med i ";
        print $fh "avleveringen:\n\n";

        for my $table (@{$args->{missing}}) {
            print $fh "        ".$table->get_noark_name()."\n";
        }
        print $fh "\n";
    }
        
    $self->increment_subsection();
    $sec = $self->get_section_string(2);

    print $fh "    $sec Ekstra tabeller som er tatt med\n\n";

    if (scalar @{$args->{extra}} < 1) {
        print $fh " " x 6;
        print $fh "Ingen ekstra tabeller var med i avleveringen.\n\n";
    }
    else {
        print $fh " " x 6;
        print $fh "Flgende ikke obligatoriske tabeller var med i avleveringen:\n\n";
        for my $table (@{$args->{extra}}) {
            print $fh "        ".$table->get_noark_name()."\n";
                
        }
    }

    return 1;
}

=item write_doccount_test

Writes number of ducuments reported in DOKVERSJON.XML

=cut
sub write_doccount_test {
    my ($self, %args) = @_;
    
    my $fh = $self->get_report_handle();

    $self->increment_subsection();
    my $sec = $self->get_section_string(2);

    print $fh "  $sec Antall dokumenter i DOKVERSJON\n\n";
    print $fh "    Antall dokumenter...: ", $args{'NUMDOCS'}, "\n\n";
}


=item write_extrafiles_test ($files)

Short test to output the value of extra files.

=cut
sub write_extrafiles_test {
    my ($self, $files) = @_;

    return 1 unless (scalar(@$files));
    
    $self->increment_subsubsection();
    my $sec = $self->get_section_string(3);
    
    my $fh = $self->get_report_handle();
    
    print $fh "    ".$sec." Ekstra filer i avleveringen\n\n";
    print $fh "      Flgende filer i avleveringen er ikke definert i NOARKIH:";
    print $fh "\n";
    for (@{$files}) {
        print $fh " " x 8, "'$_'\n";
    }
    print $fh "\n"

}  


sub write_documents_check {
    my ($self, $an) = @_;

    my $total        = $an->get_documents_mentioned();
    my $seen         = $an->get_documents_seen();
    my $missing      = $an->get_documents_missing();
    my $found        = $an->get_documents_found();
    my $types        = $an->get_document_filetypes();
    my @extra_list   = $an->get_documents_extra_list();
    my @missing_list = $an->get_documents_missing_list();

    my $max = $self->get_max_errors();
    my $fh  = $self->get_report_handle();

    $self->increment_subsection();
    my $sec = $self->get_section_string(2);

    #-------------------------------------------------------------------------
    # Sammendrag:
    print $fh "  $sec  Sammendrag\n\n";

    if (!$total) {
        print $fh "    Ingen dokumenter var nevnt i DOKVERSJON.\n";
    }
    if (!$seen) {
        print $fh "    Ingen dokumenter ble funnet i avleveringen.\n\n";
    }

    print $fh " " x 24, 'Totalt   Funnet   Mangler  Ekstra', "\n";
    print $fh " " x 24, '-------  -------  -------  -------', "\n";
    printf $fh "%23s %7d  %7d  %7d  %7d\n", 'Filer i DOKVERSJON:', 
        $total, $found, $missing, 0;
    printf $fh "%23s %7d  %7d  %7d  %7d\n", 'Filer i avlevering:',
        $seen, $found, 0, ($seen-$found);

    print $fh "\n";

    return if (!$total && !$seen);


    #-------------------------------------------------------------------------
    # Oversikt over filer som mangler
    $self->increment_subsection();
    $sec = $self->get_section_string(2);
    print $fh "  $sec  Oversikt over filer som mangler\n\n";
    print $fh "    Referansen finnes i DOKVERSJON, men fila finnes ikke.\n\n";
    my $j = 0;
    for (@missing_list) {
        last if $j >= $max;
        print $fh '    '.$_."\n";
        $j += 1;
    }
    print $fh "\n";

    #-------------------------------------------------------------------------
    # Oversikt over filer som er lagt ved ekstra
    $self->increment_subsection();
    $sec = $self->get_section_string(2);
    print $fh "  $sec  Oversikt over filer som er lagt ved ekstra\n\n";
    print $fh "    Dokumentfila finnes, men har ingen referanse i DOKVERSJON";
    print $fh "\n\n";

    my $i = 0;
    for (@extra_list) {
        last if $i >= $max;
        print $fh '    '.$_."\n";
        $i += 1;
    }
    print $fh "\n";

    #-------------------------------------------------------------------------
    # Fordeling av filtyper.
    $self->increment_subsection();
    $sec = $self->get_section_string(2);
    
    print $fh "  $sec  Fordeling av filtyper\n\n";

    for (keys %{$types}) {
        printf $fh "%22s: %7d\n", $_, $types->{$_};    
    }
    print $fh "\n";

    return 1;
}

=item write_noarksak_test ($res)

Writes results from the noarksak test.

=cut
sub write_noarksak_test {
    my ($self, %args) = @_;

    my $fh = $self->get_report_handle();
    
    $self->increment_subsection();
    my $sec = $self->get_section_string(2);
    my $nih = $args{'NOARKIH'};
    my $sak = $args{'SAKDOKOVERSIKT'};

    print $fh "  $sec Sammenlikning av antall NOARKSAK i NOARKIH og ";
    print $fh "SAKDOK\n\n";

    if (! $args{'SAKDOKOVERSIKT'}) {
        print $fh "    SAKDOK ble ikke funnet i avleveringen.\n";
        print $fh "    Dette er en obligatorisk rapport som alltid skal ";
        print $fh "vedlegges.\n";
        print $fh "    Jamfr Noark-4.1 K15.18\n\n";

        return undef;
    }
    
    print $fh "    ", "Tabell", " " x 12, "Saker    Poster\n";
    print $fh "    ", "-" x 16,         "  -------  -------\n";

    my $diff = ($sak>$nih) ? $sak-$nih : $nih-$sak;
    for (keys %args) {
        my $dots = "." x (16-length($_));
        my $name = $_.$dots;
        my $nsak = sprintf("%7s", $args{$_}->{'NOARKSAK'}  || "n/a");
        my $post = sprintf("%7s", $args{$_}->{'JOURNPOST'} || "n/a");
        
        print $fh "    ", $name, ": ", $nsak, "  ", $post, "\n";
    }
}


=item write_interval_test (%args)

Arguments: 
  FROM - FIX
  TO - FIX
  EIFROM - FIX
  EITO - FIX

=cut
sub write_interval_test {
    my ($self, %args) = @_;

    my $fh = $self->get_report_handle();

    $self->increment_section();
    my $sec = $self->get_section_string(2);

    my $fra = sprintf("%-4s", ($args{'FROM'}) ? "OK" : "FEIL");
    my $til = sprintf("%-4s", ($args{'TO'}) ? "OK" : "FEIL");

    $args{'EIFROM'}   =~ s/(\d{4})(\d{2})(\d{2})/$1-$2-$3/;
    $args{'EITO'}     =~ s/(\d{4})(\d{2})(\d{2})/$1-$2-$3/;
    $args{'FROMDATE'} =~ s/(\d{4})(\d{2})(\d{2})/$1-$2-$3/;
    $args{'TODATE'}   =~ s/(\d{4})(\d{2})(\d{2})/$1-$2-$3/;
    
    my $fra_sak = ($args{'FROMCTR'} == 1) ? "sak" : "saker";
    my $til_sak = ($args{'TOCTR'} == 1)   ? "sak" : "saker";

    $args{'FROMCTR'} = "Ingen" if ($args{'FROMCTR'} == 0);
    $args{'TOCTR'}   = "Ingen" if ($args{'TOCTR'} == 0);

    print $fh SECSEP;

    print $fh <<"EOF";
$sec Test av periodisering

    Periode.......: $args{'EIFROM'} - $args{'EITO'}
    Nedre grense..: $fra - $args{'FROMCTR'} $fra_sak utenfor periode.
    Frste sak....: $args{'FROMDATE'}
    vre grense...: $til - $args{'TOCTR'} $til_sak utenfor periode.
    Siste sak.....: $args{'TODATE'}

EOF

    return 1;

}


=item write_numbering_test ($miss)

Generates a test of case numbering.

Arguments:

  $miss - HASHREF with number of missing numbers.

=cut
sub write_numbering_test {
    my ($self, $miss) = @_;

    my $fh = $self->get_report_handle();

    $self->increment_section();
    my $sec = $self->get_section_string(2);

    print $fh SECSEP;
    print $fh "$sec Analyse av hull i saksnummerrekkeflge.\n\n";

    print $fh "  r    Antall  Frste ID   Siste ID     ID/sak  Ant.Hull\n";
    print $fh "  ----  ------  ---------  ---------  ---------  --------\n";

    my ($year, $cnt, $fra, $til, $frq, $mng) = undef; 

  format NT =
  @>>>@>>>>>>>@>>>>>>>>>>@>>>>>>>>>>@>>>>>>>>>>@>>>>>>>>>
  $year, $cnt, $fra,      $til,      $frq,      $mng
.

    $fh->format_name('NT');

    for (sort {$a <=> $b} keys %{$miss->{'YEARS'}}) {
        $year = $_;
        $cnt = $miss->{'YEARS'}->{$_}->{'COUNT'};
        $frq = sprintf("%.2f", $miss->{'YEARS'}->{$_}->{'IDFREQ'});
        $mng = scalar(@{$miss->{'YEARS'}->{$_}->{'MISSING'}});
        $fra = $miss->{'YEARS'}->{$_}->{'FIRSTID'};
        $til = $miss->{'YEARS'}->{$_}->{'LASTID'};

        $fh->format_write();
    }  
    print $fh "\n";

    print $fh "  ID Hopp  SEKNR Fra  SEKNR Til\n";
    print $fh "  -------  ---------  ---------\n"; 
    my $i = 0;
    for $year (keys %{$miss->{'YEARS'}}) {
        last if $i > 100;
        my @list = sort { $b->{'GAP'} <=> $a->{'GAP'} } map { 
            { %$_, GAP => ($_->{'LAST'}-$_->{'FIRST'})-1 } 
        } @{$miss->{'YEARS'}->{$year}->{'MISSING'}}; 
        for (@list) {
            last if $_->{'GAP'} < 10; # Skips gaps less than 10.
            last if $i > 100;
            $fra = $_->{'FIRST'};
            $til = $_->{'LAST'};
            my $fx = 9-length($fra);
            my $tx = 9-length($til);

            my $bor = $_->{'GAP'};
            my $bx = 7-length($bor);

            # $fh->format_write();
            print $fh "   " x $bx,$bor, "  "," " x $tx,$til,"  "," " x $fx,$fra,"\n";
            $i += 1;
        }
    }

    print $fh "\n";
    
} 

sub get_max {
    my ($self, $hs) = @_;
    
    my $max = 0;
    for (keys(%{$hs})) {
        $max = $hs->{$_} if $hs->{$_} >= $max;
    }

    return $max;
    
}


=item write_count ($tabellinfo)

Writes information about antall poster found.

=cut
sub write_count {
    my ($self, $info) = @_;

    $self->increment_subsubsection();
    my $sec = $self->get_section_string(3);
    my $fh  = $self->get_report_handle();
    
    print $fh "    ".$sec." Opptelling av antall poster\n\n";
    for (@{$info->get_files_ref}) {
        my $xml = $_->{'XML.ANTPOSTER'} || '<ikke funnet>';
        my $ih  = $_->{'TI.ANTPOSTER'} || '';
        my $fil = $_->{'TI.FILNAVN'};

        print STDERR '$xml is undef' if not defined $xml;
        print STDERR '$ih  is undef' if not defined $ih;
        print STDERR '$fil is undef' if not defined $fil;
        
        print STDERR Dumper($info) if not defined $xml;

        print $fh <<EOF;
      Antall poster oppgitt i NOARKIH: $ih
      Antall poster i $fil: $xml

EOF
        if (! $_->{EXISTS}) {
            print $fh " " x 6;
            print $fh $fil." ble ikke funnet i avleveringen.\n\n";
        }
    }
}

=item write_table ($err, $info, $msg)

Main function for writing error messages.

=cut
sub write_table {
    my ($self, $err, $info) = @_;
    assert(defined $info);

    return 1 if not defined $err;

    my $fh = $self->get_report_handle();

    $self->increment_subsection();
    my $sec = $self->get_section_string(2);

    # Kludge to handle NOARKIH errors aswell as ordinary tabellinfo errors
    if (ref $info eq 'Noark::Parser::Tabellinfo') { 
        print $fh "  $sec ".$info->get_table_name."\n\n";
    } 
    elsif (defined($info->{'NAME'})) {
        print $fh "  $sec ".$info->{'NAME'}."\n\n";
    } 
    else {
        confess("Oops: got unknown info: ".Dumper($info));
    }

    # type_handler er en handler for nye typer kumulative feilmeldinger.
    my %type_handler = (
        'Noark::Error::XMLMissingMandatoryAttribute' 
            => \&write_xml_missing_mandatory_attribute_error,
        'Noark::Error'                                    
            => \&write_common_error,
        'HASH'                                            
            => \&write_common_error,
    );
    for my $e (@{$err}) {
        
        if ($type_handler{ref($e)}) {
            $type_handler{ref($e)}($self, $e, $info);
            next;
        }
        
        if (grep {ref $e eq $_} qw|
            Noark::Error::BrokenDocumentReference
            Noark::Error::CursiveAttribute
            Noark::Error::DuplicateKey
            Noark::Error::DuplicateUnique
            Noark::Error::EmptyAttribute
            Noark::Error::ExtraAttribute
            Noark::Error::ForeignKey
            Noark::Error::MissingAttribute
            Noark::Error::MissingFile
            Noark::Error::MissingFileFound
            Noark::Error::Null
            Noark::Error::PotentialDuplicateKey
            Noark::Error::PrimaryKey
            Noark::Error::UnknownAttribute
            Noark::Error::UnknownTable
            Noark::Error::ValidData
            Noark::Error::WrongFilename
        |) {

            $self->increment_subsubsection();
            my $sec = $self->get_section_string(3);

            print $fh $e->to_text($sec, 4);
            print $fh "\n";
            next;
        }
        
    }

    return 1;
} 

#------------------------------------------------------------------------------
# Replacement for FORMATS. Simple formatting of text strings.
# Does not return a newline at the end.
sub __format_text {
    my ($self, $text, $indent) = @_;

    $indent ||= 0;
    my $pre  = " " x $indent;

    my @words = split / /, $text;
    my $done  = "";
    my $line  = "";

    while (my $word = shift @words) {
        my $return = undef;
        $line .= $pre if length($line) == 0;
        if ($word =~ /(\S*)(\n+)(\S*)/) {
            # word with newline in it
            $word = $1 || "";
            $return = $2;
            unshift(@words, $3) if defined $3;
            $word =~ s/\s*//g;
        }
        if (length($line.' '.$word) > 78) {
            $done .= $line."\n";
            $line = '';
            redo;
        }

        $line .= $word." ";
        if (defined $return) {
            $done .= $line.$return;
            $line = '';
        }
    }

    $done .= $line;

    return $done;
}


sub write_common_error {
    my ($self, $error, $info) = @_;

    if ($error->{mesg}) {
        confess("Ooops. Got call to write_message_error");
    }

    $self->increment_subsubsection();

    my $fh  = $self->get_report_handle();
    my $sec = $self->get_section_string(3); 

    if ($error->{name} =~ /^(OK|FAIL)$/) {
        print $fh "    ".$sec." [".$error->{'name'}."] ".$error->{'text'}."\n\n";
        return;
    } 

    if (ref($error) eq 'Noark::Error') {
        my $max = $self->{MAX_ERRORS};
        if ($error->get_error_count() < $self->{MAX_ERRORS}) {
            $max = $error->get_error_count();
        }
                   
        $error->add_variable(ERRORS_TOTAL      => $error->get_error_count(),
                             REPORT_MAX_ERRORS => $max);
    }

    my $file = $error->{'file'} || "ikke angitt";
    my $text = (ref($error) eq 'HASH') ? $error->{'text'} : $error->get_text();

    if ((ref($error) eq 'HASH') && (defined $error->{KEYS})) {
	$text .= $error->{KEYS};
    }

    my $name = (ref($error) eq 'HASH') ? $error->{'name'} : $error->get_name();
    $name = $sec." ".$name;

    format IH =
    @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<...
    $name

      ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      $text
      ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<~~
      $text

.

    $fh->format_name('IH');
    $fh->format_write();

    return 1 if ref($error) ne 'Noark::Error';

    $self->_write_error_listing($error);

    return 1;
}


#-----------------------------------------------------------------------------
# Printing out error listing:
#-----------------------------------------------------------------------------
sub _write_error_listing {
    my ($self, $error) = @_;
    
    assert(ref($error) eq 'Noark::Error');

    my @list = @{$error->get_listing()};
    return 1 if ! scalar @list;

    my $fh     = $self->get_report_handle();
    my $indent = ' ' x 8;

    my @labels  = @{$error->get_list_labels()};
    my @lengths = @{$error->get_list_label_lengths()};

    my $head = $indent; # Starting with our indent level.
    $head .= join('  ', map { "%-".$_."s" } @lengths );
    
    $fh->printf($head."\n", @labels);

    my $str = $indent;
    for my $length (@lengths) {
        $str .= '-' x $length .'  ';
    }
    $str =~ s/\s+$//;
    $str .= "\n";

    $fh->print($str);

    $str = '';
    my $error_count = 0;
    for my $entry (@list) {
        last if $error_count >= $self->{'MAX_ERRORS'};
        $error_count += 1;

        $str  = $indent;
        $str .= join('  ', map { "%-".$_."s"} @lengths);

        $fh->printf($str."\n", @{$entry});
    }
    print $fh "\n" x 2;
    return 1;    
} 


sub write_xml_missing_mandatory_attribute_error {
    my ($self, $error, $info) = @_;
    
    $self->increment_subsubsection();
    
    my $sec = $self->get_section_string(3);
    my $fh  = $self->get_report_handle();
    
    printf $fh "    %s %s\n\n", $sec, $error->get_name;
    printf $fh $self->__format_text($error->get_text(),6)."\n\n";
    print $fh " " x 8, "Attributt                   Antall\n";
    print $fh " " x 8, "-" x 24, "  ", "-" x 8, "\n";

    my $list = {};
    for my $a (@{$error->get_attributes}) {
        my ($key, $value) = each(%{$a});
        $list->{$key} += 1;                    
    }

    for my $k (keys %$list) {
        printf $fh "        %-24s  %8d\n", $k, $list->{$k};        
    }

    printf $fh "\n";
    return 1;
}


=item write_header ($version, $dir)

Writes the report title

=cut
sub write_header {
    my $self    = shift;
    my $version = shift;
    my $dir     = shift;

    my $fh = $self->get_report_handle();

    $self->increment_subsection();

    my $sec = $self->get_section_string(2);
    
    my $date = localtime();
    my $vers = "importer ".$version;
    format HEADER=

                      TESTRAPPORT FOR ARKIVUTTREKK
                Kontroll av uttrekk fra Noark-4.1 system

@<<<< OM RAPPORTEN
$sec

  Dato..............: @<<<<<<<<<<<<<<<<<<<<<<<<<<<< 
        $date
  Versjon...........: @<<<<<<<<<<<<<<<
        $vers,
  Importkatalog.....: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        $dir

.
    
    
    $fh->format_name('HEADER');
    $fh->format_write();
    
    return 1;
}

sub DESTROY {
    my $self = shift;
}
 
1; 
