#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: SQLDep.pm 493 2007-03-01 10:19:44Z mtr $
#---------------------------------------------------------------------
# Noark::SQLDep - Parses SQL to build an dependency tree.
#
# Author: Thomas Malt <tm@linpro.no>
#
# Copyright (c) 2004-2005  Riksarkivaren
#
#---------------------------------------------------------------------

use strict;
use warnings;

package Noark::SQLDep;

=head1 NAME

Noark::SQLDep - Parses the SQL-files to build a dependency tree.

=head1 SYNOPSIS

  my $sd = Noark::SQLDep->new(dir     => <dir>, 
                              verbose => 0,
                              debug   => 0);

=head1 DESCRIPTION

bar

=cut

use Carp;

my ($VERSION) = '$Revision: 493 $' =~ /: (\S+)/;
my %SEEN = ();


=head1 SUBROUTINES

=over

=item new ()

Creates a new SQLDep object and returns it as its reference.

=cut
sub new {
    my $proto = shift;
    my %args  = @_;

    my $class = ref($proto) || $proto;
    my $self  = {};

    $self->{VERB}    = $args{verbose} || 0;
    $self->{DEBUG}   = $args{debug}   || 0;
    $self->{DIR}     = $args{dir}     || "";    
    $self->{VERSION} = $VERSION;
    $self->{TREE}    = {};
    $self->{DBMAP}   = {};

    bless($self, $class);

    $self->_build_tree();

    return $self;
}



=item get_priority_list

Returns a list of tables sorting according to the amount of other
tables they reference to.

  return ARRAYREF

=cut
sub get_priority_list {
    my ($self) = @_;

    my @list = ();
    for (keys %{$self->{TREE}}) {
        %SEEN = ();
        push (@list, {NAME => $_, COUNT => $self->get_priority_count($_)});
    }

    @list = sort {$a->{COUNT} <=> $b->{COUNT}} @list;
    return \@list;

    # my @res = ();
    # for (@list) {
    #    push (@res, $_->{NAME});
    # }
    # return \@res;
}



=item get_priority_count ($tab, $max)

Returns the priority count of a given table.

  param STRING  $tab -- Table name.
  param INTEGER $max -- Maximum value when the table is already 
                        seen in this scope.

=cut
sub get_priority_count {
    my ($self, $tab, $max) = @_;
    $max = $max || 500;

    return $max if ($SEEN{$tab});
    $SEEN{$tab} = 1;

    return undef unless $self->{TREE}->{$tab};

    my $list = $self->{TREE}->{$tab};
    my $cnt = 0;
    for (keys %{$list}) {
        $cnt += 1;
        $cnt += $self->get_priority_count($list->{$_});
        # last if $cnt > $max;
    }
    return $cnt;
}



=item get_references_by_table ($tab)

Returns a tree of HASHES with the references a given table depends on.

  param  STRING  $tab -- Table name (i.e NOARK Doctype.)
  return HASHREF      -- Tree of HASHES

=cut
sub get_references_by_table {
    my ($self, $tab) = @_;

    $tab = $tab.'.TAB' unless ($tab =~ /.TAB$/);
    
    %SEEN = ();

    my $tree = $self->_get_ref_by_table($tab, 0, \%SEEN);

    %SEEN = ();
    return $tree;
}


sub _get_ref_by_table {
    my ($self, $tab, $ctr, $seen) = @_;

    confess("Got counter > 99 deep: $!") if ($ctr > 99);
    
    $ctr += 1;
    if ($seen->{$tab}) {
        if ($self->{VERB}) {
            print " " x ($ctr*2);
            print "$ctr|".$seen->{$tab}.": $tab : CIRCULAR", "\n";
        }
        return { DEPTH => 1, TREE => undef};
    }


    $seen->{$tab} = $ctr;
    if ($self->{VERB}) {
        print " " x ($ctr*2);
        print "$ctr: $tab:\n";
    }

    my $list = $self->{TREE}->{$tab};
    my $res = {DEPTH => $ctr};
    for (keys(%{$list})) {
        my $r = $self->_get_ref_by_table($list->{$_}, $ctr, $seen);
        $res->{TREE}->{$list->{$_}} = $r;
        $res->{DEPTH} += $r->{DEPTH} if ref($r) =~ /HASH/;
    }

    return $res;
}

sub _build_tree {
    my ($self) = shift;

    opendir(D, $self->{DIR}) || confess("Oops: $!");
    for (readdir(D)) {
        next unless /^modul.*tabeller.sql$/;
        $self->_parse_tables($self->{DIR}.'/'.$_);
    }
    rewinddir(D);
    for (readdir(D)) {
        next unless /^modul.*referanser.sql$/;
        $self->_parse_references($self->{DIR}.'/'.$_);
    }
    closedir(D);
}


sub _parse_tables {
    my ($self, $file) = @_;

    open(F, $file) or confess("Oops: '$file': $!");
    my $got = 0;
    my $len = 30;
    while (<F>) {
        if (/create\stable\s+(\w+)\(/) { # +;
            my $TAB = uc($1).'.TAB';
            $self->{DBMAP}->{$1} = $TAB;
            $self->{TREE}->{$TAB}  = {};
        }
        next;  
    }
    close(F); 
}


sub _parse_references {
    my ($self, $file) = @_;

    open(F, $file) || confess("Oops: '$file': $!");
    
    while (<F>) {
        next unless /^alter\stable\s+(\w+)\s+add\sforeign\skey\s?\((\w+)\)\s+references\s(\w+)\s?\((\w+)\);/; # ?;  
        next unless (defined($1) && defined($2) &&  
                     defined($3) && defined($4)); 
        print "$1($2) => $3($4)\n" if ($self->{VERB} > 1); 
        $self->{TREE}->{$self->{DBMAP}->{$1}}->{$2} = $self->{DBMAP}->{$3}; 
        
    }

    close(F); 
}


=pod

=back

=head1 AUTHOR

Thomas Malt <tm@linpro.no>

=head1 COPYRIGHT

Copyright (c) 2004-2005  Riksarkivaren

=head1 LICENCE

Whatever Riksarkivaren chooses to do.

=cut

1;

__END__

#---------------------------------------------------------------------
# $Log$
# Revision 1.1  2005/09/29 13:55:14  tm
# Added all files to repository
#
# Revision 1.3  2004/12/22 09:36:20  elark
# Siste endringer.
#
#
