#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Schema.pm 781 2011-06-23 14:45:21Z tm $
#---------------------------------------------------------------------
# Noark::Schema - Contains the database schema for the Noark archive.
#
# Author: Thomas Malt <tm@linpro.no>
# Author: Kjell-Magne ierud <kjellm@linpr.no>
#
#-----------------------------------------------------------------------------
package Noark::Schema;

use strict;
use warnings;

use Carp;
use Carp::Assert;
use Data::Dumper;
use File::Basename;

use Noark::Config qw($tables); ## Remove when ready
use Noark::Schema::Table;
use Noark::Schema::Map;

my $debug = 1;
my $self = undef;

=head1 NAME

 Noark::Schema

=head1 SYNOPSIS

my $obj = Noark::Schema->new(dir => <dir>,
                             verbosity => 0,
                             use_xml_names => 1,
                             );

=head1 DESCRIPTION

Contains the database schema for the Noark archive.

=cut



=head1 CONSTRUCTOR

=over

=item new(%args)

Creates a new Schema object and returns it as its reference.
Noark::Schema is a singleton.

=cut
sub new {
    my $proto = shift;
    my %args  = @_;

    # Enforces the singelton property
    return $self if defined $self;

    my $class = ref($proto) || $proto;
    $self  = {};

    $self->{SQLDIR}      = dirname($INC{'Noark/Schema.pm'}).'/SQL';
    $self->{DEBUG}       = $args{debug}       || 0;
    $self->{VERBOSE}     = $args{verbosity}   || 0;
    $self->{EKSPORTINFO} = $args{eksportinfo} || undef;

    $self->{RETURN_XML_NAMES} 
        = defined $args{use_xml_names} ? $args{use_xml_names} : 1;

    bless($self, $class);

    $self->{TABLES}  = {};

    croak("Unable to find SQL-dir: ".$self->{SQLDIR}) if !(-d $self->{SQLDIR});

    $self->parse($self->{SQLDIR});
    $self->parse_ref($self->{SQLDIR});

    return $self;
}


=back

=head1 SUBROUTINES

=over

=item set_return_xml($val)

Sets return_xml to $val.

Returns the old value.

=cut
sub set_return_xml {
    my ($self, $val) = @_;

    my $ret = $self->{RETURN_XML_NAMES};
    $self->{RETURN_XML_NAMES} = $val;

    return $ret;
}

=item parse($dir)

Parses the SQL table definitions found in directory `$dir'.

Dies on parse error.

=cut
sub parse {
    my ($self, $dir) = @_;

    my @files = <$dir/*tabeller*.sql>;

    croak("No SQL found in dir `$dir'") unless (@files);

    for my $file (@files) {
        open my $fh, "< $file" or die($!);
        print "Parsing $file\n" if $self->{DEBUG};

        while (<$fh>) {
            print "S: ", $_ if $self->{VERBOSE} == 4;
            /drop/i or next;

            my $tab = Noark::Schema::Table->new(verbosity => $self->{VERBOSE});
            eval {
                my $name = $tab->parse($fh);
                $self->{TABLES}->{$name} = $tab;
            };
            if ($@) {
                die ("In file `$file'\n" . $@);
            }
        }
        close $fh or die ($!);
    }

    return 1;
}


=item parse_ref($dir)

Parses references/foreign keys for a module and adds them to the data model.

=cut
sub parse_ref {
    my ($self, $dir) = @_;

    my @files = <$dir/*referanser.sql>;

    croak("No SQL found in dir `$dir'") unless (@files);

    my $regex = qr{^\s*alter\s+table\s*(\w+)
                   \s+add\s+foreign\s+key\s*\((\w+)\)
                   \s+references\s+(\w+)\s*\((\w+)\)
                  }xi;

    for my $file (@files) {
        open my $fh, "< $file" or die($!);
        print "Parsing $file\n" if $self->{DEBUG};

        for ($_ = readline($fh); defined; $_ = readline($fh)) {
            print "S: ", $_ if $self->{VERBOSE} == 4;
            my ($tab, $attr, $ref_tab, $ref_attr) = $_ =~ $regex or next;

            unless ($self->has_table($tab)) {
                die ("Parse error: Undeclared tablename `$tab'!\n"
                     . "On line $. in file `$file'.");
            }
            unless ($self->has_table($ref_tab)) {
                die ("Parse error: Undeclared tablename `$ref_tab'!\n"
                     . "On line $. in file `$file'.");
            }

            my $t = $self->get_table($tab);
            my $rt = $self->get_table($ref_tab);

            eval {
                $t->add_foreign_key($attr, $rt, $ref_attr);
            };
            if ($@) {
                die ("In file `$file'\n" . $@);
            }
            eval {
                $rt->add_reference($ref_attr, $t, $attr);
            };
            if ($@) {
                die ("In file '$file'\n".$@);
            }
        }
    }

    return 1;
}


=item get_table($name)

Returns the Table object identified by `$name'.
Will die hard if invalid table name is given.

=cut
sub get_table {
    my ($self, $table) = @_;

    croak("You must provide a table name as argument") if !$table;

    if ($table eq uc($table) || $table =~ /\.TAB$/) {
        my $dbmap = Noark::Schema::Map->new();
        eval {
            $table  = $dbmap->get_schema_table_name($table);
        };
        return undef if $@;
    }

    assert($self->{TABLES}->{$table});
    
    return $self->{TABLES}->{$table};

}

=item is_tablename_valid ($table)

Checks if a given table name exists in noark4
accepts string like <TABLE>.TAB, <TABLE> and <table>

=cut
sub is_tablename_valid {
    my ($self, $table) = @_;
    return $self->get_table($table) ? 1 : 0
}

=item is_attributename_valid ($table, $attribute)

Checks if a given attribute exists in a table.

=cut
sub is_attributename_valid {
    my ($self, $table, $attribute) = @_;
    croak "Table ($table) must be a valid table" if not $self->is_tablename_valid($table);
    
    return $self->get_attribute($table, $attribute) ? 1 : 0;
}



=item get_tables()

Returns all table objects as an HASHREF;

=cut
sub get_tables {
    my ($self) = @_;
    return $self->{TABLES};
}

=item get_table_list

Returns all table objects as a list (arrayref)

=cut
sub get_table_list {
    my $self = shift;
    
    my @list = values %{$self->{TABLES}};

    return \@list;
}


=item get_attribute($tab, $attr)

Returns the Attribute object with name `$attr' as found in the Table with
name `$tab`.

=cut
sub get_attribute {
    my ($self, $tab, $attr) = @_;

    my $t = $self->get_table($tab);
    return undef if not defined $t;
    
    # unless ($t->has_attribute($attr)) {
    #     my $map = Noark::Schema::Map->new();
    #     eval {
    #         $attr = $map->get_schema_field_name($attr);
    #     };
    #     return undef if $@;
    # } 

    return $t->get_attribute($attr);
}


=item get_mandatory_tables

Returns a list with mandatory Table objects

=cut
sub get_mandatory_tables {
    my ($self) = @_;
    my @mt = ();

    for (values %{$self->{TABLES}}) {
        push(@mt, $_) if $_->is_mandatory();
    }

    return @mt;
}

=item get_pk_names($tab)

Returns a list with the names of the attributes in the primary key in
Table `$tab'.

=cut
sub get_pk_names {
    my ($self, $tab) = @_;

    my $t = $self->get_table($tab);

    return $self->_make_retval($t->get_pk_names());
}


=item get_pk_names_as_string($table)

Returns a string representation of the primary key(s) for a table

=cut
sub get_pk_names_as_string {
    my ($self, $table) = @_;
    
    croak("You must give a table name as argument.") if (!$table);

    my $t = $self->get_table($table);
    
    return join(', ', @{$self->_make_retval($t->get_pk_names())});    
}

=item is_attr_not_null($tab, $attr)

Returns FALSE if the value of the attribute can be NULL.

=cut
sub is_attr_not_null {
    my ($self, $tab, $attr) = @_;

    my $rec = $self->get_attribute($tab, $attr);

    return not $rec->get_nullness();
}


=item is_attr_null ($tab, $attr)

Returns TRUE if the value of an attribute can be NULL, FALSE otherwise.

=cut
sub is_attr_null {
    my ($self, $tab, $attr) = @_;

    my $rec = $self->get_attribute($tab, $attr);
    return $rec->get_nullness();
}

=item is_attr_auto_increment ($tab, $attr)

Returns whether auto_increment is sat for the attribute.

=cut
sub is_attr_auto_increment {
    my ($self, $tab, $attr) = @_;

    my $rec = $self->get_attribute($tab, $attr);
    return $rec->is_auto_increment();
}


=item get_not_null_attrs($tab)

FOO

=cut
sub get_not_null_attrs {
    my ($self, $tab) = @_;

    my $t = $self->get_table($tab);

    return $self->_make_retval($t->get_not_null_attrs());
}


=item get_all_attrs($tab)

BAR

=cut
sub get_all_attrs {
    my ($self, $tab) = @_;

    my $t = $self->get_table($tab);

    return $self->_make_retval($t->get_all_attrs());
}


=item get_attr_data_type($tab, $attr)

Returns the attribute's data type.

=cut
sub get_attr_data_type {
    my ($self, $tab, $attr) = @_;

    my $rec = $self->get_attribute($tab, $attr);

    return $rec->get_data_type();
}


sub get_all_data_types {
    my $self = shift;

    my %types = ();
  TABLE:
    for my $table (@{$self->get_table_list()}) {
        
      ATTRIBUTE:
        for my $attr (@{$table->get_attribute_list()}) {
            my $type = $attr->get_data_type();
            if ($type =~ m/(\w+)\((\d+)\)/) {
                $type = $1;
            }
            next ATTRIBUTE if $types{$type};

            $types{$type} = 1;
        }
    }

    my @list = keys %types;
    return \@list;
}

=item is_attr_data_valid ($tab, $attr, $tata)

Checks that a given attribute is valid for a given table.

Currently the following checks are implementet:
  VARCHAR(len) - Checks that data is within length of varchar.
  BOOL         - Checks that data is 0 - FALSE or 1 -TRUE. 

=cut
sub is_attr_data_valid {
    my ($self, $tab, $attr, $data) = @_;
    if (!defined($data) || $data eq '') {
        return 1 if $self->is_attr_null($tab, $attr);
        return 0;
    }

    my $sepp = "-" x 78;
    my $type = $self->get_attr_data_type($tab, $attr);
    my $size = undef;
    if ($type =~ m/(\w+)\((\d+)\)/) {
        $type = $1;
        $size = $2;
    }

    my %handlers = (VARCHAR => \&is_varchar_valid,
                    BOOL    => \&is_bool_valid,
                    DATE    => \&is_date_valid,
                    INT     => \&is_int_valid,
                    TEXT    => sub { return 1; },
                    BLOB    => sub { return 1; },
                    TIME    => \&is_time_valid,
                   );
    

    confess( "\n$sepp\n  TODO: missing handler for data type: '$type', value: "
           . "'$data'\n$sepp\n") if ! $handlers{$type};

    return $handlers{$type}($self, $data, $size);
}

=item is_bol_valid ($data)

Tests if the value is acceptabel boolean data (0 or 1)

=cut
sub is_bool_valid {
    my ($self, $data) = @_;

    return 1 if $data =~ /^[01]$/;
    return 0;
}


=item is date valid ($date)

Sjekker hvorvidt en dato er gyldig ihht. noark4.
Merk at datoer kan vre fram i tid, slik at det i seg selv ikke er
diskvalifiserende her. Hvis vi skal legge inn sjekk av hvilke datoer som
skal vre fram i tid etc. m det legges et annet sted.

=cut
sub is_date_valid {
    my ($self, $date) = @_;

    my ($year, $month, $day) = $date =~ m/^(\d{4})(\d{2})(\d{2})/;

    # my ($r_year, $r_month, $r_day) = $self->__get_real_date();

    # return 0 if ($year > $r_year);
    return 0 if ($month < 1 || $month > 12);
    # return 0 if ($year == $r_year && $month > $r_month);
    return 0 if ($day < 1 || $day > 31);    # Not caring about odd months
    # return 0 if ($year == $r_year && $month == $r_month && $day > $r_day);
    return 0 if ($month == 2 && $day > 29); # Not caring about leap years
    return 0 if (!($date =~ m/^\d{8}$/));
    return 1;
} 


=item is_time_valid ($time)

Checks if time is valid 

=cut
sub is_time_valid {
    my ($self, $time) = @_;

    my ($hour, $min) = $time =~ m/(\d\d):?(\d\d)/;
    return 0 if (!defined $hour || !defined $min);
    return 0 if ($hour < 0  || $hour > 23);
    return 0 if ($min  < 0  || $min  > 59);

    return 1;
    
}

=item is_varchar_valid ($string, $length)

Checks if varchar is valid

=cut
sub is_varchar_valid {
    my ($self, $string, $length) = @_;

    return 0 if (length($string) > $length);
    return 1;
}

=item is_int_valid ($int, $length)

Checks if an int is valid

=cut
sub is_int_valid {
    my ($self, $int, $length) = @_;

    return 0 if (!($int =~ m/^\d+$/));
    return 0 if (length($int) > $length);

    return 1;
}

=item get_attr_requirement($tab, $attr)

Returns the attribute's requirement type.

=cut
sub get_attr_requirement {
    my ($self, $tab, $attr) = @_;

    my $rec = $self->get_attribute($tab, $attr);

    return $rec->get_requirement();
}


=item is_attr_mandatory($tab, $attr)

Returns TRUE if attribute is mandatory.

=cut
sub is_attr_mandatory {
    my ($self, $tab, $attr) = @_;

    my $rec = $self->get_attribute($tab, $attr);

    return $rec->get_attribute($tab, $attr)->is_mandatory();
}


=item is_attr_unique($table, $attribute)

Returns TRUE if the attribute has to be unique.

=cut
sub is_attr_unique {
    my ($self, $table, $attribute) = @_;

    my $attribute_object = $self->get_attribute($table, $attribute);

    return $attribute_object->is_unique();
}

=item is_attr_exportable($tab, $attr)

Returns FALSE if the attribute shouldn't be exported.

=cut
sub is_attr_exportable {
    my ($self, $tab, $attr) = @_;

    my $rec = undef;
    
    eval {
        $rec = $self->get_attribute($tab, $attr);
    };
    return undef if ! defined($rec);
    return $rec->in_export();
}



=item get_attr_exportable_list ()

Returns a list of all exportable attributes in a table.

=cut
sub get_attr_exportable_list {
    my ($self, $table_name) = @_;

    my $table = $self->get_table($table_name);    
    return $table->get_exportable_attribute_list();
}


=item dump_me

Dumps the database schema to stdout.

=cut
sub dump_me {
    my ($self) = @_;

    for my $table (keys(%{$self->{TABLES}})) {
        $self->{TABLES}->{$table}->dump_me();
        print "\n";
    }
}


=item has_table

FOO BAR

=cut
sub has_table {
    my ($self, $tab) = @_;

    return undef unless defined $tab;

    return defined $self->{TABLES}->{$tab};
}

sub _make_retval {
    my ($self, $val) = @_;

    return undef unless defined $val;

    if (ref($val) eq 'ARRAY') {
        my @ret = ();
        for (@$val) {
            push @ret, $self->_make_retval($_);
        }
        return \@ret;
    } elsif (ref($val) eq 'HASH') {
        my %ret = ();
        while (my ($k, $v) = each %$val) {
            $ret{$self->_make_retval($k)} = $v;
        }
        return \%ret;
    } else {
        my $map = Noark::Schema::Map->new();
        return $map->get_noark_field_name($val) if $self->{RETURN_XML_NAMES};
        return $map->get_schema_field_name($val);
    }
}


=pod

=back

=head1 SEE ALSO

Noark::Schema::Table, Noark::Schema::Attribute

=head1 AUTHOR

Kjell-Magne ierud <kjellm@linpro.no>

=head1 COPYRIGHT

Copyright (c) 2004-2005  Riksarkivaren

=head1 LICENCE

Whatever Riksarkivaren chooses to do.

=cut

1;
