#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Attribute.pm 797 2011-06-28 13:41:28Z tm $
#---------------------------------------------------------------------
use warnings;
use strict;

package Noark::Schema::Attribute;

# use Noark::Config qw($tables);
use Noark::Schema::Map;



=head1 NAME

Noark::Schema::Attribute

=head1 SYNOPSIS

The definition of an attribute in arkn4 according to Noark-4.1

=head1 DESCRIPTION

Keeps information about an Attribute in the NOARK-4 datamodel.

=cut

my $parse_rx = qr{^\s*(\w+)                      # Name
                  \s+([[:alpha:]]+(?:\(\d+\))?)  # Data type
                  \s+(?:default\s['"]?([\w\s\d]+)['"]?)? # Default value
                  \s+((?:not\s+)?null)           # Nullness
                  \s*((?:unique)?)                # Uniqueness
                  \s*(?:primary\s+key)?          # PK (ignored)
                  \s*((?:auto_increment)?)\s*,?
                  \s+--\s*([/OASEU12]+)/?        # Requirement type
                  \s*$
                 }xi;
# };

# Sometimes parts of Composite Primary Keys can be null in Noark-4.1, but needs 
# to be set for the database constraints to add up.
my %default_values = (
    nu_aar => 0,
);

=head1 SUBROUTINES

=over

=item new(verbosity => 0)

Instantiates a new Attribute object.

=cut
sub new {
    my $class = shift;
    my %args  = @_;
    my $self  = {};

    $self->{VERBOSE} = $args{verbosity} || 0;

    $self->{DATA_TYPE}        = undef;
    $self->{NAME}             = undef;
    $self->{NULLNESS}         = undef;
    $self->{UNIQUE}           = undef;
    $self->{AUTO_INCREMENT}   = undef;
    $self->{REQUIREMENT_TYPE} = undef;
    $self->{EXPORT}           = undef;
    $self->{EXPORTED}         = undef;
    $self->{DEFAULT}          = undef;
    $self->{PRIMARY_KEY}      = undef;
    $self->{REFERENCE}        = undef;
    bless ($self, $class);

    return $self;
}


=item get_name

Returns the attribute's name.

=cut
sub get_name {
    my ($self) = @_;

    return $self->{NAME};
}


=item get_noark_name

Returns the attribute name according to the NOARK spesification.

=cut
sub get_noark_name {
    my ($self) = @_;

    my $map = Noark::Schema::Map->new();
    return $map->get_noark_field_name($self->{NAME});
    
}

=item get_data_type

Returns the attribute's data type.

=cut
sub get_data_type {
    my ($self) = @_;

    return $self->{DATA_TYPE};
}


=item get_nullness

Returns the attribute's nullness. In other words: Returns TRUE if the
attribute's value can be NULL.

=cut
sub get_nullness {
    my ($self) = @_;

    return $self->{NULLNESS};
}

=item can_be_null

Returns true if the attribute can be null.

=cut
sub can_be_null {
    my ($self) = @_;

    return $self->{NULLNESS};
}


=item is_unique

Return whether the attribute has to be unique or not.
Returns TRUE if the attribute must be unique.

=cut
sub is_unique {
    my ($self) = @_;
    return $self->{UNIQUE};
}


=item is_auto_increment

Return whether auto_increment is set for the attribute or not.

=cut
sub is_auto_increment {
    my ($self) = @_;
    return $self->{AUTO_INCREMENT};
}

=item get_requirement

Returns the attribute's requirement type.

=cut
sub get_requirement {
    my ($self) = @_;

    return $self->{REQUIREMENT_TYPE};
}


=item is_mandatory

Returns TRUE if attribute is mandatory.

A very simple test. We only check if the attribute is of type 'O'. In
the future we might add attributes of other types as mandatory.

=cut
sub is_mandatory {
    my ($self) = @_;
    return ($self->{REQUIREMENT_TYPE} eq 'O') ? 1 : 0;
}


=item in_export

Returns FALSE if the attribute shouldn't be exported.

=cut
sub in_export {
    my ($self) = @_;

    return $self->{EXPORTED};
}

=item is_exportable

Returns TRUE if attribute is/should be exported.

=cut
sub is_exportable {
    my ($self) = @_;

    return $self->{EXPORTED};
}



=item set_primary_key_status($status)

Sets the attribute as a primary key or not.

=cut
sub set_primary_key_status {
    my ($self, $status) = @_;
    
    $self->{PRIMARY_KEY} = ($status) ? 1 : 0;
}


=item is_primary_key ()

Returns true if attribute is, or is part of,  primary key.
Returns false otherwise.

=cut 
sub is_primary_key {
    my $self = shift;
    
    return $self->{PRIMARY_KEY};
}

=item set_fk_reference 

Adds a fk_reference to the attribute

=cut
sub set_fk_reference {
    my ($self, $tab, $attr) = @_;
    
    $self->{REFERENCE} = [$tab, $attr];
}

=item get_fk_reference 

Fetches the foreign key reference if the attribute has one.
Returns ARRAYREF [referenced_table, referenced_attribute], or undef.

=cut
sub get_fk_reference {
    my $self = shift;
    return $self->{REFERENCE};
} 

=item get_default_value

Reads the %default_values hash and returns the default value for this 
attribute or undef if none is set.

=cut
sub get_default_value {
    my $self = shift;
    
    if (defined($default_values{$self->get_name()})) {
        return $default_values{$self->get_name()};        
    } else {
        return undef;
    }
}

=item parse($line)

Parses the sql line and fetches the content. A side effect of this
function is that it massages information about the variable and
inserts that into the variable object.

Returns the name of the variable as a string

=cut
sub parse {
    my ($self, $line) = @_;

    # Fetching variables from line:
    ($self->{NAME},
     $self->{DATA_TYPE},
     $self->{DEFAULT},
     $self->{NULLNESS},
     $self->{UNIQUE},
     $self->{AUTO_INCREMENT},
     $self->{REQUIREMENT_TYPE}
    ) = $line =~ $parse_rx or die("Parse error on line $.: $line\n");

    # Massaging variables:
    $self->{DATA_TYPE} = uc($self->{DATA_TYPE});
    $self->{NULLNESS}  = (index(lc($self->{NULLNESS}), 'not') == -1);
    $self->{UNIQUE}    = ($self->{UNIQUE}) ? 1 : 0;
    $self->{AUTO_INCREMENT} = ($self->{AUTO_INCREMENT}) ? 1 : 0;

    if (index($self->{REQUIREMENT_TYPE}, '/') >= 0) {
        $self->{EXPORT}    = 1;
        $self->{REQUIREMENT_TYPE} = uc(substr($self->{REQUIREMENT_TYPE},1,-1));
    } else {
        $self->{EXPORT}    = 0;
        $self->{REQUIREMENT_TYPE} = uc($self->{REQUIREMENT_TYPE});
    }

    $self->{EXPORTED} = $self->{EXPORT} ? 0 : 1;
    #returns the name.
    return $self->{NAME};
}


=item dump_me

bar

=cut
sub dump_me {
    my ($self) = @_;
    my $req = $self->{REQUIREMENT_TYPE};
    my $def = ($self->{DEFAULT})        ? "DEFAULT ".$self->{DEFAULT} : "";
    my $man = ($self->{NULLNESS})       ? "NULL" : "NOT NULL";
    my $uni = ($self->{UNIQUE})         ? "UNIQUE" : "";
    my $aut = ($self->{AUTO_INCREMENT}) ? "AUTO_INCREMENT" : "";

    if ($self->{EXPORT}) {
        $req = "/$req/";
    }

    printf("% -15s", $self->{NAME});
    printf("% -15s", uc($self->{DATA_TYPE}));
    printf("% -15s", $def);
    printf("% -10s", $man);
    printf("% -8s", $uni);
    printf("% -16s", $aut);
    print ",  -- $req";
    # printf "% -15s% -15s% -5s% -10s% -15s% -10s-- %s",
    #   ($self->{NAME}, uc($self->{DATA_TYPE}), $def, $man,$uni,$aut.',', $req);
}

=pod

=back

=head1 SEE ALSO

Noark::Schema, Noark::Schema::Table

=head1 AUTHOR

Kjell-Magne ierud <kjellm@linpro.no>

=head1 COPYRIGHT

Copyright (c) 2004-2005  Riksarkivaren

=head1 LICENCE

Whatever Riksarkivaren chooses to do.

=cut

1;

