#!/usr/bin/perl
#-----------------------------------------------------------------------------
# $Id: Map.pm 797 2011-06-28 13:41:28Z tm $
#-----------------------------------------------------------------------------
# Noark::Schema::Map - Object for mapping between NOARK-4 and our db-model
#
# Author: Thomas Malt <thomas.malt@linpro.no>
#
# Copyright (c) 2006-2009 Riksarkivaren
#
#-----------------------------------------------------------------------------

=head1 NAME

Noark::Schema::Map - Object for mapping between NOARK-4 and our model

=head1 SYNOPSIS

TODO

=head1 DESCRIPTION

TODO

=cut

package Noark::Schema::Map;

use strict;
use warnings;

use Carp;
use Carp::Assert;

use Noark::Config qw($tables);

our ($VERSION) = q$Revision: 797 $ =~ /: (\S+)/;

my $self = undef; # Singleton


=head1 SUBROUTINES

=over

=item new() 

Constructor returns a new object. This module is a singleton, so if an
instance already exists a reference to a common instance will be returned

=cut
sub new {
    my ($class, %args) = @_;

    return $self if (defined $self && ref $self eq __PACKAGE__); # Singleton
    croak("Cannot clone object from self.") if ($class ne __PACKAGE__);

    $self = {};
    bless($self, $class);

    $self->{NOARK_TO_SCHEMA} = $tables;
    $self->{SCHEMA_TO_NOARK} = $self->_map_db_to_schema();
    return $self;
}


sub _map_db_to_schema {
    my $self = shift;

    my %map = ();
    while (my ($key, $value) = each %{$tables}) {
        $map{$value} = $key;
    }
    return \%map;
}

=item get_noark_table_name (<table name>)

Returns the table name as defined in Noark-4.1

=cut
sub get_noark_table_name {
    my ($self, $name) = @_;

    my $hit = $self->_map_lookup(type => 'NOARK', name => $name);
    $hit =~ s/.TAB$//;

    return $hit;
} 


=item get_schema_table_name (<table name>)

Returns the table name in our data model.

=cut
sub get_schema_table_name {
    my ($self, $name) = @_;

    $name .= '.TAB' if (!($name =~ m/.TAB$/));

    my $hit = $self->_map_lookup(type => 'SCHEMA', name => $name);

    return $hit;
}

=item is_noark_table_name ($name)

Returns TRUE if the given name is a valid Noark-4.1 table name. Otherwise
returns FALSE.

=cut
sub is_noark_table_name {
    my ($self, $name) = @_;

    croak("Table name must be defined") if ! defined($name);
    eval {
        $self->get_schema_table_name($name);
    };    
    return 0 if $@;
    return 1;
}

=item get_noark_field_name (<field name>)

Returns the field (or attribute if you like) name as it is defined in NOARK.

=cut
sub get_noark_field_name {
    my ($self, $name) = @_;
    
    return $self->_map_lookup(type => 'NOARK', name => $name);
}

=item get_schema_field_name (<field_name>)

Returns the field name as it is defined in our data model

=cut
sub get_schema_field_name {
    my ($self, $name) = @_;

    return $self->_map_lookup(type => 'SCHEMA', name => $name);
} 

sub _map_lookup {
    my ($self, %args) = @_;

    my $type  = uc $args{type};
    my $name  = $args{name};
    my $reply = undef;

    $reply = $self->{NOARK_TO_SCHEMA}->{$name} if $type eq 'SCHEMA';
    $reply = $self->{SCHEMA_TO_NOARK}->{$name} if $type eq 'NOARK';

    if (!defined($reply)) {
        croak( qq{Did not find correct mapping for type:'$type', }
             . qq{name: '$name'.\nThis should NEVER happen.}); #});  
    
    }

    return $reply;
} 

# sub benchmark {
#     my $self = shift;
# 
#     use Benchmark qw(cmpthese);
# 
#     my @db_to_schema = keys %{$self->{NOARK_TO_SCHEMA}};
#     my @schema_to_db = keys %{$self->{SCHEMA_TO_NOARK}};
# 
#     cmpthese (-10, 
#               {
#                'old_map' => sub { 
#                    for my $i (@db_to_schema) {
#                        _map_lookup($self, type => 'schema', name => $i)}
#                    for my $i (@schema_to_db) {
#                        _map_lookup($self, type => 'noark', name => $i);
#                    }
#                },
#                'new_map' => sub {
#                    for my $i (@db_to_schema) {
#                        _map_lookup_2($self, type => 'schema', name => $i)}
#                    for my $i (@schema_to_db) {
#                        _map_lookup_2($self, type => 'noark', name => $i);
#                    }
#                },
#               });
# }
#
 
sub VERSION {
    return $VERSION;
}

1;

__END__

=pod

=back

=cut

