#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: Table.pm 797 2011-06-28 13:41:28Z tm $
#---------------------------------------------------------------------
# 
# Author: Thomas Malt <thomas@malt.no>
# 
#
use warnings;
use strict;

package Noark::Schema::Table;

=head1 NAME

Noark::Schema::Table

=head1 SYNOPSIS

my $obj = Noark::Schema::Table->new();

=head1 DESCRIPTION

Holdes information about a Noark table.

=cut

use Noark::Schema::Attribute;
use Noark::Schema::Map;
use Data::Dumper;

=head1 SUBROUTINES

=over

=item new(verbosity => 0)

Instantiates a new Table object.

=cut
sub new {
    my $class = shift;
    my %args = @_;
    my $self = {};

    $self->{VERBOSE}      = $args{verbosity} || 0;
    $self->{ATTRIBUTES}   = {};
    $self->{NAME}         = undef;
    $self->{PRIMARY_KEY}  = [];
    $self->{FOREIGN_KEYS} = [];
    $self->{REFERENCES}   = [];

    bless ($self, $class);

    return $self;
}

=item parse

Parses a SQL tabel definition.

Dies on parse error.

=cut
sub parse {
    my ($self, $fh) = @_;

    for (;;) {
        $_ = readline($fh);
        print ("T: ", $_) if $self->{VERBOSE} == 4;
        last if index(lc($_), "create") >= 0;
    }

    /create\s+table\s+(\w+)/i;
    $self->{NAME} = $1;

    # parse attributes
    for (;;) {
        $_ = readline($fh);
        print "T: ", $_ if $self->{VERBOSE} == 4;
        /^\s*(?:\)|primary|index)/ and last;

       my $rec = Noark::Schema::Attribute->new(verbosity => $self->{VERBOSE});
       my $name = $rec->parse($_);
       $self->{ATTRIBUTES}->{$name} = $rec;

       if (/primary\s+key/i) {
            push(@{$self->{PRIMARY_KEY}}, $rec);
            $rec->set_primary_key_status(1);
        }
    }

    unless (@{$self->{PRIMARY_KEY}}) {
        /primary/i or die("Parse error on line $.: No primary key found!");
        /primary\s+key\s*\((.+)\)/i or die("Parse error on line $.: $_");

        for (split(/,\s*/, $1)) {
            defined $self->{ATTRIBUTES}->{$_} or 
              die("Parse error on line $.: `$_' is not an attribute.");
            push(@{$self->{PRIMARY_KEY}}, $self->{ATTRIBUTES}->{$_});
            $self->{ATTRIBUTES}->{$_}->set_primary_key_status(1);
        }
    }

    return $self->{NAME};
}


=item add_foreign_key($attr, $ref_tab, $ref_attr)

Stores a foreign key reference in the data model

=cut
sub add_foreign_key {
    my ($self, $attr, $ref_tab, $ref_attr) = @_;

    my $a = $self->get_attribute($attr);
    my $ra = $ref_tab->get_attribute($ref_attr);

    defined $a or die("Parse error on line $.: `$attr' is not an attribute in $self->{NAME}.");
    defined $ra or die("Parse error on line $.: `$ref_attr' is not an attribute in " . $ref_tab->get_name());

    push @{$self->{FOREIGN_KEYS}}, [$a, $ref_tab, $ra];
    $a->set_fk_reference($ref_tab, $ra);
}


=item add_reference($attr, $ref_tab, $ref_attr)

Stores a reference from another table _to_ this table in the model.

=cut
sub add_reference {
    my ($self, $attr, $ref_tab, $ref_attr) = @_;

    my $a = $self->get_attribute($attr);
    my $ra = $ref_tab->get_attribute($ref_attr);

    defined $a or die("Parse error on line $.: `$attr' is not an attribute in $self->{NAME}.");
    defined $ra or die("Parse error on line $.: `$ref_attr' is not an attribute in " . $ref_tab->get_name());

    push @{$self->{REFERENCES}}, [$a, $ref_tab, $ra];
    
}

=item get_attribute($name)

Returns the Attribute object with name `$name'.

=cut
sub get_attribute {
    my ($self, $name) = @_;
    
    if (not $self->has_attribute($name)) {
        my $map = Noark::Schema::Map->new();
        eval {
            $name = $map->get_schema_field_name($name);
        };
        return undef if $@;
    } 
    return $self->{ATTRIBUTES}->{$name};
}

=item get_attributes()

Returns all attribute objects as an HASHREF indexed by name.

=cut
sub get_attributes {
    my ($self) = shift;
    
    return $self->{ATTRIBUTES};
}

=item get_attribute_list()

Returns all attribute objects as a list (ARRAYREF)

=cut
sub get_attribute_list {
    my $self = shift;

    my @list = values %{$self->{ATTRIBUTES}};
    return \@list;
}


=item get_attribute_noark_names()

Returns an ARRAYREF containing all the attribute names of table according to
Noark-4.1

=cut
sub get_attribute_noark_names {
    my $self = shift;

    my @names = map { $_->get_noark_name() } @{$self->get_attribute_list()};
    return \@names;
}

=item get_exportable_attribute_list

Returns a list of all attribute objects that are exportable.

=cut
sub get_exportable_attribute_list {
    my $self = shift;

    my @list = ();
    for (@{$self->get_attribute_list}) {
        push @list, $_ if $_->is_exportable();
    }
    return \@list;
}

=item get_foreign_keys()

Returns all foreign keys in a table as an ARRAYREF.

=cut
sub get_foreign_keys {
    my ($self) = shift;
    return $self->{FOREIGN_KEYS};
}

=item get_references()

Returns all references to this table by attributes in other tables as
an ARRAYREF.

=cut
sub get_references {
    my $self = shift;
    return $self->{REFERENCES}
}

=item get_name

Returns the table's name.

=cut
sub get_name {
    my ($self) = @_;

    return $self->{NAME};
}

=item get_noark_name

Returns the noark name (<NAME>.TAB) of the table.

=cut
sub get_noark_name {
    my ($self) = @_;

    my $map = new Noark::Schema::Map;
    return $map->get_noark_table_name($self->{NAME});
}


=item get_not_null_attrs

FOO

=cut
sub get_not_null_attrs {
    my ($self) = @_;
    my %ret = ();

    while (my ($key, $val) = each %{$self->{ATTRIBUTES}}) {
        unless ($val->get_nullness()) {
            $ret{$key} = $val;
        }
    }

    return \%ret;
}


=item get_all_attrs

FOO

=cut
sub get_all_attrs {
    my ($self) = @_;

    return $self->{ATTRIBUTES};
}


=item has_attribute($name)

Returns TRUE if `$name' is an attribute in this table.

=cut
sub has_attribute {
    my ($self, $name) = @_;

    return defined $self->{ATTRIBUTES}->{$name};
}


=item is_mandatory

Returns true if the table is mandatory. This means it's requirement
type is 'O' according to Noark-4.1.

=cut
sub is_mandatory {
    my ($self) = @_;

    for (values %{$self->{ATTRIBUTES}}) {
        return 1 if $_->is_mandatory();
    }

    return 0;
}



=item is_rec_mandatory($name)

Returns TRUE if attribute identified by `$name' is mandatory.

=cut
sub is_rec_mandatory {
    my ($self, $name) = @_;

    return $self->get_attribute($name)->is_mandatory();
}


sub get_primary_key {
    my $self = shift;

    return $self->{PRIMARY_KEY};
}

=item get_pk_names

Returns a list with the names of the attributes in the primary key.

=cut
sub get_pk_names {
    my ($self) = @_;

    my @list = ();
    for (@{$self->{PRIMARY_KEY}}) {
        push @list, $_->get_name();
    }
    return \@list;
}

=item get_noark_pk_names

Get the name of the primary key as it is defined in Noark-4.1

=cut
sub get_noark_pk_names {
    my ($self) = @_;
    
    my @list = ();
    for (@{$self->{PRIMARY_KEY}}) {
        push @list, $_->get_noark_name();
    }

    return \@list;
}


=item get_pk_names_as_string

Returns a string with the concatenated keys as a string.

=cut
sub get_pk_names_as_string {
    my $self = shift;

    return join(', ', @{$self->get_pk_names()});
}


=item dump_me

Dumps the table definition to stdout.

=cut
sub dump_me {
    my ($self) = @_;

    print "CREATE TABLE ", $self->{NAME}, "(\n";
    for (keys(%{$self->{ATTRIBUTES}})) {
        print "    ";
        $self->{ATTRIBUTES}->{$_}->dump_me();
        print "\n";
    }
    print "    primary key (", join(', ', @{$self->get_pk_names()}), ")\n";
    print ");\n";

    #print Dumper($self->{FOREIGN_KEYS});

    for my $a (@{$self->{FOREIGN_KEYS}}) {
        print "ALTER TABLE $self->{NAME} ADD FOREIGN KEY (", $a->[0]->get_name();
        print ") REFERENCES ", $a->[1]->get_name(), "(", $a->[2]->get_name(), ")\n";
    }
    print "\n";
}


=pod

=back

=head1 SEE ALSO

Noark::Schema, Noark::Schema::Attribute

=head1 AUTHOR

    Kjell-Magne ierud <kjellm@linpro.no>
    Thomas Malt <thomas.malt@linpro.no>

=head1 COPYRIGHT

Copyright (c) 2004-2005  Riksarkivaren

=head1 LICENCE

Whatever Riksarkivaren chooses to do.

=cut

1;
