#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: 008.schema.t 579 2011-05-31 12:15:39Z tm $
#---------------------------------------------------------------------
# Schema tests - These tests test the Noark::Schema classes
#

use strict;
use warnings;

use Test::More tests => 61;
use Test::Differences;
use File::Basename;
use Data::Dumper;

BEGIN {
    use_ok('Noark::Schema');
}

END {
    unlink("/tmp/schema_test.out");
}

my $schema = Noark::Schema->new();
isa_ok($schema, 'Noark::Schema');

my @methods 
    = qw(new set_return_xml parse parse_ref get_table get_tables
         get_attribute
         get_mandatory_tables
         get_pk_names
         is_attr_not_null
         is_attr_null
         is_attr_auto_increment
         get_not_null_attrs
         get_all_attrs
         get_attr_data_type
         is_attr_data_valid
         is_bool_valid
         is_varchar_valid
         get_attr_requirement
         is_attr_mandatory
         is_attr_unique
         is_attr_exportable
         dump_me
         has_table); #);

can_ok($schema, @methods);

# tester get_pk_names
my $pk_names_noarksak =  ['SA.ID'];
eq_or_diff_data($schema->get_pk_names('NOARKSAK'), $pk_names_noarksak,
                "Testing get_pk_names 'NOARKSAK'");

my $pk_kpautoris = ['KA.ADRID', 'KA.TGKODE', 'KA.TILDATO'];
eq_or_diff_data($schema->get_pk_names('KPAUTORIS'), $pk_kpautoris,
                "Testing get_pk_names 'KPAUTORIS'");


# tester get_attribute
my @table_attribute_list = ([ qw(ADRTYPE   AT.BETEGN   1  )],
                            [ qw(POSTNR    PO.POSTSTED 1  )],
                            [ qw(ORDNPRINS OP.LTEKST   1  )],
                            [ qw(ADRPERSON PA.ADRID    1  )],
                            [ qw(UTDOKTYP  DU.MAL      1  )],
                            [ qw(NOARKSAK  SA.FUCK     0  )],
                            [ qw(KASKRAON  SA.DUCK     0  )],
                            [ qw(NUMSERIE  NU.AAR      1 0)]                        
                           );

for my $test (@table_attribute_list) {
    my ($tab, $attr, $status, $default) = @$test;
    my $obj_ref = $schema->get_attribute($tab, $attr);
    if ($status) {
        isa_ok($obj_ref, 'Noark::Schema::Attribute');
        is($obj_ref->get_noark_name(), $attr, "Tester navn == $attr");
        is($obj_ref->get_default_value(), $default, "Tester default value ($tab->$attr)");            
    }
    else {
        ok(!defined($obj_ref), "får undef når ugyldig tabell/attributt ($tab, $attr)");
    }
}

#-----------------------------------------------------------------------------
# Tester Noark::Schema::Table
#-----------------------------------------------------------------------------
my $notable = $schema->get_table('FLUBBER');
is ($notable, undef, "test get_table('FLUBBER') returns undef");
eval {my $notabl2 = $schema->get_table();};
ok ($@, "Test get_table(undef) croaks properly");
$@ = undef;

my $table = $schema->get_table('KLASSERING');
isa_ok($table, 'Noark::Schema::Table');
eq_or_diff($table->get_pk_names(), [qw(kl_said kl_ordnpri kl_ordnver)],
          'Testing Noark::Schema::Table::get_pk_names');
ok($table->get_pk_names_as_string() eq 'kl_said, kl_ordnpri, kl_ordnver',
  'Testing Noark::Schema::Table::get_pk_names_as_string');

ok( $schema->get_pk_names_as_string('KLASSERING') eq 
    'KL.SAID, KL.ORDNPRI, KL.ORDNVER',
    'Testing Noark::Schema::get_pk_names_as_string' );

#-----------------------------------------------------------------------------
# Tester is_tablename_valid + is_attributename_valid
#-----------------------------------------------------------------------------
is($schema->is_tablename_valid('NOARKSAK'), 1, "is_tablename_valid(NOARKSAK)");
is($schema->is_tablename_valid('FLUBBER'), 0, "is_tablename_valid(FLUBBER)");
$@ = undef;
eval {$schema->is_tablename_valid();};
ok($@, "is_tablename_valid(undef) croaks");
$@ = undef;

is($schema->is_attributename_valid('NOARKSAK', 'SA.ID'), 1, "is_attributename_valid(NOARKSAK, SA.ID)");
is($schema->is_attributename_valid('NOARKSAK', 'SA.FOOBAR'), 0, "is_attributename_valid(NOARKSAK, SA.FOOBAR)");
eval { $schema->is_attributename_valid('FLUBBER', 'SA.ID'); };
ok($@, "is_attributename - Croaks on invalid table");
$@ = undef;
eval { $schema->is_attributename_valid('FLUBBER', undef); };
ok($@, "is_attributename_valid - croaks on undef attribute");
$@ = undef;
eval { $schema->is_attributename_valid(undef, undef); };
ok($@, "is_attributename_valid - croaks on undef table and attribute");
$@ = undef;

#-----------------------------------------------------------------------------
# Tester validering av data typer:
#-----------------------------------------------------------------------------
my $types  = [qw(BLOB TEXT VARCHAR BOOL TIME DATE INT)];
eq_or_diff($types, $schema->get_all_data_types(), 
           'Testing get_all_data_types' );

my $set = {
           BLOB    => [ ["DIGISERT",   "SE.SERT",    "FOOBAR",          1], ],
           TEXT    => [ ["KLASSERING", "KL.MERKNAD", "Alt er lov",      1], ],
           VARCAHR => [ ['NOARKSAK',   'SA.ARKDEL',  'Arkdel 7',        1],
                        ['NOARKSAK',   'SA.ARKDEL',  '',                0],
                        ['NOARKSAK',   'SA.ARKDEL',  'Arkdel Arkdel',   0], ],
           BOOL    => [ ['NOARKSAK',   'SA.PAPIR',   '1',               1],
                        ['NOARKSAK',   'SA.PAPIR',   '',                1],
                        ['NOARKSAK',   'SA.PAPIR',   2,                 0], ],
           TIME    => [ ['MERKNAD',    'ME.REGKL',   '1345',            1],
                        ['MERKNAD',    'ME.REGKL',   '13:45',           1],
                        ['MERKNAD',    'ME.REGKL',   '45:13',           0],
                        ['MERKNAD',    'ME.REGKL',   'trettenf�rtifem', 0], ],
           DATE    => [ ['NOARKSAK',   'SA.DATO',    '20061122',        1],
                        ['NOARKSAK',   'SA.DATO',    '20071122',        1],
                        ['NOARKSAK',   'SA.DATO',    '20062211',        0], ],
           INT     => [ ['NOARKSAK',   'SA.ID',      '123456789',       1],
                        ['NOARKSAK',   'SA.ID',      '1234567901234',   0],
                        ['NOARKSAK',   'SA.ID',      'ab0eaebf',        0], ],
          };

for my $type (@{$types}) {
    for my $test (@{$set->{$type}}) {
        my ($table, $attr, $value, $valid) = @{$test};
        if ($valid) {
            ok($schema->is_attr_data_valid($table, $attr, $value),
               "Testing is_attr_data_valid('$table', '$attr', '$value')");
        } 
        else {
            ok(!$schema->is_attr_data_valid($table, $attr, $value),
               "Testing NOT is_attr_data_valid('$table', '$attr', '$value')");
            
        } 
        
    } 
} 

#-----------------------------------------------------------------------------
# tester get_attr_data_type
#-----------------------------------------------------------------------------
is($schema->get_attr_data_type('NUMSERIE', 'NU.ID'), "INT(10)", "get_attr_data_type - numserie nu_id");
is($schema->get_attr_data_type('NUMSERIE', 'NU.AAR'), "INT(4)", "get_attr_data_type - numserie nu_aar");
is($schema->get_attr_data_type('numserie', 'nu_betegn'), "VARCHAR(70)", "get_attr_data_type - numserie nu_betegn");
$@ = undef;
eval {$schema->get_attr_data_type('NUMSERIE', 'NU.FOO'); };
ok($@, "get_attr_data_type - croak on invalid attribute");
$@ = undef;
eval {$schema->get_attr_data_type('numXXserie', 'NU.FOO'); };
ok($@, "get_attr_data_type - croak on invalid table");
$@ = undef;

#-----------------------------------------------------------------------------
# Tester dump av hele greia
#-----------------------------------------------------------------------------
my $s = Noark::Schema->new(use_xml_names => 0);
my $new = "";
# {
#     no warnings;
# close(STDOUT);
open(OUT, ">/tmp/schema_test.out") or die("Oops: $!");
select(OUT);
$s->dump_me();
select(STDOUT);
close(OUT);
open(IN, "/tmp/schema_test.out") or die("Oops: $!");
while (<IN>) {
    $new .= $_;
}
close(IN);

my $scriptdir = dirname($0);
my $fh = undef;
{
    no warnings;
    open($fh, "$scriptdir/schema.out") or die ("no dump to test");
}
my $org = "";
while (<$fh>) {
    $org .= $_;
}
close($fh);
# print STDERR "-" x 72, "\n";
# print STDERR $org;
eq_or_diff_data($org, $new, "testing parsing of data structure.");

