#!/usr/bin/perl
#---------------------------------------------------------------------
# $Id: 060.noark_analyzer.t 780 2011-06-23 14:45:01Z tm $
#---------------------------------------------------------------------
# Automatic tester for ArkN4.
# These tests test the Analyzer object.
#
use strict;
use warnings;

use Time::HiRes qw(time);
use Data::Dumper;

use Test::More tests => 21;
use Test::Differences;
use Noark::Config qw($config);

BEGIN {
    use_ok('Noark::Analyzer');
}

my $an = Noark::Analyzer->new(
    dir    => "./uttrekk/test",
    xmldir    => 'DATA',
    docdir    => 'DOKUMENT',
    reportdir => 'RAPPORT',
    sqldir => "./database",
    dtddir => "./dtd",
    usedb  => 0
);

isa_ok($an, 'Noark::Analyzer');

my $info = $an->get_tabellinfo_list();
ok($an->is_noarkih_valid(), 'testing noarkih');
ok($an->is_dokver_valid(), 'testing dokver');
ok($an->is_journal_valid(), 'testing journal');
ok($an->is_sakdok_valid(), 'testing sakdok');
ok($an->_add_missing_journal_error(), 'testing add_missing_journal_error');
ok($an->_handle_no_noarkih_error(), 'testing _handle_no_noarkih_error');
ok($an->_add_sakdok_files_warning(), 'testing _add_sakdok_files_warning');
# ok($an->validate_import(), 'testing validate_import');
ok(!$an->are_files_valid($info->get_tabellinfo_by_name('NOARKSAK')), 'testing are_files_valid');

ok($an->is_file_valid("./uttrekk/test/DATA/ORDNPRI.XML"), 
   "Test 'Noark::Analyzer::is_file_valid(): valid XML with wrong DTD.");

ok(!$an->is_file_valid("./uttrekk/test/DATA/UTENDTD.XML"),
   "Test 'Noark::Analyzer::is_file_valid(): XML without valid DTD");

test_handle_deposited_table(); # Runs 2 tests

ok(test_get_xmlmeta_info(), 
   "Test 'Noark::Analyzer::get_xmlmeta_info()'");

ok(test_handle_bad_order(),
   "Test 'Noark::Analyzer::handle_bad_order()'");

#-----------------------------------------------------------------------------
# Test: 14-16: Noark::Analyzer::UnknownTableError
#-----------------------------------------------------------------------------
ok($an->check_for_nonexisting_tables(), "Checking nonexisting tables");

#---------------------------------------------------------------------
# SUBROUTINES
#---------------------------------------------------------------------

sub test_get_xmlmeta_info {

    my $info = $an->get_xmlmeta_info("ORDNPRI.XML");

    my $corr = {'DTDLINE' => 2,
                'DTDREF' => 'Re4LLyBaD.dtd',
                'FILE' => '/tmp/arkn4.'.$ENV{USER}.'.'.$$ 
                        . '/ORDNPRI.XML.'.$$.'.DTDFIX.XML',
                'LINE' => 94,
                'TABLE' => 'ORDNPRINS.TAB',};
    
    for (keys %{$info}) {
        return 0 unless ($info->{$_} eq $corr->{$_});
     }

    return 1;
}


sub test_handle_deposited_table {
    my $list = $an->get_tabellinfo_list();

    my $info_expect
        = bless( {
                  'FILES' => [
                              {
                               'EXISTS' => 1,
                               'UPCASE' => 1,
                               'VALID' => 0,
                               'XMLNAME' => 'NOARKSAK.XML',
                               'PRELEN' => 8,
                               'OK' => 1,
                               'BASENAME' => 'NOARKSAK.XML',
                               'TI.FILNAVN' => 'NOARKSAK.XML',
                               'SUFLEN' => 3,
                               'FPRELEN' => 8,
                               'TI.FILDEL' => 1,
                               'SAME' => 1,
                               'FSUFLEN' => 3,
                               'TI.ANTPOSTER' => '100'
                              },
                              {
                               'EXISTS' => 1,
                               'UPCASE' => 1,
                               'VALID' => 0,
                               'XMLNAME' => 'NOARKSA2.XML',
                               'PRELEN' => 8,
                               'OK' => 1,
                               'BASENAME' => 'NOARKSA2.XML',
                               'TI.FILNAVN' => 'NOARKSA2.XML',
                               'SUFLEN' => 3,
                               'FPRELEN' => 8,
                               'TI.FILDEL' => 2,
                               'SAME' => 1,
                               'FSUFLEN' => 3,
                               'TI.ANTPOSTER' => '32'
                              }
                            ],
                  'FILE_COUNT'   => 2,
                  'LINE_NUMBER'  => 134,
                  'TI.ANTFILER'  => 2,
                  'TI.TABELL'    => 'NOARKSAK',
                  'FILDEL_OK'    => 1,
                  'SORTED_FILES' => undef,
                  'ATTRIBUTTER'  => [
                                  'SA.DATO',
                                  'SA.ANTJP',
                                  'SA.ARKDEL',
                                  'SA.SAAR',
                                  'SA.ID',
                                  'SA.JENHET',
                                  'SA.SEKNR',
                                  'SA.PAPIR',
                                  'SA.SISTEJP',
                                  'SA.ADMID',
                                  'SA.TYPE',
                                  'SA.TITTEL',
                                  'SA.ANSVID',
                                  'SA.STATUS',
                                  'SA.TGKODE',
                                  'SA.U1',
                                  'SA.TGGRUPPE',
                                  'SA.UOFF',
                                  'SA.BEVTID',
                                  'SA.KASSKODE',
                                  'SA.PROSJEKT',
                                  'SA.KASSDATO',
                                  'SA.FRARKDEL'
                                ],
                 'TI.COUNT' => 1
               }, 'Noark::Parser::Tabellinfo' );
 

    
    my $info = $list->get_tabellinfo_by_name('NOARKSAK');    
    $an->check_tabellinfo( $info );

    $info_expect->{SORTED_FILES} = [$info_expect->{'FILES'}[0],
                                    $info_expect->{'FILES'}[1],];
    # $info_expect->{ELEMENT}      = $info->{ELEMENT};

    eq_or_diff($info, $info_expect, "Testing info");

    ok( !$an->handle_deposited_table($info), "Testing handle_deposited_table");
}



sub test_handle_bad_order {
    my $msg = <<EOF;
bad order of Elements Found=[SA.ID,SA.SAAR,SA.SEKNR,SA.PAPIR,SA.DATO,SA.TITTEL,SA.U1,SA.STATUS,SA.ARKDEL,SA.TYPE,SA.ADMID,SA.ANSVID,SA.TGKODE,SA.UOFF,SA.TGGRUPPE] RE=[((SA.ID)(SA.SAAR)(SA.SEKNR)(SA.PAPIR)?(SA.DATO)(SA.TITTEL)(SA.U1)(SA.STATUS)(SA.ARKDEL)(SA.TYPE)?(SA.JENHET)?(SA.OPSAKSDEL)?(SA.ADMID)(SA.ANSVID)(SA.TGKODE)?(SA.UOFF)?(SA.TGGRUPPE)?(SA.SISTEJP)?(SA.ANTJP)(SA.BEVTID)?(SA.KASSKODE)?(SA.KASSDATO)?(SA.PROSJEKT)?(SA.PRES)?(SA.OBS)?(SA.FRARKDEL)?(SA.UTLDATO)?(SA.UTLTIL)?)]
EOF
    my %info = (
                'byte' => 6452,
                'line' => 205,
                'column' => 3,
               );

    $an->empty_current_errorlist();
    my $res = $an->handle_bad_order($msg, 'NOARKSA2.XML', %info);

    return 0 if $res;

    my $err = $an->get_current_errorlist();
    
    return 0 if (scalar(@{$err}) > 1);

    my $e = shift @$err;
    isa_ok($e, 'Noark::Error::XMLMissingMandatoryAttribute');
    is(scalar(@{$e->get_attributes()}), 1, "count number of attributes");
    for my $row (@{$e->get_attributes()}) {
        my ($key, $value) = each(%$row);
        is($key, 'SA.ANTJP', "attribute is sa.antjp");
        is($value, 205, "line is 205");        
    }

    return 1;
}


